
(** This file contains the definition of an OWhile, a While language
  with some very basic JavaScript-like pointer manipulation. **)

Set Implicit Arguments.
Require Export Utf8.
Require Export PrettyBigStep.

Require Export LibHeap TLC.LibStream.

Module Heap := HeapList.


Section OWhile.

(** * OWhile Definition **)

(** ** Syntax **)

Definition basic_val := int.
Definition var := string.
Definition property := string.

Inductive operator :=
  | op_add
  | op_sub
  | op_greater
  .

Inductive expr :=
  | expr_cst : basic_val -> expr
  | expr_random : expr
  | expr_var : var -> expr

  | expr_new_obj : expr
  | expr_property : expr -> property -> expr
  | expr_in : property -> expr -> expr

  | expr_op : expr -> operator -> expr -> expr
  | expr_not : expr -> expr
  | expr_eq : expr -> expr -> expr

  | expr_lambda : var -> prog -> expr
  | expr_app : expr -> expr -> expr

with stat :=
  | stat_skip : stat
  | stat_seq : stat -> stat -> stat

  | stat_if : expr -> stat -> stat -> stat
  | stat_while : expr -> stat -> stat
  | stat_throw : stat

  | stat_property_asgn : expr -> property -> expr -> stat
  | stat_delete : expr -> property -> stat

  | stat_return : expr -> stat
  | stat_run : expr -> stat

with prog :=
  | prog_stat : stat -> prog
  .

Scheme expr_ind' := Induction for expr Sort Prop
  with stat_ind' := Induction for stat Sort Prop
  with prog_ind' := Induction for prog Sort Prop.
Combined Scheme expr_stat_prog_mutind from expr_ind', stat_ind', prog_ind'.

Inductive ext_expr :=
  | expr_property_1 : property -> ext_expr
  | expr_in_1 : property -> ext_expr

  | expr_op_1 : operator -> expr -> ext_expr
  | expr_op_2 : operator -> ext_expr
  | expr_not_1 : ext_expr
  | expr_eq_1 : expr -> ext_expr
  | expr_eq_2 : ext_expr

  | expr_app_1 : expr -> ext_expr
  | expr_app_2 : ext_expr
  | expr_app_3 : ext_expr
  .

Inductive ext_stat :=
  | stat_seq_1 : stat -> ext_stat

  | stat_if_1 : stat -> stat -> ext_stat
  | stat_while_1 : expr -> stat -> ext_stat
  | stat_while_2 : expr -> stat -> ext_stat

  | stat_property_asgn_1 : property -> expr -> ext_stat
  | stat_property_asgn_2 : property -> ext_stat
  | stat_delete_1 : property -> ext_stat

  | stat_return_1 : ext_stat
  | stat_run_1 : ext_stat
  .

Inductive owhile_terms :=
  | owhile_terms_expr : expr -> owhile_terms
  | owhile_terms_ext_expr : ext_expr -> owhile_terms
  | owhile_terms_stat : stat -> owhile_terms
  | owhile_terms_ext_stat : ext_stat -> owhile_terms
  | owhile_terms_prog : prog -> owhile_terms
  .

Coercion owhile_terms_expr : expr >-> owhile_terms.
Coercion owhile_terms_ext_expr : ext_expr >-> owhile_terms.
Coercion owhile_terms_stat : stat >-> owhile_terms.
Coercion owhile_terms_ext_stat : ext_stat >-> owhile_terms.
Coercion owhile_terms_prog : prog >-> owhile_terms.


Global Instance operator_comparable : Comparable operator.
  prove_comparable_simple_inductive. Defined.


Fixpoint compare_expr e1 e2 :=
  match e1, e2 with
  | expr_cst c1, expr_cst c2 => decide (c1 = c2)
  | expr_random, expr_random => true
  | expr_var x1, expr_var x2 => decide (x1 = x2)

  | expr_new_obj, expr_new_obj => true
  | expr_property e1 f1, expr_property e2 f2 =>
    decide (f1 = f2) && compare_expr e1 e2
  | expr_in f1 e1, expr_in f2 e2 =>
    decide (f1 = f2) && compare_expr e1 e2

  | expr_op e11 op1 e21, expr_op e12 op2 e22 =>
    decide (op1 = op2) && compare_expr e11 e12 && compare_expr e21 e22
  | expr_not e1, expr_not e2 =>
    compare_expr e1 e2
  | expr_eq e11 e21, expr_eq e12 e22 =>
    compare_expr e11 e12 && compare_expr e21 e22

  | expr_lambda x1 p1, expr_lambda x2 p2 =>
    decide (x1 = x2) && compare_prog p1 p2
  | expr_app e11 e21, expr_app e12 e22 =>
    compare_expr e11 e12 && compare_expr e21 e22

  | _, _ => false
  end

with compare_stat s1 s2 :=
  match s1, s2 with
  | stat_skip, stat_skip => true
  | stat_seq s11 s12, stat_seq s21 s22 =>
    compare_stat s11 s21 && compare_stat s12 s22

  | stat_if e1 s11 s12, stat_if e2 s21 s22 =>
    compare_expr e1 e2 && compare_stat s11 s21 && compare_stat s12 s22
  | stat_while e1 s1, stat_while e2 s2 =>
    compare_expr e1 e2 && compare_stat s1 s2
  | stat_throw, stat_throw => true

  | stat_property_asgn e11 f1 e12, stat_property_asgn e21 f2 e22 =>
    decide (f1 = f2) && compare_expr e11 e21 && compare_expr e12 e22
  | stat_delete e1 f1, stat_delete e2 f2 =>
    decide (f1 = f2) && compare_expr e1 e2

  | stat_return e1, stat_return e2 =>
    compare_expr e1 e2
  | stat_run e1, stat_run e2 =>
    compare_expr e1 e2

  | _, _ => false
  end

with compare_prog p1 p2 : bool :=
  match p1, p2 with
  | prog_stat s1, prog_stat s2 =>
    compare_stat s1 s2
  end.


Lemma expr_stat_prog_compare_correct :
  (forall e1 e2, compare_expr e1 e2 ↔ e1 = e2)
  ∧ (forall s1 s2, compare_stat s1 s2 ↔ s1 = s2)
  ∧ (forall p1 p2, compare_prog p1 p2 ↔ p1 = p2).
Proof.
  apply expr_stat_prog_mutind;
    intros; let t := get_last_hyp tt in destruct t; simpls;
    repeat (rew_refl in *; try rewrite <- and_andb in *);
    iff; jauto_set; tryfalse; substs~; fequals*;
    repeat match goal with H: _ = _ |- _ => inverts H end;
    solve [ apply* H | apply* H0 | apply* H1 ].
Qed.

Global Instance expr_Comparable : Comparable expr.
  applys comparable_beq compare_expr. apply expr_stat_prog_compare_correct.
Defined.

Global Instance stat_Comparable : Comparable stat.
  applys comparable_beq compare_stat. apply expr_stat_prog_compare_correct.
Defined.

Global Instance prog_Comparable : Comparable prog.
  applys comparable_beq compare_prog. apply expr_stat_prog_compare_correct.
Defined.

Global Instance ext_expr_Comparable : Comparable ext_expr.
  prove_comparable_simple_inductive. Defined.

Global Instance ext_stat_Comparable : Comparable ext_stat.
  prove_comparable_simple_inductive. Defined.

Global Instance owhile_terms_Comparable : Comparable owhile_terms.
  prove_comparable_simple_inductive. Defined.


Global Instance expr_Inhab : Inhab expr.
  applys prove_Inhab expr_random. Defined.

Global Instance stat_Inhab : Inhab stat.
  applys prove_Inhab stat_skip. Defined.

Global Instance prog_Inhab : Inhab prog.
  applys prove_Inhab (prog_stat arbitrary). Defined.


(** ** Structure **)

Inductive owhile_name :=
  | name_abort_ext_expr : ext_expr -> owhile_name
  | name_abort_ext_stat : ext_stat -> owhile_name

  | name_cst : basic_val -> owhile_name
  | name_random : bool -> owhile_name
  | name_var : var -> owhile_name
  | name_varUndef : var -> owhile_name

  | name_new_obj : owhile_name
  | name_property : expr -> property -> owhile_name
  | name_property_1 : property -> owhile_name
  | name_property_1_no_loc : property -> owhile_name
  | name_property_1_undef : property -> owhile_name
  | name_in : property -> expr -> owhile_name
  | name_in_1_true : property -> owhile_name
  | name_in_1_false : property -> owhile_name
  | name_in_1_no_loc : property -> owhile_name

  | name_op : expr -> operator -> expr -> owhile_name
  | name_op_1 : operator -> expr -> owhile_name
  | name_op_1_error : operator -> expr -> owhile_name
  | name_op_2_error : operator -> owhile_name
  | name_add_2 : owhile_name
  | name_sub_2 : owhile_name
  | name_greater_2_greater : owhile_name
  | name_greater_2_lesser_eq : owhile_name
  | name_not : expr -> owhile_name
  | name_not_1_true : owhile_name
  | name_not_1_false : owhile_name
  | name_not_1_error : owhile_name
  | name_eq : expr -> expr -> owhile_name
  | name_eq_1 : expr -> owhile_name
  | name_eq_2_basic_val_eq : owhile_name
  | name_eq_2_basic_val_neq : owhile_name
  | name_eq_2_loc_eq : owhile_name
  | name_eq_2_loc_neq : owhile_name
  | name_eq_2_mistype_basic_val_loc : owhile_name
  | name_eq_2_mistype_loc_basic_val : owhile_name
  | name_eq_2_closure : owhile_name

  | name_lambda : var -> prog -> owhile_name
  | name_app : expr -> expr -> owhile_name
  | name_app_1 : expr -> owhile_name
  | name_app_1_error : expr -> owhile_name
  | name_app_2 : prog -> owhile_name
  | name_app_3_return : owhile_name
  | name_app_3_no_return : owhile_name

  | name_skip : owhile_name
  | name_seq : stat -> stat -> owhile_name
  | name_seq_1 : stat -> owhile_name
  | name_throw : owhile_name

  | name_if : expr -> stat -> stat -> owhile_name
  | name_if_1_true : stat -> stat -> owhile_name
  | name_if_1_false : stat -> stat -> owhile_name
  | name_if_1_error : stat -> stat -> owhile_name
  | name_while : expr -> stat -> owhile_name
  | name_while_1 : expr -> stat -> owhile_name
  | name_while_2_true : expr -> stat -> owhile_name
  | name_while_2_false : expr -> stat -> owhile_name
  | name_while_2_error : expr -> stat -> owhile_name

  | name_property_asgn : expr -> property -> expr -> owhile_name
  | name_property_asgn_1 : property -> expr -> owhile_name
  | name_property_asgn_1_error : property -> expr -> owhile_name
  | name_property_asgn_2 : property -> owhile_name
  | name_delete : expr -> property -> owhile_name
  | name_delete_1 : property -> owhile_name
  | name_delete_1_error : property -> owhile_name

  | name_return : expr -> owhile_name
  | name_return_1 : owhile_name
  | name_run : expr -> owhile_name
  | name_run_1 : owhile_name

  | name_prog_stat : stat -> owhile_name
  .


Definition owhile_left n : owhile_terms :=
  match n with
  | name_abort_ext_expr e => e
  | name_abort_ext_stat s => s

  | name_cst c => expr_cst c
  | name_random b => expr_random
  | name_var x => expr_var x
  | name_varUndef x => expr_var x

  | name_new_obj => expr_new_obj
  | name_property e f => expr_property e f
  | name_property_1 f => expr_property_1 f
  | name_property_1_no_loc f => expr_property_1 f
  | name_property_1_undef f => expr_property_1 f
  | name_in f e => expr_in f e
  | name_in_1_true f => expr_in_1 f
  | name_in_1_false f => expr_in_1 f
  | name_in_1_no_loc f => expr_in_1 f

  | name_op e1 op e2 => expr_op e1 op e2
  | name_op_1 op e2 => expr_op_1 op e2
  | name_op_1_error op e2 => expr_op_1 op e2
  | name_op_2_error op => expr_op_2 op
  | name_add_2 => expr_op_2 op_add
  | name_sub_2 => expr_op_2 op_sub
  | name_greater_2_greater => expr_op_2 op_greater
  | name_greater_2_lesser_eq => expr_op_2 op_greater
  | name_not e => expr_not e
  | name_not_1_true => expr_not_1
  | name_not_1_false => expr_not_1
  | name_not_1_error => expr_not_1
  | name_eq e1 e2 => expr_eq e1 e2
  | name_eq_1 e2 => expr_eq_1 e2
  | name_eq_2_basic_val_eq => expr_eq_2
  | name_eq_2_basic_val_neq => expr_eq_2
  | name_eq_2_loc_eq => expr_eq_2
  | name_eq_2_loc_neq => expr_eq_2
  | name_eq_2_mistype_basic_val_loc => expr_eq_2
  | name_eq_2_mistype_loc_basic_val => expr_eq_2
  | name_eq_2_closure => expr_eq_2

  | name_lambda x p => expr_lambda x p
  | name_app e1 e2 => expr_app e1 e2
  | name_app_1 e2 => expr_app_1 e2
  | name_app_1_error e2 => expr_app_1 e2
  | name_app_2 p => expr_app_2
  | name_app_3_return => expr_app_3
  | name_app_3_no_return => expr_app_3

  | name_skip => stat_skip
  | name_seq s1 s2 => stat_seq s1 s2
  | name_seq_1 s2 => stat_seq_1 s2
  | name_throw => stat_throw

  | name_if e s1 s2 => stat_if e s1 s2
  | name_if_1_true s1 s2 => stat_if_1 s1 s2
  | name_if_1_false s1 s2 => stat_if_1 s1 s2
  | name_if_1_error s1 s2 => stat_if_1 s1 s2
  | name_while e s => stat_while e s
  | name_while_1 e s => stat_while_1 e s
  | name_while_2_true e s => stat_while_2 e s
  | name_while_2_false e s => stat_while_2 e s
  | name_while_2_error e s => stat_while_2 e s

  | name_property_asgn e1 f e2 => stat_property_asgn e1 f e2
  | name_property_asgn_1 f e2 => stat_property_asgn_1 f e2
  | name_property_asgn_1_error f e2 => stat_property_asgn_1 f e2
  | name_property_asgn_2 f => stat_property_asgn_2 f
  | name_delete e f => stat_delete e f
  | name_delete_1 f => stat_delete_1 f
  | name_delete_1_error f => stat_delete_1 f

  | name_return e => stat_return e
  | name_return_1 => stat_return_1
  | name_run e => stat_run e
  | name_run_1 => stat_run_1

  | name_prog_stat s => prog_stat s
  end.

Definition owhile_struct n : Rule_struct owhile_terms :=
  match n with
  | name_abort_ext_expr e => Rule_struct_Ax _
  | name_abort_ext_stat s => Rule_struct_Ax _

  | name_cst c => Rule_struct_Ax _
  | name_random b => Rule_struct_Ax _
  | name_var x => Rule_struct_Ax _
  | name_varUndef x => Rule_struct_Ax _

  | name_new_obj => Rule_struct_Ax _
  | name_property e f => Rule_struct_R2 (term := owhile_terms) e (expr_property_1 f)
  | name_property_1 f => Rule_struct_Ax _
  | name_property_1_no_loc f => Rule_struct_Ax _
  | name_property_1_undef f => Rule_struct_Ax _
  | name_in f e => Rule_struct_R2 (term := owhile_terms) e (expr_in_1 f)
  | name_in_1_true f => Rule_struct_Ax _
  | name_in_1_false f => Rule_struct_Ax _
  | name_in_1_no_loc f => Rule_struct_Ax _

  | name_op e1 op e2 => Rule_struct_R2 (term := owhile_terms) e1 (expr_op_1 op e2)
  | name_op_1 op e2 => Rule_struct_R2 (term := owhile_terms) e2 (expr_op_2 op)
  | name_op_1_error op e2 => Rule_struct_Ax _
  | name_op_2_error op => Rule_struct_Ax _
  | name_add_2 => Rule_struct_Ax _
  | name_sub_2 => Rule_struct_Ax _
  | name_greater_2_greater => Rule_struct_Ax _
  | name_greater_2_lesser_eq => Rule_struct_Ax _
  | name_not e => Rule_struct_R2 (term := owhile_terms) e expr_not_1
  | name_not_1_true => Rule_struct_Ax _
  | name_not_1_false => Rule_struct_Ax _
  | name_not_1_error => Rule_struct_Ax _
  | name_eq e1 e2 => Rule_struct_R2 (term := owhile_terms) e1 (expr_eq_1 e2)
  | name_eq_1 e2 => Rule_struct_R2 (term := owhile_terms) e2 expr_eq_2
  | name_eq_2_basic_val_eq => Rule_struct_Ax _
  | name_eq_2_basic_val_neq => Rule_struct_Ax _
  | name_eq_2_loc_eq => Rule_struct_Ax _
  | name_eq_2_loc_neq => Rule_struct_Ax _
  | name_eq_2_mistype_basic_val_loc => Rule_struct_Ax _
  | name_eq_2_mistype_loc_basic_val => Rule_struct_Ax _
  | name_eq_2_closure => Rule_struct_Ax _

  | name_lambda x p => Rule_struct_Ax _
  | name_app e1 e2 => Rule_struct_R2 (term := owhile_terms) e1 (expr_app_1 e2)
  | name_app_1 e2 => Rule_struct_R2 (term := owhile_terms) e2 expr_app_2
  | name_app_1_error e2 => Rule_struct_Ax _
  | name_app_2 p => Rule_struct_R2 (term := owhile_terms) p expr_app_3
  | name_app_3_return => Rule_struct_Ax _
  | name_app_3_no_return => Rule_struct_Ax _

  | name_skip => Rule_struct_Ax _
  | name_seq s1 s2 => Rule_struct_R2 (term := owhile_terms) s1 (stat_seq_1 s2)
  | name_seq_1 s2 => Rule_struct_R1 (term := owhile_terms) s2
  | name_throw => Rule_struct_Ax _

  | name_if e s1 s2 => Rule_struct_R2 (term := owhile_terms) e (stat_if_1 s1 s2)
  | name_if_1_true s1 s2 => Rule_struct_R1 (term := owhile_terms) s1
  | name_if_1_false s1 s2 => Rule_struct_R1 (term := owhile_terms) s2
  | name_if_1_error s1 s2 => Rule_struct_Ax _
  | name_while e s => Rule_struct_R1 (term := owhile_terms) (stat_while_1 e s)
  | name_while_1 e s => Rule_struct_R2 (term := owhile_terms) e (stat_while_2 e s)
  | name_while_2_true e s => Rule_struct_R2 (term := owhile_terms) s (stat_while_1 e s)
  | name_while_2_false e s => Rule_struct_Ax _
  | name_while_2_error e s => Rule_struct_Ax _

  | name_property_asgn e1 f e2 => Rule_struct_R2 (term := owhile_terms) e1 (stat_property_asgn_1 f e2)
  | name_property_asgn_1 f e2 => Rule_struct_R2 (term := owhile_terms) e2 (stat_property_asgn_2 f)
  | name_property_asgn_1_error f e2 => Rule_struct_Ax _
  | name_property_asgn_2 f => Rule_struct_Ax _
  | name_delete e f => Rule_struct_R2 (term := owhile_terms) e (stat_delete_1 f)
  | name_delete_1 f => Rule_struct_Ax _
  | name_delete_1_error f => Rule_struct_Ax _

  | name_return e => Rule_struct_R2 (term := owhile_terms) e stat_return_1
  | name_return_1 => Rule_struct_Ax _
  | name_run e => Rule_struct_R2 (term := owhile_terms) e stat_run_1
  | name_run_1 => Rule_struct_Ax _

  | name_prog_stat s => Rule_struct_R1 (term := owhile_terms) s
  end.


Definition owhile_str := {|

    term := owhile_terms ;
    name := owhile_name ;

    left := owhile_left ;

    rule_struct := owhile_struct

  |}.


(** ** Rules **)

(** *** Semantic Definitions **)

(** Object locations **)
Definition loc := nat.

(** Closure locations (they point to an environment) **)
Definition floc := nat.

(** Values **)
Inductive val :=
  | val_basic : basic_val -> val
  | val_loc : loc -> val
  | val_closure : var -> prog -> floc -> val
  .

Global Instance val_Inhab : Inhab val.
  apply prove_Inhab. repeat constructors.
Defined.

Definition Env := Heap.heap var val.

Definition fHeap := Heap.heap floc Env.

Definition obj := Heap.heap property val.

Definition Heap := Heap.heap loc obj.

Inductive error :=
  | Err_undefined_variable : var -> error
  | Err_val_basic_no_property : property -> error
  | Err_undefined_property : property -> error
  | Err_not_a_basic_value
  | Err_comparing_values_of_different_types
  | Err_comparing_functions
  | Err_not_a_function
  | Err_no_return_at_the_end_of_a_function
  | Err_throw
  | Err_write_property_of_something_not_object : property -> error
  | Err_delete_property_of_something_not_object : property -> error
  .

Global Instance error_Inhab : Inhab error.
  apply prove_Inhab. repeat constructors.
Defined.

Record State := {
    state_heap : Heap ;
    state_closures : fHeap ;
    (** These unfortunately can't be streams, as they have to be comparable. **)
    state_fresh_loc : loc ;
    state_fresh_floc : loc
  }.

Definition emptyState := {|
    state_heap := Heap.empty ;
    state_closures := Heap.empty ;
    state_fresh_loc := 1 ;
    state_fresh_floc := 1
  |}.

Definition alloc_state (St : State) :=
  let l := state_fresh_loc St in
  let St := {|
      state_heap := write (state_heap St) l Heap.empty ;
      state_closures := state_closures St ;
      state_fresh_loc := S l ;
      state_fresh_floc := state_fresh_floc St
    |} in
  (St, l).

Definition initialEnv : Env :=
  let l := snd (alloc_state emptyState) in
  write Heap.empty "global"%string (val_loc l).

Definition initialState :=
  let St := fst (alloc_state emptyState) in
  let fl := state_fresh_floc St in {|
    state_heap := state_heap St ;
    state_closures := write (state_closures St) fl initialEnv ;
    state_fresh_loc := state_fresh_loc St ;
    state_fresh_floc := S fl
  |}.

Definition initialFloc :=
  state_fresh_floc (fst (alloc_state emptyState)).


Definition indom_property St l f : Prop :=
  exists o, binds (state_heap St) l o ∧ indom o f.

Global Instance indom_property_decidable : forall St l f,
  Decidable (indom_property St l f).
Proof.
  introv. destruct (read_option (state_heap St) l) eqn:E.
   apply read_option_binds in E. applys decidable_make (decide (indom o f)).
    rew_refl. fequals. extens. iff I.
     exists~ o.
     lets (o'&B&I'): (rm I). forwards: @binds_func E B; try typeclass. substs~.
   apply read_option_not_indom in E. applys decidable_make false.
    fold_bool. rew_refl. introv (o&B&I). apply E. applys @binds_indom B.
Defined.


Inductive oute :=
  | oute_val : State -> val -> oute
  | oute_error : error -> oute
  .

Coercion oute_error : error >-> oute.

Definition is_val o :=
  match o with
  | oute_val St v => True
  | oute_error err => False
  end.

Definition is_basic o :=
  match o with
  | oute_val St v =>
    match v with
    | val_basic v => True
    | val_loc l => False
    | val_closure v p l => False
    end
  | oute_error err => False
  end.

(** Defined but not an basic value. **)
Definition is_not_basic o :=
  match o with
  | oute_val St v =>
    match v with
    | val_basic v => False
    | val_loc l => True
    | val_closure v p l => True
    end
  | oute_error err => False
  end.

Definition is_loc o :=
  match o with
  | oute_val St v =>
    match v with
    | val_basic v => False
    | val_loc l =>
      (** Even if this is correct because of the semantics, we have to
         also check that the location is well-defined. **)
      indom (state_heap St) l
    | val_closure v p l => False
    end
  | oute_error err => False
  end.

(** Defined but not an object location. **)
Definition is_not_loc o :=
  match o with
  | oute_val St v =>
    match v with
    | val_basic v => True
    | val_loc l => False
    | val_closure v p l => True
    end
  | oute_error err => False
  end.

Definition is_closure o :=
  match o with
  | oute_val St v =>
    match v with
    | val_basic v => False
    | val_loc l => False
    | val_closure v p l =>
      (** Even if this is true by observing closely the semantics,
         we have to also check that the location is well-defined. **)
      indom (state_closures St) l
    end
  | oute_error err => False
  end.

(** Defined but not a function location. **)
Definition is_not_closure o :=
  match o with
  | oute_val St v =>
    match v with
    | val_basic v => True
    | val_loc l => True
    | val_closure v p l => False
    end
  | oute_error err => False
  end.

(** A simpler version of [is_closure] when we do not need to fetch the environment. **)
Definition val_is_closure v :=
  match v with
  | val_basic v => False
  | val_loc l => False
  | val_closure v p l => True
  end.

Definition is_defined_oute l o :=
  match o with
  | oute_val St v =>
    indom (state_heap St) l
  | oute_error err => False
  end.

Definition env_is_defined_in_oute fl o :=
  match o with
  | oute_val St v =>
    indom (state_closures St) fl
  | oute_error err => False
  end.


Inductive aborte : oute -> Prop :=
  | aborte_error : forall err,
    aborte (oute_error err)
  .


Inductive outs :=
  | outs_normal : State -> outs
  | outs_return : State -> val -> outs
  | outs_error : error -> outs
  .

Coercion outs_error : error >-> outs.

Definition is_normal o :=
  match o with
  | outs_normal St => True
  | outs_return St v => False
  | outs_error err => False
  end.

Definition is_return o :=
  match o with
  | outs_normal St => False
  | outs_return St v => True
  | outs_error err => False
  end.

Inductive aborts : outs -> Prop :=
  | aborts_error : forall err,
    aborts (outs_error err)
  | aborts_return : forall St v,
    aborts (outs_return St v)
  .


Definition env_is_defined_in_outs fl o :=
  match o with
  | outs_normal St =>
    indom (state_closures St) fl
  | outs_return St v =>
    indom (state_closures St) fl
  | outs_error err => False
  end.


Inductive owhile_res : Type :=
  | owhile_res_expr : oute -> owhile_res
  | owhile_res_stat : outs -> owhile_res
  .

Coercion owhile_res_expr : oute >-> owhile_res.
Coercion owhile_res_stat : outs >-> owhile_res.

Inductive abort_res : owhile_res -> Prop :=
  | abort_res_expr : forall o,
      aborte o →
      abort_res o
  | abort_res_stat : forall o,
      aborts o →
      abort_res o
  .

Inductive owhile_st : Type :=
  | owhile_st_expr : floc -> State -> owhile_st
  | owhile_st_stat : floc -> State -> owhile_st
  | owhile_st_prog : floc -> State -> owhile_st

  | owhile_st_in_1 : owhile_res -> owhile_st

  | owhile_st_property_1 : owhile_res -> owhile_st
  | owhile_st_op_1 : floc -> owhile_res -> owhile_st
  | owhile_st_op_2 : basic_val -> owhile_res -> owhile_st
  | owhile_st_not_1 : owhile_res -> owhile_st
  | owhile_st_eq_1 : floc -> owhile_res -> owhile_st
  | owhile_st_eq_2 : val -> owhile_res -> owhile_st

  | owhile_st_app_1 : floc -> owhile_res -> owhile_st
  | owhile_st_app_2 : var -> prog -> floc -> owhile_res -> owhile_st
  | owhile_st_app_3 : owhile_res -> owhile_st

  | owhile_st_seq_1 : floc -> owhile_res -> owhile_st

  | owhile_st_if_1 : floc -> owhile_res -> owhile_st
  | owhile_st_while_1 : floc -> owhile_res -> owhile_st
  | owhile_st_while_2 : floc -> owhile_res -> owhile_st

  | owhile_st_property_asgn_1 : floc -> owhile_res -> owhile_st
  | owhile_st_property_asgn_2 : loc -> owhile_res -> owhile_st
  | owhile_st_delete_1 : owhile_res -> owhile_st

  | owhile_st_return_1 : owhile_res -> owhile_st
  | owhile_st_run_1 : owhile_res -> owhile_st
  .

Definition abort_st : owhile_st -> Prop.
  introv sigma. destruct sigma;
    match goal with
    | o : owhile_res |- _ =>
      exact (abort_res o)
    | _ => exact False
    end.
Defined.

Inductive abort_intercepted : owhile_terms -> owhile_st -> Prop :=
  | abort_intercepted_app_3 : forall St v,
    abort_intercepted expr_app_3 (owhile_st_app_3 (outs_return St v))
  .


(** *** Side-conditions **)

(** Side conditions have to ensure that the operations which we want
  to perfom in the transfer function can be applied. They are thus
  divided into two parts: [owhile_cond_well_formed] verifies that the
  conditions required by [owhile_rule] are met, while
  [owhile_cond_applies] verifies that the rule should indeed
  apply. **)

Definition owhile_cond_well_formed n sigma : Prop :=
  match n, sigma with
  | name_abort_ext_expr e, sigma =>
      True
  | name_abort_ext_stat s, sigma =>
      True

  | name_cst c, owhile_st_expr fl St =>
      indom (state_closures St) fl
  | name_random b, owhile_st_expr fl St =>
      indom (state_closures St) fl
  | name_var x, owhile_st_expr fl St =>
      indom (state_closures St) fl
  | name_varUndef x, owhile_st_expr fl St =>
      indom (state_closures St) fl

  | name_new_obj, owhile_st_expr fl St =>
      indom (state_closures St) fl
  | name_property e f, owhile_st_expr fl St =>
      indom (state_closures St) fl
  | name_property_1 f, owhile_st_property_1 (oute_val St (val_loc l)) =>
      indom_property St l f
  | name_property_1_no_loc f, owhile_st_property_1 (owhile_res_expr o) =>
      True
  | name_property_1_undef f, owhile_st_property_1 (oute_val St (val_loc l)) =>
      indom (state_heap St) l
  | name_in f e, owhile_st_expr fl St =>
      indom (state_closures St) fl
  | name_in_1_true f, owhile_st_in_1 (oute_val St (val_loc l)) =>
      indom_property St l f
  | name_in_1_false f, owhile_st_in_1 (oute_val St (val_loc l)) =>
      True
  | name_in_1_no_loc f, owhile_st_in_1 (owhile_res_expr o) =>
      True

  | name_lambda v p, owhile_st_expr fl St =>
      indom (state_closures St) fl
  | name_app e1 e2, owhile_st_expr fl St =>
      indom (state_closures St) fl
  | name_app_1 e2, owhile_st_app_1 fl (owhile_res_expr o) =>
      env_is_defined_in_oute fl o
  | name_app_1_error e2, owhile_st_app_1 fl (owhile_res_expr o) =>
      env_is_defined_in_oute fl o
  | name_app_2 p1', owhile_st_app_2 x1 p1 fl1 (owhile_res_expr o) =>
      env_is_defined_in_oute fl1 o
  | name_app_3_return, owhile_st_app_3 (owhile_res_stat o) =>
      True
  | name_app_3_no_return, owhile_st_app_3 (owhile_res_stat o) =>
      True

  | name_op e1 op e2, owhile_st_expr fl St =>
      indom (state_closures St) fl
  | name_op_1 op e2, owhile_st_op_1 fl (owhile_res_expr o) =>
      env_is_defined_in_oute fl o
  | name_op_1_error op e2, owhile_st_op_1 fl (owhile_res_expr o) =>
      env_is_defined_in_oute fl o
  | name_op_2_error op, owhile_st_op_2 v1 (owhile_res_expr o) =>
      True
  | name_add_2, owhile_st_op_2 v1 (owhile_res_expr o) =>
      True
  | name_sub_2, owhile_st_op_2 v1 (owhile_res_expr o) =>
      True
  | name_greater_2_greater, owhile_st_op_2 v1 (oute_val St (val_basic v2)) =>
      True
  | name_greater_2_lesser_eq, owhile_st_op_2 v1 (oute_val St (val_basic v2)) =>
      True
  | name_not e, owhile_st_expr fl St =>
      indom (state_closures St) fl
  | name_not_1_true, owhile_st_not_1 (oute_val St (val_basic v)) =>
      True
  | name_not_1_false, owhile_st_not_1 (oute_val St (val_basic v)) =>
      True
  | name_not_1_error, owhile_st_not_1 (owhile_res_expr o) =>
      True
  | name_eq e1 e2, owhile_st_expr fl St =>
      indom (state_closures St) fl
  | name_eq_1 e2, owhile_st_eq_1 fl (owhile_res_expr o) =>
      env_is_defined_in_oute fl o
  | name_eq_2_basic_val_eq, owhile_st_eq_2 (val_basic v1) (oute_val St (val_basic v2)) =>
      True
  | name_eq_2_basic_val_neq, owhile_st_eq_2 (val_basic v1) (oute_val St (val_basic v2)) =>
      True
  | name_eq_2_loc_eq, owhile_st_eq_2 (val_loc l1) (oute_val St (val_loc l2)) =>
      True
  | name_eq_2_loc_neq, owhile_st_eq_2 (val_loc l1) (oute_val St (val_loc l2)) =>
      True
  | name_eq_2_mistype_basic_val_loc, owhile_st_eq_2 (val_basic v1) (oute_val St (val_loc l2)) =>
      True
  | name_eq_2_mistype_loc_basic_val, owhile_st_eq_2 (val_loc l1) (oute_val St (val_basic v2)) =>
      True
  | name_eq_2_closure, owhile_st_eq_2 v1 (oute_val St v2) =>
      True

  | name_skip, owhile_st_stat fl St =>
      indom (state_closures St) fl
  | name_seq s1 s2, owhile_st_stat fl St =>
      indom (state_closures St) fl
  | name_seq_1 s2, owhile_st_seq_1 fl (owhile_res_stat o) =>
      env_is_defined_in_outs fl o

  | name_if e s1 s2, owhile_st_stat fl St =>
      indom (state_closures St) fl
  | name_if_1_true s1 s2, owhile_st_if_1 fl (oute_val St (val_basic v)) =>
      indom (state_closures St) fl
  | name_if_1_false s1 s2, owhile_st_if_1 fl (oute_val St (val_basic v)) =>
      indom (state_closures St) fl
  | name_if_1_error s1 s2, owhile_st_if_1 fl (owhile_res_expr o) =>
      env_is_defined_in_oute fl o
  | name_while e s, owhile_st_stat fl St =>
      indom (state_closures St) fl
  | name_while_1 e s, owhile_st_while_1 fl (owhile_res_stat o) =>
      env_is_defined_in_outs fl o
  | name_while_2_true e s, owhile_st_while_2 fl (oute_val St (val_basic v)) =>
      indom (state_closures St) fl
  | name_while_2_false e s, owhile_st_while_2 fl (oute_val St (val_basic v)) =>
      indom (state_closures St) fl
  | name_while_2_error e s, owhile_st_while_2 fl (owhile_res_expr o) =>
      env_is_defined_in_oute fl o
  | name_throw, owhile_st_stat fl St =>
      indom (state_closures St) fl

  | name_property_asgn e1 f e2, owhile_st_stat fl St =>
      indom (state_closures St) fl
  | name_property_asgn_1 f e2, owhile_st_property_asgn_1 fl (owhile_res_expr o) =>
      env_is_defined_in_oute fl o
  | name_property_asgn_1_error f e2, owhile_st_property_asgn_1 fl (owhile_res_expr o) =>
      env_is_defined_in_oute fl o
  | name_property_asgn_2 f, owhile_st_property_asgn_2 l (owhile_res_expr o) =>
      is_defined_oute l o
  | name_delete e f, owhile_st_stat fl St =>
      indom (state_closures St) fl
  | name_delete_1 f, owhile_st_delete_1 (owhile_res_expr o) =>
      True
  | name_delete_1_error f, owhile_st_delete_1 (owhile_res_expr o) =>
      True

  | name_return e, owhile_st_stat fl St =>
      indom (state_closures St) fl
  | name_return_1, owhile_st_return_1 (owhile_res_expr o) =>
      True
  | name_run e, owhile_st_stat fl St =>
      indom (state_closures St) fl
  | name_run_1, owhile_st_run_1 (owhile_res_expr o) =>
      True

  | name_prog_stat s, owhile_st_prog fl St =>
      indom (state_closures St) fl

  | _, _ =>
      False
  end.

Definition owhile_cond_applies n sigma : Prop :=
  match n, sigma with
  | name_abort_ext_expr e, sigma =>
      abort_st sigma ∧ ~ abort_intercepted e sigma
  | name_abort_ext_stat s, sigma =>
      abort_st sigma ∧ ~ abort_intercepted s sigma

  | name_cst c, owhile_st_expr fl St =>
      True
  | name_random b, owhile_st_expr fl St =>
      True
  | name_var x, owhile_st_expr fl St =>
      exists E, binds (state_closures St) fl E ∧ indom E x
  | name_varUndef x, owhile_st_expr fl St =>
      exists E, binds (state_closures St) fl E ∧ ~ indom E x

  | name_new_obj, owhile_st_expr fl St =>
      True
  | name_property e f, owhile_st_expr fl St =>
      True
  | name_property_1 f, owhile_st_property_1 (oute_val St (val_loc l)) =>
      True
  | name_property_1_no_loc f, owhile_st_property_1 (owhile_res_expr o) =>
      is_not_loc o
  | name_property_1_undef f, owhile_st_property_1 (oute_val St (val_loc l)) =>
      ~ indom_property St l f
  | name_in f e, owhile_st_expr fl St =>
      True
  | name_in_1_true f, owhile_st_in_1 (oute_val St (val_loc l)) =>
      indom_property St l f
  | name_in_1_false f, owhile_st_in_1 (oute_val St (val_loc l)) =>
      ~ indom_property St l f
  | name_in_1_no_loc f, owhile_st_in_1 (owhile_res_expr o) =>
      is_not_loc o

  | name_lambda v p, owhile_st_expr fl St =>
      True
  | name_app e1 e2, owhile_st_expr fl St =>
      True
  | name_app_1 e2, owhile_st_app_1 fl (owhile_res_expr o) =>
      is_closure o
  | name_app_1_error e2, owhile_st_app_1 fl (owhile_res_expr o) =>
      is_not_closure o
  | name_app_2 p1', owhile_st_app_2 x1 p1 fl1 (owhile_res_expr o) =>
      p1 = p1' ∧ is_val o
  | name_app_3_return, owhile_st_app_3 (owhile_res_stat o) =>
      is_return o
  | name_app_3_no_return, owhile_st_app_3 (owhile_res_stat o) =>
      is_normal o

  | name_op e1 op e2, owhile_st_expr fl St =>
      True
  | name_op_1 op e2, owhile_st_op_1 fl (owhile_res_expr o) =>
      is_basic o
  | name_op_1_error op e2, owhile_st_op_1 fl (owhile_res_expr o) =>
      is_not_basic o
  | name_op_2_error op, owhile_st_op_2 v1 (owhile_res_expr o) =>
      is_not_basic o
  | name_add_2, owhile_st_op_2 v1 (owhile_res_expr o) =>
      is_basic o
  | name_sub_2, owhile_st_op_2 v1 (owhile_res_expr o) =>
      is_basic o
  | name_greater_2_greater, owhile_st_op_2 v1 (oute_val St (val_basic v2)) =>
      v1 > v2
  | name_greater_2_lesser_eq, owhile_st_op_2 v1 (oute_val St (val_basic v2)) =>
      v1 <= v2
  | name_not e, owhile_st_expr fl St =>
      True
  | name_not_1_true, owhile_st_not_1 (oute_val St (val_basic v)) =>
      v ≠ 0
  | name_not_1_false, owhile_st_not_1 (oute_val St (val_basic v)) =>
      v = 0
  | name_not_1_error, owhile_st_not_1 (owhile_res_expr o) =>
      is_not_basic o
  | name_eq e1 e2, owhile_st_expr fl St =>
      True
  | name_eq_1 e2, owhile_st_eq_1 fl (owhile_res_expr o) =>
      is_val o
  | name_eq_2_basic_val_eq, owhile_st_eq_2 (val_basic v1) (oute_val St (val_basic v2)) =>
      v1 = v2
  | name_eq_2_basic_val_neq, owhile_st_eq_2 (val_basic v1) (oute_val St (val_basic v2)) =>
      v1 ≠ v2
  | name_eq_2_loc_eq, owhile_st_eq_2 (val_loc l1) (oute_val St (val_loc l2)) =>
      l1 = l2
  | name_eq_2_loc_neq, owhile_st_eq_2 (val_loc l1) (oute_val St (val_loc l2)) =>
      l1 ≠ l2
  | name_eq_2_mistype_basic_val_loc, owhile_st_eq_2 (val_basic v1) (oute_val St (val_loc l2)) =>
      True
  | name_eq_2_mistype_loc_basic_val, owhile_st_eq_2 (val_loc l1) (oute_val St (val_basic v2)) =>
      True
  | name_eq_2_closure, owhile_st_eq_2 v1 (oute_val St v2) =>
      val_is_closure v1 ∨ val_is_closure v2

  | name_skip, owhile_st_stat fl St =>
      True
  | name_seq s1 s2, owhile_st_stat fl St =>
      True
  | name_seq_1 s2, owhile_st_seq_1 fl (owhile_res_stat o) =>
      is_normal o

  | name_if e s1 s2, owhile_st_stat fl St =>
      True
  | name_if_1_true s1 s2, owhile_st_if_1 fl (oute_val St (val_basic v)) =>
      v ≠ 0
  | name_if_1_false s1 s2, owhile_st_if_1 fl (oute_val St (val_basic v)) =>
      v = 0
  | name_if_1_error s1 s2, owhile_st_if_1 fl (owhile_res_expr o) =>
      is_not_basic o
  | name_while e s, owhile_st_stat fl St =>
      True
  | name_while_1 e s, owhile_st_while_1 fl (owhile_res_stat o) =>
      is_normal o
  | name_while_2_true e s, owhile_st_while_2 fl (oute_val St (val_basic v)) =>
      v ≠ 0
  | name_while_2_false e s, owhile_st_while_2 fl (oute_val St (val_basic v)) =>
      v = 0
  | name_while_2_error e s, owhile_st_while_2 fl (owhile_res_expr o) =>
      is_not_basic o
  | name_throw, owhile_st_stat fl St =>
      True

  | name_property_asgn e1 f e2, owhile_st_stat fl St =>
      True
  | name_property_asgn_1 f e2, owhile_st_property_asgn_1 fl (owhile_res_expr o) =>
      is_loc o
  | name_property_asgn_1_error f e2, owhile_st_property_asgn_1 fl (owhile_res_expr o) =>
      is_not_loc o
  | name_property_asgn_2 f, owhile_st_property_asgn_2 l (owhile_res_expr o) =>
      is_val o
  | name_delete e f, owhile_st_stat fl St =>
      True
  | name_delete_1 f, owhile_st_delete_1 (owhile_res_expr o) =>
      is_loc o
  | name_delete_1_error f, owhile_st_delete_1 (owhile_res_expr o) =>
      is_not_loc o

  | name_return e, owhile_st_stat fl St =>
      True
  | name_return_1, owhile_st_return_1 (owhile_res_expr o) =>
      is_val o
  | name_run e, owhile_st_stat fl St =>
      True
  | name_run_1, owhile_st_run_1 (owhile_res_expr o) =>
      is_val o

  | name_prog_stat s, owhile_st_prog fl St =>
      True

  | _, _ =>
      False
  end.

Definition owhile_cond n sigma :=
  owhile_cond_well_formed n sigma ∧ owhile_cond_applies n sigma.


(** *** Monads **)

Definition get_abort_oute o :=
  match o with
  | oute_val St v => None
  | oute_error err => Some (o : owhile_res)
  end.

Definition get_abort_outs o :=
  match o with
  | outs_normal St => None
  | outs_return St v => Some (o : owhile_res)
  | outs_error err => Some (o : owhile_res)
  end.

Definition get_abort_res o :=
  match o with
  | owhile_res_expr o =>
      get_abort_oute o
  | owhile_res_stat o =>
      get_abort_outs o
  end.

Definition get_abort_st : owhile_st -> option owhile_res.
  introv sigma. destruct sigma;
    match goal with
    | o : owhile_res |- _ =>
      exact (get_abort_res o)
    | _ => exact None
    end.
Defined.

Definition convert_abort_expr o :=
  match o with
  | owhile_res_expr o => Some (owhile_res_expr o)
  | owhile_res_stat o =>
    match o with
    | outs_normal St => None
    | outs_return St v =>
      (** This case never happens if the derivation starts from the initial state. **)
      Some (owhile_res_stat o)
    | outs_error err => Some (owhile_res_expr err)
    end
  end.

Definition convert_abort_stat o :=
  option_map owhile_res_stat
    match o with
    | owhile_res_expr o =>
      match o with
      | oute_val St v => None
      | oute_error err => Some (err : outs)
      end
    | owhile_res_stat o => Some o
    end.


Definition if_st_expr A (f : _ -> _ -> option A) sigma :=
  match sigma with
  | owhile_st_expr fl St => f fl St
  | _ => None
  end.

Definition if_st_stat A (f : _ -> _ -> option A) sigma :=
  match sigma with
  | owhile_st_stat fl St => f fl St
  | _ => None
  end.

Definition if_st_prog A (f : _ -> _ -> option A) sigma :=
  match sigma with
  | owhile_st_prog fl St => f fl St
  | _ => None
  end.

Definition if_res_expr A (f : _ -> option A) o :=
  match o with
  | owhile_res_expr o => f o
  | _ => None
  end.

Definition if_res_stat A (f : _ -> option A) o :=
  match o with
  | owhile_res_stat o => f o
  | _ => None
  end.

Definition if_st_in_1 A (f : _ -> _ -> option A) o :=
  match o with
  | owhile_st_in_1 (oute_val St (val_loc l)) => f St l
  | _ => None
  end.

Definition if_st_property_1 A (f : _ -> _ -> option A) o :=
  match o with
  | owhile_st_property_1 (oute_val St (val_loc l)) => f St l
  | _ => None
  end.

Definition if_st_op_1 A (f : _ -> _ -> _ -> option A) o :=
  match o with
  | owhile_st_op_1 fl (oute_val St (val_basic v)) => f fl St v
  | _ => None
  end.

Definition if_st_op_2 A (f : _ -> _ -> _ -> option A) o :=
  match o with
  | owhile_st_op_2 v1 (oute_val St (val_basic v2)) => f St v1 v2
  | _ => None
  end.

Definition if_st_not_1 A (f : _ -> _ -> option A) o :=
  match o with
  | owhile_st_not_1 (oute_val St (val_basic v)) => f St v
  | _ => None
  end.

Definition if_st_eq_1 A (f : _ -> _ -> _ -> option A) o :=
  match o with
  | owhile_st_eq_1 fl (oute_val St v) => f fl St v
  | _ => None
  end.

Definition if_st_eq_2 A (f : _ -> _ -> _ -> option A) o :=
  match o with
  | owhile_st_eq_2 v1 (oute_val St v2) => f St v1 v2
  | _ => None
  end.

Definition if_st_app_1 A (f : _ -> _ -> _ -> option A) o :=
  match o with
  | owhile_st_app_1 fl (oute_val St v1) => f fl St v1
  | _ => None
  end.

Definition if_st_app_2 A (f : _ -> _ -> _ -> _ -> _ -> option A) o :=
  match o with
  | owhile_st_app_2 x1 p1 fl1 (oute_val St v2) => f St x1 p1 fl1 v2
  | _ => None
  end.

Definition if_st_app_3 A (f : _ -> _ -> option A) o :=
  match o with
  | owhile_st_app_3 (outs_return St v) => f St v
  | _ => None
  end.

Definition if_st_seq_1 A (f : _ -> _ -> option A) o :=
  match o with
  | owhile_st_seq_1 fl (outs_normal St) => f fl St
  | _ => None
  end.

Definition if_st_if_1 A (f : _ -> _ -> _ -> option A) o :=
  match o with
  | owhile_st_if_1 fl (oute_val St v) => f fl St v
  | _ => None
  end.

Definition if_st_while_1 A (f : _ -> _ -> option A) o :=
  match o with
  | owhile_st_while_1 fl (outs_normal St) => f fl St
  | _ => None
  end.

Definition if_st_while_2 A (f : _ -> _ -> _ -> option A) o :=
  match o with
  | owhile_st_while_2 fl (oute_val St v) => f fl St v
  | _ => None
  end.

Definition if_st_property_asgn_1 A (f : _ -> _ -> _ -> option A) o :=
  match o with
  | owhile_st_property_asgn_1 fl (oute_val St (val_loc l)) => f fl St l
  | _ => None
  end.

Definition if_st_property_asgn_2 A (f : _ -> _ -> _ -> option A) o :=
  match o with
  | owhile_st_property_asgn_2 l (oute_val St v) => f St l v
  | _ => None
  end.

Definition if_st_delete_1 A (f : _ -> _ -> option A) o :=
  match o with
  | owhile_st_delete_1 (oute_val St (val_loc l)) => f St l
  | _ => None
  end.

Definition if_st_return_1 A (f : _ -> _ -> option A) o :=
  match o with
  | owhile_st_return_1 (oute_val St v) => f St v
  | _ => None
  end.

Definition if_st_run_1 A (f : _ -> option A) o :=
  match o with
  | owhile_st_run_1 (oute_val St v) => f St
  | _ => None
  end.


(** *** Transfer Functions **)

Definition owhile_rule n : Rule owhile_st owhile_res :=
  match n with
  | name_abort_ext_expr e =>
    let ax sigma :=
      LibOption.apply convert_abort_expr (get_abort_st sigma) in
    Rule_Ax ax
  | name_abort_ext_stat s =>
    let ax sigma :=
      LibOption.apply convert_abort_stat (get_abort_st sigma) in
    Rule_Ax ax

  | name_cst c =>
    let ax :=
      if_st_expr (fun fl St =>
        Some (oute_val St (val_basic c) : owhile_res)) in
    Rule_Ax ax
  | name_random b =>
    let n :=
      if b then 1 else 0 in
    let ax :=
      if_st_expr (fun fl St =>
        Some (oute_val St (val_basic n) : owhile_res)) in
    Rule_Ax ax
  | name_var x =>
    let ax :=
      if_st_expr (fun fl St =>
        match read_option (state_closures St) fl with
        | Some E => option_map (oute_val St : _ -> owhile_res) (read_option E x)
        | None => None
        end) in
    Rule_Ax ax
  | name_varUndef x =>
    let ax _ :=
      Some (owhile_res_expr (Err_undefined_variable x)) in
    Rule_Ax ax

  | name_new_obj =>
    let ax :=
      if_st_expr (fun fl St =>
        let '(St, l) := alloc_state St in
        Some (oute_val St (val_loc l) : owhile_res)) in
    Rule_Ax ax
  | name_property e f =>
    let up :=
      if_st_expr (fun fl St =>
        Some (owhile_st_expr fl St)) in
    let next sigma o :=
      Some (owhile_st_property_1 o) in
    Rule_R2 up next
  | name_property_1 f =>
    let ax :=
      if_st_property_1 (fun St l =>
        match read_option (state_heap St) l with
        | Some o =>
          match read_option o f with
          | Some v =>
            Some (oute_val St v : owhile_res)
          | None => None
          end
        | None => None
        end) in
    Rule_Ax ax
  | name_property_1_no_loc f =>
    let ax _ :=
      Some (owhile_res_expr (Err_val_basic_no_property f)) in
    Rule_Ax ax
  | name_property_1_undef f =>
    let ax _ :=
      Some (owhile_res_expr (Err_undefined_property f)) in
    Rule_Ax ax
  | name_in f e =>
    let up :=
      if_st_expr (fun fl St =>
        Some (owhile_st_expr fl St)) in
    let next sigma o :=
      Some (owhile_st_in_1 o) in
    Rule_R2 up next
  | name_in_1_true f =>
    let ax :=
      if_st_in_1 (fun St l =>
        Some (oute_val St (val_basic 1) : owhile_res)) in
    Rule_Ax ax
  | name_in_1_false f =>
    let ax :=
      if_st_in_1 (fun St l =>
        Some (oute_val St (val_basic 0) : owhile_res)) in
    Rule_Ax ax
  | name_in_1_no_loc f =>
    let ax _ :=
      Some (owhile_res_expr (Err_val_basic_no_property f)) in
    Rule_Ax ax

  | name_op e1 op e2 =>
    let up :=
      if_st_expr (fun fl St =>
        Some (owhile_st_expr fl St)) in
    let next sigma o :=
      if_st_expr (fun fl St =>
        Some (owhile_st_op_1 fl o)) sigma in
    Rule_R2 up next
  | name_op_1 op e2 =>
    let up :=
      if_st_op_1 (fun fl St v =>
        Some (owhile_st_expr fl St)) in
    let next sigma o :=
      if_st_op_1 (fun fl St v =>
        Some (owhile_st_op_2 v o)) sigma in
    Rule_R2 up next
  | name_op_1_error op e2 =>
    let ax _ :=
      Some (owhile_res_expr Err_not_a_basic_value) in
    Rule_Ax ax
  | name_op_2_error op =>
    let ax _ :=
      Some (owhile_res_expr Err_not_a_basic_value) in
    Rule_Ax ax
  | name_add_2 =>
    let ax :=
      if_st_op_2 (fun St v1 v2 =>
        Some (oute_val St (val_basic (v1 + v2)) : owhile_res)) in
    Rule_Ax ax
  | name_sub_2 =>
    let ax :=
      if_st_op_2 (fun St v1 v2 =>
        Some (oute_val St (val_basic (v1 - v2)) : owhile_res)) in
    Rule_Ax ax
  | name_greater_2_greater =>
    let ax :=
      if_st_op_2 (fun St v1 v2 =>
        Some (oute_val St (val_basic 1) : owhile_res)) in
    Rule_Ax ax
  | name_greater_2_lesser_eq =>
    let ax :=
      if_st_op_2 (fun St v1 v2 =>
        Some (oute_val St (val_basic 0) : owhile_res)) in
    Rule_Ax ax
  | name_not e =>
    let up :=
      if_st_expr (fun fl St =>
        Some (owhile_st_expr fl St)) in
    let next sigma o :=
      Some (owhile_st_not_1 o) in
    Rule_R2 up next
  | name_not_1_true =>
    let ax :=
      if_st_not_1 (fun St v =>
        Some (oute_val St (val_basic 0) : owhile_res)) in
    Rule_Ax ax
  | name_not_1_false =>
    let ax :=
      if_st_not_1 (fun St v =>
        Some (oute_val St (val_basic 1) : owhile_res)) in
    Rule_Ax ax
  | name_not_1_error =>
    let ax _ :=
      Some (owhile_res_expr Err_not_a_basic_value) in
    Rule_Ax ax
  | name_eq e1 e2 =>
    let up :=
      if_st_expr (fun fl St =>
        Some (owhile_st_expr fl St)) in
    let next sigma o :=
      if_st_expr (fun fl St =>
        Some (owhile_st_eq_1 fl o)) sigma in
    Rule_R2 up next
  | name_eq_1 e2 =>
    let up :=
      if_st_eq_1 (fun fl St v =>
        Some (owhile_st_expr fl St)) in
    let next sigma o :=
      if_st_eq_1 (fun fl St v =>
        Some (owhile_st_eq_2 v o)) sigma in
    Rule_R2 up next
  | name_eq_2_basic_val_eq =>
    let ax :=
      if_st_eq_2 (fun St v1 v2 =>
        Some (oute_val St (val_basic 1) : owhile_res)) in
    Rule_Ax ax
  | name_eq_2_basic_val_neq =>
    let ax :=
      if_st_eq_2 (fun St v1 v2 =>
        Some (oute_val St (val_basic 0) : owhile_res)) in
    Rule_Ax ax
  | name_eq_2_loc_eq =>
    let ax :=
      if_st_eq_2 (fun St v1 v2 =>
        Some (oute_val St (val_basic 1) : owhile_res)) in
    Rule_Ax ax
  | name_eq_2_loc_neq =>
    let ax :=
      if_st_eq_2 (fun St v1 v2 =>
        Some (oute_val St (val_basic 0) : owhile_res)) in
    Rule_Ax ax
  | name_eq_2_mistype_basic_val_loc =>
    let ax _ :=
      Some (owhile_res_expr Err_comparing_values_of_different_types) in
    Rule_Ax ax
  | name_eq_2_mistype_loc_basic_val =>
    let ax _ :=
      Some (owhile_res_expr Err_comparing_values_of_different_types) in
    Rule_Ax ax
  | name_eq_2_closure =>
    let ax _ :=
      Some (owhile_res_expr Err_comparing_functions) in
    Rule_Ax ax

  | name_lambda x p =>
    let ax :=
      if_st_expr (fun fl St =>
        Some (oute_val St (val_closure x p fl) : owhile_res)) in
    Rule_Ax ax
  | name_app e1 e2 =>
    let up :=
      if_st_expr (fun fl St =>
        Some (owhile_st_expr fl St)) in
    let next sigma o :=
      if_st_expr (fun fl St =>
        Some (owhile_st_app_1 fl o)) sigma in
    Rule_R2 up next
  | name_app_1 e2 =>
    let up :=
      if_st_app_1 (fun fl St v1 =>
        Some (owhile_st_expr fl St)) in
    let next sigma o :=
      if_st_app_1 (fun fl0 St v1 =>
        match v1 with
        | val_closure x p fl =>
          Some (owhile_st_app_2 x p fl o)
        | _ => None
        end) sigma in
    Rule_R2 up next
  | name_app_1_error e2 =>
    let ax _ :=
      Some (owhile_res_expr Err_not_a_function) in
    Rule_Ax ax
  | name_app_2 p =>
    let up :=
      if_st_app_2 (fun St x p fl v2 =>
        match read_option (state_closures St) fl with
        | Some E =>
          let E' := write E x v2 in
          let fl' := state_fresh_floc St in
          let St' := {|
              state_heap := state_heap St ;
              state_closures := write (state_closures St) fl' E' ;
              state_fresh_loc := state_fresh_loc St ;
              state_fresh_floc := S fl'
            |} in
          Some (owhile_st_prog fl' St')
        | None => None
        end) in
    let next sigma o :=
      Some (owhile_st_app_3 o) in
    Rule_R2 up next
  | name_app_3_return =>
    let ax :=
      if_st_app_3 (fun St v =>
        Some (oute_val St v : owhile_res)) in
    Rule_Ax ax
  | name_app_3_no_return =>
    let ax _ :=
      Some (owhile_res_expr Err_no_return_at_the_end_of_a_function) in
    Rule_Ax ax

  | name_skip =>
    let ax :=
      if_st_stat (fun fl St =>
        Some (outs_normal St : owhile_res)) in
    Rule_Ax ax
  | name_seq s1 s2 =>
    let up :=
      if_st_stat (fun fl St =>
        Some (owhile_st_stat fl St)) in
    let next sigma o :=
      if_st_stat (fun fl St =>
        Some (owhile_st_seq_1 fl o)) sigma in
    Rule_R2 up next
  | name_seq_1 s2 =>
    let up :=
      if_st_seq_1 (fun fl St =>
        Some (owhile_st_stat fl St)) in
    Rule_R1 _ up

  | name_if e s1 s2 =>
    let up :=
      if_st_stat (fun fl St =>
        Some (owhile_st_expr fl St)) in
    let next sigma o :=
      if_st_stat (fun fl St =>
        Some (owhile_st_if_1 fl o)) sigma in
    Rule_R2 up next
  | name_if_1_true s1 s2 =>
    let up :=
      if_st_if_1 (fun fl St v =>
        Some (owhile_st_stat fl St)) in
    Rule_R1 _ up
  | name_if_1_false s1 s2 =>
    let up :=
      if_st_if_1 (fun fl St v =>
        Some (owhile_st_stat fl St)) in
    Rule_R1 _ up
  | name_if_1_error s1 s2 =>
    let ax _ :=
      Some (owhile_res_stat Err_not_a_basic_value) in
    Rule_Ax ax
  | name_while e s =>
    let up :=
      if_st_stat (fun fl St =>
        Some (owhile_st_while_1 fl (outs_normal St))) in
    Rule_R1 _ up
  | name_while_1 e s =>
    let up :=
      if_st_while_1 (fun fl St =>
        Some (owhile_st_expr fl St)) in
    let next sigma o :=
      if_st_while_1 (fun fl St =>
        Some (owhile_st_while_2 fl o)) sigma in
    Rule_R2 up next
  | name_while_2_true s1 s2 =>
    let up :=
      if_st_while_2 (fun fl St v =>
        Some (owhile_st_stat fl St)) in
    let next sigma o :=
      if_st_while_2 (fun fl St v =>
        Some (owhile_st_while_1 fl o)) sigma in
    Rule_R2 up next
  | name_while_2_false s1 s2 =>
    let ax :=
      if_st_while_2 (fun fl St v =>
        Some (outs_normal St : owhile_res)) in
    Rule_Ax ax
  | name_while_2_error s1 s2 =>
    let ax _ :=
      Some (owhile_res_stat Err_not_a_basic_value) in
    Rule_Ax ax
  | name_throw =>
    let ax _ :=
      Some (owhile_res_stat Err_throw) in
    Rule_Ax ax

  | name_property_asgn e1 f e2 =>
    let up :=
      if_st_stat (fun fl St =>
        Some (owhile_st_expr fl St)) in
    let next sigma o :=
      if_st_stat (fun fl St =>
        Some (owhile_st_property_asgn_1 fl o)) sigma in
    Rule_R2 up next
  | name_property_asgn_1 f e2 =>
    let up :=
      if_st_property_asgn_1 (fun fl St l =>
        Some (owhile_st_expr fl St)) in
    let next sigma o :=
      if_st_property_asgn_1 (fun fl St l =>
        Some (owhile_st_property_asgn_2 l o)) sigma in
    Rule_R2 up next
  | name_property_asgn_1_error f e2 =>
    let ax _ :=
      Some (owhile_res_stat (Err_write_property_of_something_not_object f)) in
    Rule_Ax ax
  | name_property_asgn_2 f =>
    let ax :=
      if_st_property_asgn_2 (fun St l v =>
        match read_option (state_heap St) l with
        | Some obj =>
          let obj := write obj f v in
          let St := {|
              state_heap := write (state_heap St) l obj ;
              state_closures := state_closures St ;
              state_fresh_loc := state_fresh_loc St ;
              state_fresh_floc := state_fresh_floc St
            |} in
          Some (outs_normal St : owhile_res)
        | None => None
        end) in
    Rule_Ax ax
  | name_delete e f =>
    let up :=
      if_st_stat (fun fl St =>
        Some (owhile_st_expr fl St)) in
    let next sigma o :=
      Some (owhile_st_delete_1 o) in
    Rule_R2 up next
  | name_delete_1 f =>
    let ax :=
      if_st_delete_1 (fun St l =>
        match read_option (state_heap St) l with
        | Some obj =>
          let obj := rem obj f in
          let St := {|
              state_heap := write (state_heap St) l obj ;
              state_closures := state_closures St ;
              state_fresh_loc := state_fresh_loc St ;
              state_fresh_floc := state_fresh_floc St
            |} in
          Some (outs_normal St : owhile_res)
        | None => None
        end) in
    Rule_Ax ax
  | name_delete_1_error f =>
    let ax _ :=
      Some (owhile_res_stat (Err_delete_property_of_something_not_object f)) in
    Rule_Ax ax

  | name_return e =>
    let up :=
      if_st_stat (fun fl St =>
        Some (owhile_st_expr fl St)) in
    let next sigma o :=
      Some (owhile_st_return_1 o) in
    Rule_R2 up next
  | name_return_1 =>
    let ax :=
      if_st_return_1 (fun St v =>
        Some (owhile_res_stat (outs_return St v))) in
    Rule_Ax ax
  | name_run e =>
    let up :=
      if_st_stat (fun fl St =>
        Some (owhile_st_expr fl St)) in
    let next sigma o :=
      Some (owhile_st_run_1 o) in
    Rule_R2 up next
  | name_run_1 =>
    let ax :=
      if_st_run_1 (fun St =>
        Some (outs_normal St : owhile_res)) in
    Rule_Ax ax

  | name_prog_stat s =>
    let up :=
      if_st_prog (fun fl St =>
        Some (owhile_st_stat fl St)) in
    Rule_R1 _ up
  end.


(** *** Wrapping Up **)

Definition owhile_dom :=
  make_domains owhile_st owhile_res.

Definition owhile_sem :=
  make_semantics owhile_str owhile_dom owhile_cond owhile_rule.

Lemma owhile_rule_format_correct : rule_format_correct_all owhile_sem.
Proof. intro n. destruct n; constructors. Qed.

Lemma owhile_semantics_full : semantics_full owhile_sem.
Proof.
  introv C. simpls. lets (CW&CA): (rm C); destruct n eqn:EQn; destruct sigma; simpls;
    tryfalse; try solve [ repeat inverts CA as CA; constructors; simpls* ].
  (** var **)
  lets (E&BE&I): (rm CA). lets (va&B): @indom_binds (rm I). constructors~. simpl.
  rewrite (binds_read_option BE). rewrite* (binds_read_option B).
  (** property 1 **)
  destruct o as [[St [|l|]|]|]; tryfalse. lets (obj&Bobj&I): (rm CW).
  lets (v&B): @indom_binds (rm I). constructors~. simpl.
  rewrite (binds_read_option Bobj). rewrite* (binds_read_option B).
  (** in 1 true **)
  destruct o as [[St [|l|]|]|]; tryfalse. constructors*.
  (** in 1 false **)
  destruct o as [[St [|l|]|]|]; tryfalse. constructors*.
  (** op 1 **)
  destruct o0 as [[St [v| |]|]|]; tryfalse*. constructors*.
  intro o0. eexists. reflexivity.
  (** add 2 **)
  destruct o as [[St [v| |]|]|]; tryfalse. constructors*.
  (** sub 2 **)
  destruct o as [[St [v| |]|]|]; tryfalse. constructors*.
  (** greater 2 greater **)
  destruct o as [[St [v| |]|]|]; tryfalse. constructors*.
  (** greater 2 lesser eq **)
  destruct o as [[St [v| |]|]|]; tryfalse. constructors*.
  (** not 1 true **)
  destruct o as [[St [v| |]|]|]; tryfalse. constructors*.
  (** not 1 false **)
  destruct o as [[St [v| |]|]|]; tryfalse. constructors*.
  (** eq 1 **)
  destruct o as [[St v|]|]; tryfalse*. constructors*. intro o. eexists. reflexivity.
  (** eq 2 basic val eq **)
  destruct v as [v1| |]; tryfalse. destruct o as [[St [v2| |]|]|]; tryfalse. constructors*.
  (** eq 2 basic val neq **)
  destruct v as [v1| |]; tryfalse. destruct o as [[St [v2| |]|]|]; tryfalse. constructors*.
  (** eq 2 loc eq **)
  destruct v as [|l1|]; tryfalse. destruct o as [[St [|l2|]|]|]; tryfalse. constructors*.
  (** eq 2 loc neq **)
  destruct v as [|l1|]; tryfalse. destruct o as [[St [|l2|]|]|]; tryfalse. constructors*.
  (** app 1 **)
  destruct o as [[St [| |l1]|]|]; tryfalse*. constructors*. lets (E&B): @indom_binds (rm CA).
  intro o. eexists. reflexivity.
  (** app 2 **)
  destruct o as [[St v2|]|]; tryfalse*. simpls.
  destruct (read_option (state_closures St) f) eqn: Eq.
   constructors.
    simpl. rewrite* Eq.
    intro o. eexists. autos*.
   apply read_option_not_indom in Eq. false* CW.
  (** app 3 **)
  destruct o as [|[|St v|]]; tryfalse. constructors*.
  (** seq 1 **)
  destruct o as [|[St| |]]; tryfalse*. constructors*.
  (** if 1 true **)
  destruct o as [[St [v| |]|]|]; tryfalse. constructors*.
  (** if 1 false **)
  destruct o as [[St [v| |]|]|]; tryfalse. constructors*.
  (** while 1 **)
  destruct o as [|[St| |]]; tryfalse*. constructors*. simpls*.
  (** while 1 true **)
  destruct o as [[St [v| |]|]|]; tryfalse. constructors*. simpls*.
  (** while 1 false **)
  destruct o as [[St [v| |]|]|]; tryfalse. constructors*.
  (** property asgn 1 **)
  destruct o as [[St [|l|]|]|]; tryfalse*. constructors*.
  intro o. eexists. reflexivity.
  (** property asgn 2 **)
  destruct o as [[St v|]|]; tryfalse. lets (va&B): @indom_binds (rm CW).
  constructors~. simpl. rewrite* (binds_read_option B).
  (** delete 1 **)
  destruct o as [[St [|l|]|]|]; tryfalse. lets (va&B): @indom_binds (rm CA).
  constructors~. simpl. rewrite* (binds_read_option B).
  (** return 1 **)
  destruct o as [[St v|]|]; tryfalse. constructors*.
  (** run 1 **)
  destruct o as [[St v|]|]; tryfalse. constructors*.
Qed.

End OWhile.

