
(** This file contains the definitions of the separation-logic
  formulae used in the abstract semantics of the example language. **)

Set Implicit Arguments.

Require Export DefinitionsSep.

Section General.

(** [L] represents singleton locations, [K] set locations and [KE]
  environments, which behave like set locations. **)
Variable L K KE : Type. (** Typically, set of integers **)
Hypothesis L_comparable : Comparable L.
Hypothesis K_comparable : Comparable K.
Hypothesis KE_comparable : Comparable KE.

(** These sets have to be infinite. **)
Hypothesis L_infinite : forall ls,
  exists (l : L), ~ mem l ls.
Hypothesis K_infinite : forall ks,
  exists (k : K), ~ mem k ks.
Hypothesis KE_infinite : forall ks,
  exists (k : KE), ~ mem k ks.
Hypothesis L_infinite_pick : forall ls,
  Pickable (fun l : L => ~ mem l ls).
Hypothesis K_infinite_pick : forall ks,
  Pickable (fun k : K => ~ mem k ks).
Hypothesis KE_infinite_pick : forall ks,
  Pickable (fun k : KE => ~ mem k ks).

Variable ABasic : Type. (** Abstract basic values **)
Hypothesis ABasic_comparable : Comparable ABasic.
Hypothesis ABasicLat : LatticeDec.t ABasic.
Hypothesis ABasicTop : TopDec.t ABasic.
Variable ABasic_gamma : ABasic → ℘ basic_val.
Hypothesis ABasic_Gamma : Gamma.t _ _ ABasic_gamma.
Hypothesis ABasic_gamma_bot : forall b,
  ~ ABasic_gamma (⊥♯) b.
Hypothesis ABasic_gamma_top : forall b,
  ABasic_gamma (⊤♯) b.

Let ALoc := ALoc L K.
Let d_singleton_loc := singleton_loc _ : L → ALoc.
Let d_set_loc := set_loc _ (K := K) : K → ALoc.
Coercion d_singleton_loc : L >-> ALoc.
Coercion d_set_loc : K >-> ALoc.

Local Instance Local_Inhab_Comparable : Inhab ALoc.
  apply~ ALoc_inhab.
Defined.

Local Instance Local_ALoc_Comparable : Comparable ALoc.
  apply~ ALoc_comparable.
Defined.

Let DObject := DObject L K KE ABasic.
Let AEnv := AEnv L K KE ABasic.


(** * Abstract State Formulae **)

(** ** Inner Formulae **)

(** *** Definition **)

Inductive IFormula : Type :=
  | IFormula_emp : IFormula
  | IFormula_star : IFormula → IFormula → IFormula

  | IFormula_loc : ALoc → DObject → IFormula

  | IFormula_env : KE → AEnv → IFormula
  .

Notation "'emp'" := IFormula_emp : sep_scope.
Notation "phi1 '*' phi2" := (IFormula_star phi1 phi2) : sep_scope.
Notation "h '-->' o" := (IFormula_loc h o)
  (at level 39, no associativity) : sep_scope.
Notation "ke '~~>' E" := (IFormula_env ke E)
  (at level 39, no associativity) : sep_scope.

Open Scope sep_scope.

Instance IFormula_Inhab : Inhab IFormula.
  applys prove_Inhab emp.
Defined.

(** Renaming the identifiers of an inner formula. **)
Fixpoint IFormula_substs phi hs h :=
  match phi with
  | emp => emp
  | phi1 * phi2 => IFormula_substs phi1 hs h * IFormula_substs phi2 hs h
  | h' --> o =>
    let o' := DObject_substs _ _ o hs h in
    ifb h = h' then
      fold_left (fun h phi => h --> o' * phi) emp hs
    else h' --> o'
  | ke ~~> E => ke ~~> AEnv_substs _ _ E hs h
  end.

Fixpoint IFormula_Esubsts phi kes ke :=
  match phi with
  | emp => emp
  | phi1 * phi2 => IFormula_Esubsts phi1 kes ke * IFormula_Esubsts phi2 kes ke
  | h --> o => h --> DObject_Esubsts _ o kes ke
  | ke' ~~> E =>
    let E' := AEnv_Esubsts _ E kes ke in
    ifb ke = ke' then
      fold_left (fun ke phi => ke ~~> E' * phi) emp kes
    else ke' ~~> E'
  end.


(** Note that [IFormula_substs] and [IFormula_Esubsts] only perform
  one simple rewriting, and is unable to deal with the simultaneous
  rewritings of membranes.  See [passThroughMembrane] for a more
  powerful version. **)


(** Names present in an inner formula. **)
Fixpoint IFNames phi :=
  match phi with
  | emp => []
  | phi1 * phi2 => IFNames phi1 ++ IFNames phi2
  | h --> o => h :: DNames o
  | ke ~~> E => ENames E
  end.

Fixpoint IFENames phi :=
  match phi with
  | emp => []
  | phi1 * phi2 => IFENames phi1 ++ IFENames phi2
  | h --> o => DENames o
  | ke ~~> E => ke :: EENames E
  end.


(** It can be frustrating not to have the equality here, but the
  equality is not correct: if we replace an abstract location by an
  empty list, we remove every objects referred by this location, thus
  removing every location present in the object.  In other words,
  consider [(l ---> nil)(l --> {f:l'}) = emp], in which we have lost
  the location [l']. **)
Lemma IFNames_substs : forall phi h hs h0,
  mem h0 (IFNames (IFormula_substs phi hs h)) →
  mem h (IFNames phi) &&& mem h0 hs
  ||| mem h0 (IFNames phi) &&& decide (h ≠ h0).
Proof.
  induction phi; simpl; introv E; unfolds ALoc.
   autos*.
   repeat rewrite mem_app in *. rew_refl in E.
    repeat rewrite distribute_and. invert E as E.
     apply IHphi1 in E. rew_refl in E. rew_refl*.
     apply IHphi2 in E. rew_refl in E. rew_refl*.
   cases_if as I; fold_bool; rew_refl in I.
    substs. rewrite fold_left_eq_fold_right in E. sets_eq hs': (rev hs).
     asserts R: (mem h0 hs' → mem h0 hs).
      repeat rewrite <- Mem_mem. substs. apply~ Mem_rev_iff.
     clear EQhs'. gen hs. induction hs'; introv E R; tryfalse.
     rewrite fold_right_cons in E. unfold IFNames in E. fold IFNames in E. unfolds ALoc.
      rewrite mem_app in E. rew_refl in E. invert E as E.
       simpl in E. rew_refl in E. rew_refl. invert E as E.
        left. splits~. apply R. simpl. rew_refl*.
        unfolds Local_ALoc_Comparable. rewrite~ DNames_substs in E. rew_refl in E. inverts* E.
       forwards~: (rm IHhs') (rm E). introv I. apply R. simpl. rew_refl*.
    simpl in E. unfolds Local_ALoc_Comparable. rewrite~ DNames_substs in E. rew_refl in E. rew_refl. inverts* E.
   unfolds Local_ALoc_Comparable. rewrite~ ENames_substs in E.
Qed.

Lemma IFENames_Esubsts : forall phi ke kes ke0,
  mem ke0 (IFENames (IFormula_Esubsts phi kes ke)) →
  mem ke (IFENames phi) &&& mem ke0 kes
  ||| mem ke0 (IFENames phi) &&& decide (ke ≠ ke0).
Proof.
  induction phi; simpl; introv E.
   autos*.
   repeat rewrite mem_app in *. rew_refl in E.
    repeat rewrite distribute_and. invert E as E.
     apply IHphi1 in E. rew_refl in E. rew_refl*.
     apply IHphi2 in E. rew_refl in E. rew_refl*.
   rewrite~ DENames_Esubsts in E.
   cases_if as I; fold_bool; rew_refl in I.
    substs. rewrite fold_left_eq_fold_right in E. sets_eq kes': (rev kes).
     asserts R: (mem ke0 kes' → mem ke0 kes).
      repeat rewrite <- Mem_mem. substs. apply~ Mem_rev_iff.
     clear EQkes'. gen kes. induction kes'; introv E R; tryfalse.
     rewrite fold_right_cons in E. unfold IFENames in E. fold IFENames in E.
      rewrite mem_app in E. rew_refl in E. invert E as E.
       simpl in E. rew_refl in E. rew_refl. invert E as E.
        left. splits~. apply R. simpl. rew_refl*.
        rewrite~ EENames_Esubsts in E. rew_refl in E. inverts* E.
       forwards~: (rm IHkes') (rm E). introv I. apply R. simpl. rew_refl*.
    simpl in E. rewrite~ EENames_Esubsts in E. rew_refl in E. rew_refl. inverts E; autos*.
Qed.

Lemma IFNames_Esubsts : forall phi ke kes h,
  mem h (IFNames (IFormula_Esubsts phi kes ke)) →
  mem h (IFNames phi).
Proof.
  induction phi; simpl; introv E.
   autos*.
   repeat rewrite mem_app in *. rew_refl in E.
    repeat rewrite distribute_and. invert E as E.
     apply IHphi1 in E. rew_refl in E. rew_refl*.
     apply IHphi2 in E. rew_refl in E. rew_refl*.
   unfolds ALoc. unfolds Local_ALoc_Comparable. rewrite~ DNames_Esubsts in E.
   cases_if as I; fold_bool; rew_refl in I.
    substs. rewrite fold_left_eq_fold_right in E. induction (rev kes); tryfalse.
     rewrite fold_right_cons in E. unfold IFNames in E. fold IFNames in E.
     rewrite mem_app in E. rew_refl in E. invert E as E.
      unfolds ALoc. unfolds Local_ALoc_Comparable. rewrite~ ENames_Esubsts in E.
      forwards~: (rm IHl) (rm E).
    simpl in E. unfolds ALoc. unfolds Local_ALoc_Comparable. rewrite~ ENames_Esubsts in E.
Qed.

Lemma IFENames_substs : forall phi h hs ke,
  mem ke (IFENames (IFormula_substs phi hs h)) →
  mem ke (IFENames phi).
Proof.
  induction phi; simpl; introv E.
   autos*.
   repeat rewrite mem_app in *. rew_refl in E.
    repeat rewrite distribute_and. invert E as E.
     apply IHphi1 in E. rew_refl in E. rew_refl*.
     apply IHphi2 in E. rew_refl in E. rew_refl*.
   cases_if as I; fold_bool; rew_refl in I.
    substs. rewrite fold_left_eq_fold_right in E. unfolds ALoc. induction (rev hs); tryfalse.
     rewrite fold_right_cons in E. unfold IFENames in E. fold IFENames in E.
      rewrite mem_app in E. rew_refl in E. invert E as E.
       simpl in E. rew_refl in E. rew_refl.
        rewrite DENames_substs in E. rew_refl in E. inverts* E.
        typeclass.
       forwards~: (rm IHl) (rm E).
    simpl in E. rewrite~ DENames_substs in E.
   rewrite~ EENames_substs in E.
Qed.


(** *** Subformula **)

Inductive IFormula_sub_step : IFormula → IFormula → Prop :=
  | IFormula_sub_step_star_left : forall phi1 phi2,
    IFormula_sub_step (phi1 * phi2) phi1
  | IFormula_sub_step_star_right : forall phi1 phi2,
    IFormula_sub_step (phi1 * phi2) phi2
  .

(** [IFormula_sub phi1 phi2] iff [phi2] is a subformula of [phi1].  **)
Definition IFormula_sub :=
  rtclosure IFormula_sub_step.

Lemma IFormula_sub_refl : forall phi,
 IFormula_sub phi phi.
Proof. introv. apply~ rtclosure_refl. Qed.

Lemma IFormula_sub_star : forall phi0 phi1 phi2,
 IFormula_sub (phi1 * phi2) phi0
   ↔ phi0 = phi1 * phi2 ∨ IFormula_sub phi1 phi0 ∨ IFormula_sub phi2 phi0.
Proof.
  introv. iff I.
   sets_eq phi12: (phi1 * phi2). induction I; substs.
    left~.
    inverts~ H.
   inverts I as I.
    apply~ IFormula_sub_refl.
    inverts I as I; applys rtclosure_step I; constructors~.
Qed.

Lemma IFormula_sub_IFNames_incl : forall phi1 phi2,
  IFormula_sub phi1 phi2 →
  incl (IFNames phi2) (IFNames phi1).
Proof.
  introv S. induction S.
   apply* incl_refl.
   applys incl_tran IHS. inverts H.
    apply* incl_app_l. typeclass.
    apply* incl_app_r. typeclass.
Qed.

Lemma IFormula_sub_IFNames : forall phi1 phi2 h,
  IFormula_sub phi1 phi2 →
  mem h (IFNames phi2) →
  mem h (IFNames phi1).
Proof. introv S. repeat rewrite <- In_mem. applys~ IFormula_sub_IFNames_incl S. Qed.

Lemma IFormula_sub_IFENames_incl : forall phi1 phi2,
  IFormula_sub phi1 phi2 →
  incl (IFENames phi2) (IFENames phi1).
Proof.
  introv S. induction S.
   apply* incl_refl.
   applys incl_tran IHS. inverts H.
    apply* incl_app_l.
    apply* incl_app_r.
Qed.

Lemma IFormula_sub_IFENames : forall phi1 phi2 ke,
  IFormula_sub phi1 phi2 →
  mem ke (IFENames phi2) →
  mem ke (IFENames phi1).
Proof. introv S. repeat rewrite <- In_mem. applys~ IFormula_sub_IFENames_incl S. Qed.


(** *** Equivalence relation **)

(** Two inner formulae are related to each other with this predicate
  if one is the rewriting of the other through the commutativity and
  neutral of the star. We do not consider the relations of the lattice
  here. Also, partial objects can’t be merged here. **)
Inductive IFormula_star_closure : binary IFormula :=

  | IFormula_star_emp_neutral_left : forall phi,
      IFormula_star_closure (phi * emp) phi

  | IFormula_star_commut : forall phi1 phi2,
      IFormula_star_closure (phi1 * phi2) (phi2 * phi1)

  | IFormula_star_assoc_left : forall phi1 phi2 phi3,
      IFormula_star_closure ((phi1 * phi2) * phi3) (phi1 * (phi2 * phi3))

  | IFormula_star_congruence_left : forall phi phi' phic,
      IFormula_star_closure phi phi' →
      IFormula_star_closure (phi * phic) (phi' * phic)

  | IFormula_star_refl : forall phi,
      IFormula_star_closure phi phi

  | IFormula_star_sym : forall phi1 phi2,
      IFormula_star_closure phi1 phi2 →
      IFormula_star_closure phi2 phi1

  | IFormula_star_trans : forall phi1 phi2 phi3,
      IFormula_star_closure phi1 phi2 →
      IFormula_star_closure phi2 phi3 →
      IFormula_star_closure phi1 phi3
  .

Lemma IFormula_star_emp_neutral_right : forall phi,
  IFormula_star_closure phi (phi * emp).
Proof. introv. apply~ IFormula_star_sym. apply~ IFormula_star_emp_neutral_left. Qed.

Lemma IFormula_star_congruence_right : forall phi phi' phic,
  IFormula_star_closure phi phi' →
  IFormula_star_closure (phic * phi) (phic * phi').
Proof.
  introv T. apply~ IFormula_star_trans.
   apply~ IFormula_star_commut.
   apply~ IFormula_star_trans.
    applys~ IFormula_star_congruence_left T.
    apply~ IFormula_star_commut.
Qed.

Lemma IFormula_star_assoc_right : forall phi1 phi2 phi3,
  IFormula_star_closure (phi1 * (phi2 * phi3)) ((phi1 * phi2) * phi3).
Proof. introv. apply~ IFormula_star_sym. apply~ IFormula_star_assoc_left. Qed.

Lemma IFormula_sub_star_closure : forall phi phi',
  IFormula_sub phi phi' →
  exists phic, IFormula_star_closure phi (phi' * phic).
Proof.
  introv S. induction S.
   exists emp. apply~ IFormula_star_emp_neutral_right.
   lets (phic&T): (rm IHS). inverts H as S'.
    exists (phic * phi2). apply~ IFormula_star_trans.
     applys~ IFormula_star_congruence_left T.
     apply~ IFormula_star_assoc_left.
    exists (phic * phi1). apply~ IFormula_star_trans.
     apply~ IFormula_star_commut.
     apply~ IFormula_star_trans.
      applys~ IFormula_star_congruence_left T.
      apply~ IFormula_star_assoc_left.
Qed.

Lemma IFormula_sub_star_closure_loc : forall phi phi' h o,
  IFormula_star_closure phi phi' →
  IFormula_sub phi (h --> o)
  ↔ IFormula_sub phi' (h --> o).
Proof.
  introv T. induction T.
   iff S.
    do 2 (inverts S as S S'; autos~). do 2 (inverts S' as S' S''; autos~).
    repeat constructors*.
   iff S.
    do 2 (inverts S as S S'; autos~); repeat constructors*.
    do 2 (inverts S as S S'; autos~); repeat constructors*.
   iff S.
    do 2 (inverts S as S S'; autos~); repeat (inverts S' as S' S''; autos~).
     repeat constructors*.
     applys~ rtclosure_step IFormula_sub_step_star_right.
      applys~ rtclosure_step IFormula_sub_step_star_left.
     do 2 applys~ rtclosure_step IFormula_sub_step_star_right. apply IFormula_sub_refl.
     do 2 applys~ rtclosure_step IFormula_sub_step_star_right.
      repeat constructors*.
     repeat applys~ rtclosure_step IFormula_sub_step_star_right.
    do 2 (inverts S as S S'; autos~); repeat (inverts S' as S' S''; autos~).
     repeat constructors*.
     repeat applys~ rtclosure_step IFormula_sub_step_star_left.
     do 2 applys~ rtclosure_step IFormula_sub_step_star_left.
      applys~ rtclosure_step IFormula_sub_step_star_right.
     applys~ rtclosure_step IFormula_sub_step_star_left.
      applys~ rtclosure_step IFormula_sub_step_star_right.
     repeat applys~ rtclosure_step IFormula_sub_step_star_right.
   iff S.
    do 2 (inverts S as S S'; autos~); repeat (inverts S' as S' S''; autos~).
     applys~ rtclosure_step IFormula_sub_step_star_left.
      rewrite <- IHT. apply~ IFormula_sub_refl.
     applys~ rtclosure_step IFormula_sub_step_star_left.
      rewrite <- IHT. applys~ rtclosure_step IFormula_sub_step_star_left.
     applys~ rtclosure_step IFormula_sub_step_star_left.
      rewrite <- IHT. applys~ rtclosure_step IFormula_sub_step_star_right.
     applys~ rtclosure_step IFormula_sub_step_star_right. apply~ IFormula_sub_refl.
     applys~ rtclosure_step IFormula_sub_step_star_right. repeat constructors*.
     repeat applys~ rtclosure_step IFormula_sub_step_star_right.
    do 2 (inverts S as S S'; autos~); repeat (inverts S' as S' S''; autos~).
     applys~ rtclosure_step IFormula_sub_step_star_left.
      rewrite IHT. apply~ IFormula_sub_refl.
     applys~ rtclosure_step IFormula_sub_step_star_left.
      rewrite IHT. applys~ rtclosure_step IFormula_sub_step_star_left.
     applys~ rtclosure_step IFormula_sub_step_star_left.
      rewrite IHT. applys~ rtclosure_step IFormula_sub_step_star_right.
     applys~ rtclosure_step IFormula_sub_step_star_right. apply~ IFormula_sub_refl.
     applys~ rtclosure_step IFormula_sub_step_star_right. repeat constructors*.
     repeat applys~ rtclosure_step IFormula_sub_step_star_right.
   iff~.
   symmetry. autos~.
   rewrite~ IHT1.
Qed.

Lemma IFormula_sub_star_closure_env : forall phi phi' ke E,
  IFormula_star_closure phi phi' →
  IFormula_sub phi (ke ~~> E)
  ↔ IFormula_sub phi' (ke ~~> E).
Proof.
  introv T. induction T.
   iff S.
    do 2 (inverts S as S S'; autos~). do 2 (inverts S' as S' S''; autos~).
    repeat constructors*.
   iff S.
    do 2 (inverts S as S S'; autos~); repeat constructors*.
    do 2 (inverts S as S S'; autos~); repeat constructors*.
   iff S.
    do 2 (inverts S as S S'; autos~); repeat (inverts S' as S' S''; autos~).
     repeat constructors*.
     applys~ rtclosure_step IFormula_sub_step_star_right.
      applys~ rtclosure_step IFormula_sub_step_star_left.
     do 2 applys~ rtclosure_step IFormula_sub_step_star_right. apply IFormula_sub_refl.
     do 2 applys~ rtclosure_step IFormula_sub_step_star_right.
      repeat constructors*.
     repeat applys~ rtclosure_step IFormula_sub_step_star_right.
    do 2 (inverts S as S S'; autos~); repeat (inverts S' as S' S''; autos~).
     repeat constructors*.
     repeat applys~ rtclosure_step IFormula_sub_step_star_left.
     do 2 applys~ rtclosure_step IFormula_sub_step_star_left.
      applys~ rtclosure_step IFormula_sub_step_star_right.
     applys~ rtclosure_step IFormula_sub_step_star_left.
      applys~ rtclosure_step IFormula_sub_step_star_right.
     repeat applys~ rtclosure_step IFormula_sub_step_star_right.
   iff S.
    do 2 (inverts S as S S'; autos~); repeat (inverts S' as S' S''; autos~).
     applys~ rtclosure_step IFormula_sub_step_star_left.
      rewrite <- IHT. apply~ IFormula_sub_refl.
     applys~ rtclosure_step IFormula_sub_step_star_left.
      rewrite <- IHT. applys~ rtclosure_step IFormula_sub_step_star_left.
     applys~ rtclosure_step IFormula_sub_step_star_left.
      rewrite <- IHT. applys~ rtclosure_step IFormula_sub_step_star_right.
     applys~ rtclosure_step IFormula_sub_step_star_right. apply~ IFormula_sub_refl.
     applys~ rtclosure_step IFormula_sub_step_star_right. repeat constructors*.
     repeat applys~ rtclosure_step IFormula_sub_step_star_right.
    do 2 (inverts S as S S'; autos~); repeat (inverts S' as S' S''; autos~).
     applys~ rtclosure_step IFormula_sub_step_star_left.
      rewrite IHT. apply~ IFormula_sub_refl.
     applys~ rtclosure_step IFormula_sub_step_star_left.
      rewrite IHT. applys~ rtclosure_step IFormula_sub_step_star_left.
     applys~ rtclosure_step IFormula_sub_step_star_left.
      rewrite IHT. applys~ rtclosure_step IFormula_sub_step_star_right.
     applys~ rtclosure_step IFormula_sub_step_star_right. apply~ IFormula_sub_refl.
     applys~ rtclosure_step IFormula_sub_step_star_right. repeat constructors*.
     repeat applys~ rtclosure_step IFormula_sub_step_star_right.
   iff~.
   symmetry. autos~.
   rewrite~ IHT1.
Qed.

Global Instance IFormula_star_closure_decidable : forall phi1 phi2,
    Decidable (IFormula_star_closure phi1 phi2).
Admitted. (* Admitted *)


(** *** Domain **)

(** Does a given location has a given field present at left-hand side
  in the formula. **)
Definition locInDom phi h f :=
  exists o, IFormula_sub phi (h --> o) ∧ in_AObject_domain f (DObject_dom o).

(** Does a given environment location is present at left-hand side in the formula. **)
Definition envInDom phi ke :=
  exists E, IFormula_sub phi (ke ~~> E).

Lemma locInDom_emp : forall h f, ~ locInDom emp h f.
Proof. introv (o&S). forwards I: IFormula_sub_IFNames_incl S. forwards*: I. left~. Qed.

Lemma locInDom_env : forall h f ke E, ~ locInDom (ke ~~> E) h f.
Proof. introv (o&S). repeat inverts S as S. Qed.

Lemma locInDom_star : forall phi1 phi2 h f,
  locInDom (phi1 * phi2) h f →
  locInDom phi1 h f ∨ locInDom phi2 h f.
Proof.
  introv (o&S&I). sets_eq phi: (phi1 * phi2). sets_eq phi': (h --> o).
  gen EQphi EQphi'; induction S; introv E E'; substs.
   inverts E'.
   inverts H; [left|right]; exists o; splits~; apply S.
Qed.

Lemma locInDom_star_inv : forall phi1 phi2 h f,
  locInDom phi1 h f ∨ locInDom phi2 h f →
  locInDom (phi1 * phi2) h f.
Proof. introv [(o&S)|(o&S)]; exists o; rewrite* IFormula_sub_star. Qed.

Lemma locInDom_here : forall h1 h2 f o,
  locInDom (h1 --> o) h2 f →
  h1 = h2.
Proof. introv (o'&S). repeat inverts S as S. autos~. Qed.

Lemma locInDom_sub : forall phi phi' h f,
  locInDom phi h f →
  IFormula_sub phi' phi →
  locInDom phi' h f.
Proof. introv (o&S) S'. exists o. splits*. applys rtclosure_trans S' S. Qed.

Lemma envInDom_emp : forall ke, ~ envInDom emp ke.
Proof. introv (E&S). forwards I: IFormula_sub_IFENames_incl S. forwards*: I. left~. Qed.

Lemma envInDom_loc : forall ke h o, ~ envInDom (h --> o) ke.
Proof. introv (E&S). repeat inverts S as S. Qed.

Lemma envInDom_star : forall phi1 phi2 ke,
  envInDom (phi1 * phi2) ke →
  envInDom phi1 ke ∨ envInDom phi2 ke.
Proof.
  introv (E&S). sets_eq phi: (phi1 * phi2). sets_eq phi': (ke ~~> E).
  gen EQphi EQphi'; induction S; introv E1 E2; substs.
   inverts E2.
   inverts H; [left|right]; exists E; apply S.
Qed.

Lemma envInDom_star_inv : forall phi1 phi2 ke,
  envInDom phi1 ke ∨ envInDom phi2 ke →
  envInDom (phi1 * phi2) ke.
Proof. introv [(E&S)|(E&S)]; exists E; rewrite* IFormula_sub_star. Qed.

Lemma envInDom_here : forall ke1 ke2 E,
  envInDom (ke1 ~~> E) ke2 →
  ke1 = ke2.
Proof. introv (E'&S). repeat inverts S as S. autos~. Qed.

Lemma envInDom_sub : forall phi phi' ke,
  envInDom phi ke →
  IFormula_sub phi' phi →
  envInDom phi' ke.
Proof. introv (E&S) S'. exists E. applys rtclosure_trans S' S. Qed.

Local Instance ALoc_AObject_domain_inhab : Inhab (ALoc * AObject_domain).
  apply prod_inhab.
Defined.


(** The domain of a formula is the set of all left-hand side locations
  (objects and environments, objects being considered at the field
  level).  It is not defined for incoherent formulae.  It is mainly
  used to decide the incoherence of a formula.  It also returns a list
  of set locations which the formula forces to be empty. **)
Fixpoint IFormula_domain phi : option (list (ALoc * AObject_domain) * list KE * list K * list KE) :=
  match phi with
  | emp => Some (nil, nil, nil, nil)
  | phi1 * phi2 =>
    LibOption.apply (fun xhs1 : _ * _ =>
      LibOption.apply (fun xhs2 : _ * _ =>
        let '(hds1, kes1, ks1, kees1) := xhs1 in
        let '(hds2, kes2, ks2, kees2) := xhs2 in
        'let hds2_filtered :=
          filter (fun hd2 => ! mem_assoc (fst hd2) hds1) hds2 in
        'let hds1_try_to_merge :=
           map (fun hd1 : _ * _ =>
           let (h, d1) := hd1 in
           (h, if mem_assoc h hds2 then
                 let d2 := assoc h hds2 in
                 pick_option (AObject_domain_merge d1 d2)
               else Some d1)) hds1 in
        'let hs_lacuna :=
          map fst (filter (fun hd => decide (snd hd = None)) hds1_try_to_merge) in
        'let hds1_merged :=
           map (fun hd =>
             match hd with
             | (h, Some d) => (h, d)
             | (_, None) => arbitrary
             end) (filter (fun hd => decide (snd hd <> None)) hds1_try_to_merge) in
        'let kes_lacuna :=
          filter (fun ke1 => mem ke1 kes2) kes1 in
        LibOption.map (fun ks_lacuna =>
          (hds1_merged ++ hds2_filtered,
           Remove_duplicates (kes1 ++ kes2),
           Remove_duplicates (ks_lacuna ++ ks1 ++ ks2),
           kes_lacuna ++ kees1 ++ kees2)) (map_partial (fun h =>
            match h with
            | singleton_loc l => None
            | set_loc k => Some k
            end) hs_lacuna))
      (IFormula_domain phi2))
    (IFormula_domain phi1)
  | h --> o =>
    ifb DObject_incoherent _ _ _ _ o then
      match h with
      | singleton_loc l => None
      | set_loc k => Some (nil, nil, [k], nil)
      end
    else Some ([(h, DObject_dom o)], nil, nil, nil)
  | ke ~~> E =>
    Some (nil, [ke], nil, nil)
  end.

(** Here follows a simpler version which does not recognise incoherent
  formulae. **)
Fixpoint IFormula_domain_simple phi : list (ALoc * AObject_domain) * list KE :=
  match phi with
  | emp => (nil, nil)
  | phi1 * phi2 =>
    let '(hds1, kes1) := IFormula_domain_simple phi1 in
    let '(hds2, kes2) := IFormula_domain_simple phi2 in
    'let merge := fun h =>
      let d1 := if mem_assoc h hds1 then assoc h hds1 else AObject_domain_empty in
      let d2 := if mem_assoc h hds2 then assoc h hds2 else AObject_domain_empty in
      let d3 := proj1_sig (AObject_domain_simple_merge d1 d2) in
      unsome_default d3 (pick_option (AObject_domain_merge d1 d2)) in
    let hs := Remove_duplicates (map fst (hds1 ++ hds2)) in
    let hds :=
      map (fun h =>
        (h, merge h)) hs in
    (hds, Remove_duplicates (kes1 ++ kes2))
  | h --> o => ([(h, DObject_dom o)], nil)
  | ke ~~> E => (nil, [ke])
  end.

Lemma IFormula_domain_mem_assoc : forall phi hds kes ks kees h d,
  IFormula_domain phi = Some (hds, kes, ks, kees) →
  mem (h, d) hds →
  assoc h hds = d.
Proof.
  induction phi; introv E I; inverts E as E.
   false*.
   do 2 destruct IFormula_domain as [[[[? ?] ?] ?]|]; tryfalse. simpl in E. repeat let_name.
    forwards (hds1_merged_o&E1&E2): LibOption.apply_on_inv (rm E). inverts E2.
    rewrite mem_app in I. rew_refl in I. inverts I as I.
     rewrite EQhds1_merged in I. rewrite map_mem with (CA := _) in I. lets ((h'&do)&Mf&Ed): (rm I).
      rewrite filter_mem_eq in Mf. rew_refl in Mf.
      destruct do as [d'|]; tryfalse*. inverts Ed. rewrite EQhds1_try_to_merge in Mf.
      rewrite map_mem with (CA := _) in Mf. lets (((h''&d'')&Il&Eh')&_): (rm Mf). inverts Eh'.
      forwards~ A1: IHphi1 Il. asserts I1: (mem (h'', Some d') hds1_try_to_merge).
        rewrite EQhds1_try_to_merge. rewrite map_mem. eexists. splits*.
      asserts I2: (mem (h'', d') hds1_merged).
        rewrite EQhds1_merged. rewrite map_mem. exists (h'', Some d'). splits~.
        rewrite filter_mem_eq. rew_refl. splits*. discriminate.
      asserts A2: (assoc h'' hds1_try_to_merge = Some d').
       subst hds1_try_to_merge. clear - A1 H1 I1. induction l as [|[h d1] l].
        false*.
        rewrite map_cons. rewrite assoc_cons in *. case_if.
         substs~. rewrite H1. clear. repeat case_if*.
         case_if in A1; tryfalse~. apply~ IHl. rewrite map_cons in I1.
          simpl in I1. rew_refl in I1. repeat inverts I1 as I1; autos*.
      rewrite app_assoc_l.
       subst hds1_merged. clear - I1 A2. induction hds1_try_to_merge as [|[h do] l].
        false*.
        simpl in I1. rew_refl in I1. rewrite filter_cons. tests: (h = h'').
         rewrite assoc_cons in A2. case_if.
          case_if; tryfalse*. rewrite map_cons. rewrite A2. simpl. case_if~. false*.
          case_if; tryfalse*. rename n into I. false I. substs. discriminate.
         inverts I1 as I1.
          inverts I1. false*.
          case_if.
           rewrite map_cons. destruct do; tryfalse~. simpl. case_if; tryfalse~.
            apply~ IHl. simpl in A2. case_if~. false*.
           apply~ IHl. simpl in A2. case_if~. false*.
       rewrite mem_assoc_map_fst. rewrite* map_mem.
     rewrite EQhds2_filtered in I. rewrite filter_mem_eq in I.
      rew_refl in I. rewrite app_assoc_r.
       rewrite EQhds2_filtered. rewrite* assoc_filter.
        fold_bool. rew_refl. apply* I.
       introv A. subst hds1_merged. rewrite mem_assoc_map_fst in A.
        rewrite list_map_compose in A. rewrite map_mem in A. lets ((h'&do)&Id&Eh): (rm A).
        rewrite filter_mem_eq in Id. rew_refl in Id. destruct do as [d'|]; tryfalse*.
        compute in Eh. subst h' hds1_try_to_merge. rewrite map_mem in Id.
        lets (((h'&do)&Il&E)&_): (rm Id). inverts E. forwards*: mem_mem_assoc Il.
   cases_if.
    destruct a; inverts E. false*.
    inverts E. simpl in I. rew_refl in I. do 2 inverts I as I. simpl. case_if*.
   false*.
Qed.

Lemma IFormula_domain_simple_mem_assoc : forall phi hds kes h d,
  IFormula_domain_simple phi = (hds, kes) →
  mem (h, d) hds →
  assoc h hds = d.
Proof.
  introv E I. destruct phi; inverts E as E.
   false*.
   do 2 destruct IFormula_domain_simple. inverts E as E1 E2.
    applys~ unique_assoc_assoc_mem I. clear. introv N1 N2.
     forwards (hd1&E1&N1'): Nth_map_inv (rm N1). forwards (hd2&E2&N2'): Nth_map_inv (rm N2).
     inverts E1. inverts E2. applys~ No_duplicates_Nth N1' N2'.
      typeclass.
      apply~ Remove_duplicates_spec.
   simpl in I. rew_refl in I. do 2 inverts I as I. simpl. case_if*.
   false*.
Qed.

(** Relations between [IFormula_domain_simple] and [IFormula_domain]. **)
Lemma IFormula_domain_simple_IFormula_domain_obj_loc : forall phi hds1 hds2 kes1 kes2 ks2 kees2,
  IFormula_domain_simple phi = (hds1, kes1) →
  IFormula_domain phi = Some (hds2, kes2, ks2, kees2) →
  double_incl (map fst hds1) (map fst hds2 ++ map (id : K → ALoc) ks2).
Proof.
  induction phi; introv E1 E2; lets Eq0: E2; inverts E1 as E1; inverts E2 as E2.
   apply double_incl_refl.
   destruct (IFormula_domain phi1) as [hkes1|] eqn: Eq1; tryfalse.
    destruct hkes1 as [hkes1' kees1']. destruct hkes1' as [hkes1' ks1'].
    destruct hkes1' as [hds1' kes1'].
    destruct (IFormula_domain phi2) as [hkes2|] eqn: Eq2; tryfalse.
    destruct hkes2 as [hkes2' kees2']. destruct hkes2' as [hkes2' ks2'].
    destruct hkes2' as [hds2' kes2'].
    destruct (IFormula_domain_simple phi1) as [hds1'' kes1''] eqn: Eq1'.
    destruct (IFormula_domain_simple phi2) as [hds2'' kes2''] eqn: Eq2'.
    let_name. inverts E1. forwards~ I1: (rm IHphi1). forwards~ I3: (rm IHphi2).
    simpl in E2. repeat let_name.
    forwards (hds1_merged_o&E3&E4): LibOption.apply_on_inv (rm E2).
    forwards F: map_partial_inv (rm E3). inverts E4.
    apply mem_double_incl with (H := _); introv.
    repeat rewrite mem_app with (CA := _). extens. rew_refl.
    repeat rewrite map_mem with (CA := _). iff I; [| destruct* I as [I|I] ].
     lets ((h&do)&Mh&Eh): (rm I). rewrite map_mem with (CA := _) in Mh.
      lets (h'&Mh'&Eh'): (rm Mh). inverts Eh'. rewrite Remove_duplicates_mem in Mh'.
      rewrite map_mem with (CA := _) in Mh'. lets ((h''&do')&Mh''&Eh''): (rm Mh').
      simpl in Eh, Eh''. subst a h''. rewrite mem_app in Mh''. rew_refl in Mh''.
      tests N: (mem_assoc h' hds1'').
       rewrite mem_assoc_map_fst in N. rewrite map_mem with (CA := _) in N.
       lets ((h&do'')&Ih&Eh'): (rm N). simpl in Eh'. subst h.
       forwards Eq: incl_mem h' (proj1 I1).
         rewrite* map_mem.
        rewrite mem_app in Eq. rew_refl in Eq. inverts Eq as I';
          rewrite map_mem with (CA := _) in I'.
         lets ((h&do)&Ihdo&Eh): (rm I'). simpl in Eh. subst h'.
          destruct (if mem_assoc h hds2'
                    then pick_option (AObject_domain_merge do (assoc h hds2'))
                    else Some do) as [d|] eqn: E.
           left. exists (h, d). splits~. rewrite mem_app. rew_refl. left.
            rewrite EQhds1_merged. rewrite map_mem with (CA := _). exists (h, Some d).
            splits~. rewrite filter_mem_eq. rew_refl. splits; [| discriminate ].
            rewrite EQhds1_try_to_merge. rewrite map_mem. exists (h, do). splits*. fequals~.
           right. asserts I': (mem h hs_lacuna).
             rewrite EQhs_lacuna.
             asserts WTF: (exists y,
               mem y (filter (fun hd => decide (snd hd = None)) hds1_try_to_merge)
               ∧ h = fst y).
              exists (h, @None AObject_domain). splits~. rewrite filter_mem_eq. rew_refl.
               splits~. rewrite EQhds1_try_to_merge. rewrite map_mem. eexists. splits*.
               fequals~.
             rewrite <- map_mem in WTF. apply* WTF.
            lets (n&N): mem_Nth I'. forwards N': Forall2_Nth_nth_def (arbitrary : K) F N.
            forwards Eq: Forall2_forall_Nth (rm F) N N'. destruct h as [l|k]; tryfalse.
            exists k. inverts Eq. splits~. rewrite Remove_duplicates_mem.
            repeat rewrite mem_app. rew_refl. left. applys~ Nth_mem N'.
         lets (k&Mk&Ek): (rm I'). subst h'. right. exists k. splits~.
          rewrite Remove_duplicates_mem. repeat rewrite mem_app. rew_refl*.
       inverts Mh'' as I.
        false N. rewrite mem_assoc_map_fst. rewrite* map_mem.
        forwards Eq: incl_mem h' (proj1 I3).
          rewrite* map_mem.
         rewrite mem_app in Eq. rew_refl in Eq. inverts Eq as I';
           rewrite map_mem with (CA := _) in I'.
          lets ((h&do)&Ihdo&Eh): (rm I'). simpl in Eh. subst h'.
           left. exists (h, do). splits~. rewrite mem_app. rew_refl.
           right. rewrite EQhds2_filtered. rewrite filter_mem_eq. rew_refl. splits~.
           introv A. apply N. rewrite mem_assoc_map_fst. erewrite* double_incl_mem.
           rewrite mem_app. rew_refl. rewrite* mem_assoc_map_fst in A.
          lets (k&Mk&Ek): (rm I'). subst h'. right. exists k. splits~.
           rewrite Remove_duplicates_mem. repeat rewrite mem_app. rew_refl*.
     lets ((h&d)&Ih&Eh): (rm I). simpl in Eh. subst a. rewrite mem_app in Ih.
      rew_refl in Ih. exists (h, merge h). splits; [| substs~ ].
      rewrite map_mem with (CA := _). exists h. splits~. rewrite Remove_duplicates_mem.
      rewrite map_app. rewrite mem_app. rew_refl. inverts Ih as Ih.
       left. erewrite* double_incl_mem. rewrite mem_app. rew_refl. left.
        subst hds1_merged. rewrite map_mem with (CA := _) in Ih.
        lets ((h'&d')&Mf&E): (rm Ih). rewrite filter_mem_eq in Mf. rew_refl in Mf.
        subst hds1_try_to_merge. rewrite map_mem with (CA := _) in Mf.
        lets (((h''&d'')&Mh&Eh)&D): (rm Mf). destruct d'; tryfalse~.
        inverts Eh as Ed. inverts E. rewrite* map_mem.
       right. erewrite* double_incl_mem. rewrite mem_app. rew_refl. left.
        subst hds2_filtered. rewrite filter_mem_eq in Ih. rew_refl in Ih.
        rewrite* map_mem.
     lets (k&Mk&Ek): (rm I). rewrite Remove_duplicates_mem in Mk.
      repeat rewrite mem_app in Mk. rew_refl in Mk.
       asserts I: (mem (k : ALoc) (map fst (hds1'' ++ hds2''))).
        rewrite map_app. rewrite mem_app. rew_refl. repeat destruct* Mk as [Mk|Mk].
         left. erewrite* double_incl_mem. rewrite mem_app. rew_refl. left.
          lets (n&N): mem_Nth Mk. apply Forall2_swap in F.
          forwards N': Forall2_Nth_nth_def (arbitrary : ALoc) F N.
          forwards Eq: Forall2_forall_Nth (rm F) N N'. destruct nth_def; inverts Eq.
          apply Nth_mem with (CA := _) in N'. rewrite EQhs_lacuna in N'.
          asserts ((h&d)&Mf&Eh): (exists y,
            mem y (filter (fun hd => decide (snd hd = None)) hds1_try_to_merge)
            ∧ (k : ALoc) = fst y).
           rewrite <- map_mem. apply N'.
          rewrite filter_mem_eq in Mf. rew_refl in Mf. rewrite EQhds1_try_to_merge in Mf.
          rewrite map_mem in Mf. lets (((h'&d')&Mh&Ed)&EN): (rm Mf). inverts Ed.
          rewrite* map_mem.
         left. erewrite* double_incl_mem. rewrite mem_app. rew_refl. right.
          rewrite* map_mem.
         right. erewrite* double_incl_mem. rewrite mem_app. rew_refl. right.
          rewrite* map_mem.
       evar (d : AObject_domain). exists (k : ALoc, d). splits~.
        rewrite map_mem with (CA := _). eexists. unfolds d. splits*.
        rewrite Remove_duplicates_mem. rewrite map_mem. rewrite map_mem in I.
        lets* ((h&d')&Mh&Eh): (rm I).
   cases_if.
    destruct a; inverts E2. apply double_incl_refl.
    inverts E2. apply double_incl_refl.
   apply double_incl_refl.
Qed.

Lemma IFormula_domain_simple_IFormula_domain_env_loc : forall phi hds1 hds2 kes1 kes2 ks2 kees2,
  IFormula_domain_simple phi = (hds1, kes1) →
  IFormula_domain phi = Some (hds2, kes2, ks2, kees2) →
  double_incl kes1 (kes2 ++ kees2).
Proof.
  induction phi; introv E1 E2; lets Eq0: E2; inverts E1 as E1; inverts E2 as E2.
   apply double_incl_refl.
   destruct (IFormula_domain phi1) as [hkes1|] eqn: Eq1; tryfalse.
    destruct hkes1 as [hkes1' kees1']. destruct hkes1' as [hkes1' ks1'].
    destruct hkes1' as [hds1' kes1'].
    destruct (IFormula_domain phi2) as [hkes2|] eqn: Eq2; tryfalse.
    destruct hkes2 as [hkes2' kees2']. destruct hkes2' as [hkes2' ks2'].
    destruct hkes2' as [hds2' kes2'].
    destruct (IFormula_domain_simple phi1) as [hds1'' kes1''] eqn: Eq1'.
    destruct (IFormula_domain_simple phi2) as [hds2'' kes2''] eqn: Eq2'.
    let_name. inverts E1. forwards~ I2: (rm IHphi1). forwards~ I4: (rm IHphi2).
    simpl in E2. repeat let_name.
    forwards (hds1_merged_o&E3&E4): LibOption.apply_on_inv (rm E2).
    forwards F: map_partial_inv (rm E3). inverts E4.
    apply mem_double_incl with (H := _); introv.
     rewrite mem_app. repeat rewrite Remove_duplicates_mem. repeat rewrite mem_app.
      erewrite* double_incl_mem. erewrite double_incl_mem with (l1 := kes2''); autos*.
      repeat rewrite mem_app. extens. rew_refl. iff [I|I]; repeat destruct~ I as [I|I].
      substs. rewrite filter_mem_eq in I. rew_refl* in I.
   cases_if.
    destruct a; inverts E2. apply double_incl_refl.
    inverts E2. apply double_incl_refl.
   apply double_incl_refl.
Qed.

Lemma IFormula_domain_IFormula_domain_simple : forall phi hds1 hds2 kes1 kes2 ks2 kees2 h d f,
  IFormula_domain_simple phi = (hds1, kes1) →
  IFormula_domain phi = Some (hds2, kes2, ks2, kees2) →
  mem (h, d) hds2 →
  in_AObject_domain f d →
  exists d', in_AObject_domain f d' ∧ mem (h, d') hds1.
Proof.
  induction phi; introv E1 E2 I ID; lets Eq0: E2; inverts E1 as E1; inverts E2 as E2.
   false*.
   destruct (IFormula_domain phi1) as [hkes1|] eqn: Eq1; tryfalse.
    destruct hkes1 as [hkes1' kees1']. destruct hkes1' as [hkes1' ks1'].
    destruct hkes1' as [hds1' kes1'].
    destruct (IFormula_domain phi2) as [hkes2|] eqn: Eq2; tryfalse.
    destruct hkes2 as [hkes2' kees2']. destruct hkes2' as [hkes2' ks2'].
    destruct hkes2' as [hds2' kes2'].
    destruct (IFormula_domain_simple phi1) as [hds1'' kes1''] eqn: Eq1'.
    destruct (IFormula_domain_simple phi2) as [hds2'' kes2''] eqn: Eq2'.
    let_name. inverts E1. simpl in E2. repeat let_name.
    forwards (hds1_merged_o&E3&E4): LibOption.apply_on_inv (rm E2).
    forwards F: map_partial_inv (rm E3). inverts E4. rewrite mem_app in I. rew_refl in I.
    tests I': (exists d', mem (h, d') hds1_merged ∧ in_AObject_domain f d').
     clear I. lets (d'&I&Id'): (rm I'). rewrite EQhds1_merged in I.
      unfolds ALoc. rewrite map_mem with (CA := _) in I. lets ((h'&d'')&I'&E): (rm I).
      rewrite filter_mem_eq in I'. rew_refl in I'. destruct d'' as [d''|]; tryfalse*.
      inverts E. rewrite EQhds1_try_to_merge in I'. rewrite map_mem with (CA := _) in I'.
      lets (((h&d')&I&E)&D): (rm I'). inverts E as E. tests Id'': (in_AObject_domain f d').
       forwards* (d'''&C&I'): IHphi1 I Id''. exists (merge h). splits.
        rewrite EQmerge. erewrite* IFormula_domain_simple_mem_assoc. case_if as I1.
         match goal with |- context [ pick_option ?P ] =>
           destruct (pick_option P) as [d1|] eqn: Eq end; simpl.
          apply pick_option_correct in Eq.
           erewrite <- AObject_domain_merge_total; autos*. rew_refl*.
          destruct AObject_domain_simple_merge as [okd OKD]. simpl. rewrite <- OKD.
           rew_refl*.
         fold_bool. rew_refl in I1. false I1. rewrite mem_assoc_map_fst. rewrite* map_mem.
        rewrite map_mem with (CA := _). exists h. splits~. rewrite Remove_duplicates_mem.
         rewrite map_mem with (CA := _). exists (h, d'''). rewrite mem_app. rew_refl*.
       case_if as I1.
        symmetry in E. apply pick_option_correct in E.
         erewrite <- AObject_domain_merge_total in Id'; autos*. rew_refl in Id'.
         inverts Id' as Id; tryfalse~. fold_bool. rewrite mem_assoc_map_fst in I1.
         rewrite map_mem in I1. lets ((h'&d1)&I2&E'): (rm I1). simpl in E'. subst h'.
         erewrite IFormula_domain_mem_assoc in E, Id; autos*. forwards* (d2&I3&I'): IHphi2 I2.
         exists (merge h). splits~.
          rewrite EQmerge. destruct (mem_assoc h hds2'') eqn:I4.
           match goal with |- context [ pick_option ?P ] =>
             destruct (pick_option P) as [d3|] eqn: Eq end; simpl.
            apply pick_option_correct in Eq.
             erewrite <- AObject_domain_merge_total; autos*. rew_refl. right.
             erewrite* IFormula_domain_simple_mem_assoc.
            destruct AObject_domain_simple_merge as [okd OKD]. simpl. rewrite <- OKD.
             rew_refl. right. erewrite* IFormula_domain_simple_mem_assoc.
           fold_bool. rew_refl in I4. false I4. rewrite mem_assoc_map_fst. rewrite* map_mem.
         rewrite map_mem with (CA := _). exists h. splits~. rewrite Remove_duplicates_mem.
          rewrite map_mem with (CA := _). exists (h, d2). splits~. rewrite mem_app. rew_refl*.
        inverts* E.
     inverts I as I.
      false I'. splits*.
      erewrite EQhds2_filtered in I. rewrite filter_mem_eq in I. rew_refl in I.
       forwards* (d'&C&I1): IHphi2. exists (merge h). splits.
        rewrite EQmerge. erewrite (IFormula_domain_simple_mem_assoc _ _ _ Eq2' I1).
         destruct (mem_assoc h hds2'') eqn:I2.
          destruct pick_option eqn: E.
           apply pick_option_correct in E. erewrite <- AObject_domain_merge_total; autos*.
            rew_refl*.
           destruct AObject_domain_simple_merge as [okd OKD]. simpl. rewrite <- OKD.
            rew_refl*.
          fold_bool. rew_refl in I2. false I2. apply* mem_mem_assoc.
        rewrite map_mem with (CA := _). exists h. splits~. rewrite Remove_duplicates_mem.
         rewrite map_mem with (CA := _). exists (h, d'). splits~. rewrite mem_app. rew_refl*.
   cases_if.
    destruct a; inverts E2. false*.
    inverts E2. simpl in I. rew_refl in I. repeat inverts I as I. eexists. splits*.
     simpl. rew_refl*.
   false*.
Qed.

Lemma IFormula_domain_simple_IFormula_domain : forall phi hds1 hds2 kes1 kes2 ks2 kees2 h d f,
  IFormula_domain_simple phi = (hds1, kes1) →
  IFormula_domain phi = Some (hds2, kes2, ks2, kees2) →
  mem (h, d) hds1 →
  in_AObject_domain f d →
  (exists d', in_AObject_domain f d' ∧ mem (h, d') hds2)
  ∨ (exists (k : K), h = k ∧ mem k ks2).
Proof.
  induction phi; introv E1 E2 I ID; lets Eq0: E2; inverts E1 as E1; inverts E2 as E2.
   false*.
   destruct (IFormula_domain phi1) as [hkes1|] eqn: Eq1; tryfalse.
    destruct hkes1 as [hkes1' kees1']. destruct hkes1' as [hkes1' ks1'].
    destruct hkes1' as [hds1' kes1'].
    destruct (IFormula_domain phi2) as [hkes2|] eqn: Eq2; tryfalse.
    destruct hkes2 as [hkes2' kees2']. destruct hkes2' as [hkes2' ks2'].
    destruct hkes2' as [hds2' kes2'].
    destruct (IFormula_domain_simple phi1) as [hds1'' kes1''] eqn: Eq1'.
    destruct (IFormula_domain_simple phi2) as [hds2'' kes2''] eqn: Eq2'.
    let_name. inverts E1. simpl in E2. repeat let_name.
    forwards (hds1_merged_o&E3&E4): LibOption.apply_on_inv (rm E2).
    forwards F: map_partial_inv (rm E3). inverts E4.
    rewrite map_mem with (CA := _) in I. lets (h'&I2&E): (rm I). inverts E.
    rewrite Remove_duplicates_mem in I2. rewrite map_mem with (CA := _) in I2.
    lets ((h&d)&I&E): (rm I2). subst h'. rewrite mem_app in I. rew_refl in I.
    match goal with |- ?P =>
      asserts Lem: (forall d,
        mem (h, d) hds1'' ∨ mem (h, d) hds2'' →
        in_AObject_domain f d → P) end.
     simpl. clear d I ID EQmerge merge. introv I ID.
      tests T: (exists d', in_AObject_domain f d' ∧ mem (h, d') hds1').
       lets (d'&Id'&Mh): (rm T). tests Ih: (mem h hs_lacuna).
        right. lets (n&N): mem_Nth Ih. forwards N': Forall2_Nth_nth_def (arbitrary : K) F N.
         forwards Eq: Forall2_forall_Nth (rm F) N N'. destruct h; inverts Eq.
         eexists. splits*. rewrite Remove_duplicates_mem. repeat rewrite mem_app. rew_refl.
         left. apply* Nth_mem.
        left. tests MA: (mem_assoc h hds2').
         destruct (pick_option (AObject_domain_merge d' (assoc h hds2'))) eqn: E.
          exists a. splits.
           apply pick_option_correct in E. erewrite <- AObject_domain_merge_total; autos*.
            rew_refl*.
           rewrite mem_app. rew_refl. left. rewrite EQhds1_merged.
            rewrite map_mem with (CA := _). exists (h, Some a). splits~.
            rewrite filter_mem_eq. rew_refl. splits.
             rewrite EQhds1_try_to_merge. rewrite map_mem. exists (h, d'). splits*.
              fequals. case_if as I'; autos~. fold_bool. rew_refl in I'. false* I'.
             discriminate.
          false Ih. rewrite EQhs_lacuna. unfolds ALoc. rewrite map_mem with (CA := _).
           exists (h, None : option AObject_domain). splits~. rewrite filter_mem_eq. rew_refl.
           splits~. rewrite EQhds1_try_to_merge. rewrite map_mem. exists (h, d'). splits*.
           fequals. case_if as I'. rewrite~ E.
         exists d'. splits~. rewrite mem_app. rew_refl. left.
          rewrite EQhds1_merged. rewrite map_mem with (CA := _). exists (h, Some d'). splits~.
          rewrite filter_mem_eq. rew_refl. splits.
           rewrite EQhds1_try_to_merge. rewrite map_mem. exists (h, d'). splits*.
            fequals. cases_if~. fold_bool. false* MA.
           discriminate.
       inverts I as I.
        forwards* [(d'&Id'&Ih)|(k&Ek&Ik)]: IHphi1 I.
         false* T.
         subst h. right. exists k. splits~. rewrite Remove_duplicates_mem.
          repeat rewrite mem_app. rew_refl*.
        forwards* [(d'&Id'&Ih)|(k&Ek&Ik)]: IHphi2 I.
         tests T': (mem_assoc h hds1').
          rewrite mem_assoc_map_fst in T'. rewrite map_mem with (CA := _) in T'.
           lets ((h'&d0)&Id0&Eh): (rm T'). simpl in Eh. subst h'.
           asserts I': (mem (h, pick_option (AObject_domain_merge d0 d')) hds1_try_to_merge).
             rewrite EQhds1_try_to_merge. refine (proj2 (map_mem _ hds1' _) _).
             exists (h, d0). splits~. fequals. rewrite* mem_mem_assoc.
             erewrite* IFormula_domain_mem_assoc.
           destruct pick_option as [d''|] eqn: E.
            left. exists d''. splits.
             apply pick_option_correct in E. erewrite <- AObject_domain_merge_total; autos*.
              rew_refl*.
             rewrite mem_app. rew_refl. left. rewrite EQhds1_merged. rewrite map_mem.
              exists (h, Some d''). rewrite filter_mem_eq. rew_refl. repeat splits*.
              discriminate.
            right. asserts Ih': (mem h hs_lacuna).
              rewrite EQhs_lacuna. refine (proj2 (map_mem _ _ _) _).
              exists (h, @None AObject_domain). rewrite filter_mem_eq. rew_refl. splits~.
             lets (n&N): mem_Nth Ih'. forwards N': Forall2_Nth_nth_def (arbitrary : K) F N.
             forwards Eq: Forall2_forall_Nth (rm F) N N'. destruct h as [l|k]; inverts Eq.
             eexists. splits*. rewrite Remove_duplicates_mem. repeat rewrite mem_app.
             rew_refl. left. apply* Nth_mem.
          left. exists d'. splits~. rewrite mem_app. rew_refl. right. rewrite EQhds2_filtered.
           rewrite filter_mem_eq. rew_refl*.
         subst h. right. exists k. splits~. rewrite Remove_duplicates_mem.
          repeat rewrite mem_app. rew_refl*.
    rewrite EQmerge in ID. destruct pick_option eqn: E; simpl in ID.
     apply pick_option_correct in E. erewrite <- AObject_domain_merge_total in ID; autos*.
      rew_refl in ID. inverts ID as ID.
       cases_if as I'. fold_bool. rewrite mem_assoc_map_fst in I'.
        rewrite map_mem with (CA := _) in I'. lets ((h'&d')&Ih&Eh): (rm I'). simpl in Eh.
        subst h. erewrite IFormula_domain_simple_mem_assoc in ID; try eassumption; autos*.
       cases_if as I'. fold_bool. rewrite mem_assoc_map_fst in I'.
        rewrite map_mem with (CA := _) in I'. lets ((h'&d')&Ih&Eh): (rm I'). simpl in Eh.
        subst h. erewrite IFormula_domain_simple_mem_assoc in ID; try eassumption; autos*.
     destruct AObject_domain_simple_merge as [okd OKD]. simpl in ID. rewrite <- OKD in ID.
      rew_refl in ID. inverts ID as ID.
       cases_if as I'. fold_bool. rewrite mem_assoc_map_fst in I'.
        rewrite map_mem with (CA := _) in I'. lets ((h'&d')&Ih&Eh): (rm I'). simpl in Eh.
        subst h. erewrite IFormula_domain_simple_mem_assoc in ID; try eassumption; autos*.
       cases_if as I'. fold_bool. rewrite mem_assoc_map_fst in I'.
        rewrite map_mem with (CA := _) in I'. lets ((h'&d')&Ih&Eh): (rm I'). simpl in Eh.
        subst h. erewrite IFormula_domain_simple_mem_assoc in ID; try eassumption; autos*.
   simpl in I. rew_refl in I. repeat inverts I as I. cases_if.
    destruct a; inverts E2. right. exists k. splits~. simpl. rew_refl*.
    inverts E2. left. eexists. splits*. simpl. rew_refl*.
   false*.
Qed.

(** Relation between [IFormula_domain_simple] and [locInDom]. **)
Lemma IFormula_domain_simple_locInDom : forall phi h f hds kes,
  IFormula_domain_simple phi = (hds, kes) →
  locInDom phi h f ↔ exists d, mem (h, d) hds ∧ in_AObject_domain f d.
Proof.
  induction phi; introv E; simpl in E; inverts E as E; iff I.
   forwards*: locInDom_emp I.
   lets (d&A&M): (rm I). false*.
   destruct (IFormula_domain_simple phi1) as [hs1' kes1'] eqn: E1.
    destruct (IFormula_domain_simple phi2) as [hs2' kes2'] eqn: E2. let_name. inverts E.
    forwards [I'|I']: locInDom_star I.
     rewrite~ IHphi1 in I'. lets (d&Id&ID): (rm I'). exists (merge h). splits~.
      rewrite map_mem with (CA := _). exists h. splits~. rewrite Remove_duplicates_mem.
       rewrite map_mem. exists (h, d). rewrite mem_app. rew_refl*.
      rewrite EQmerge. case_if as M1.
       erewrite* IFormula_domain_simple_mem_assoc. destruct pick_option eqn: Eq; simpl.
        apply pick_option_correct in Eq. erewrite <- AObject_domain_merge_total.
         rew_refl. left*.
         autos*.
        match goal with |- context [ proj1_sig ?d ] => rewrite <- (proj2_sig d) end. rew_refl~.
       fold_bool. rew_refl in M1. false M1. rewrite mem_assoc_map_fst. rewrite* map_mem.
     rewrite~ IHphi2 in I'. lets (d&Id&ID): (rm I'). exists (merge h). splits~.
      rewrite map_mem with (CA := _). exists h. splits~. rewrite Remove_duplicates_mem.
       rewrite map_mem. exists (h, d). rewrite mem_app. rew_refl*.
      rewrite EQmerge. destruct (mem_assoc h hs2') eqn: M2.
       erewrite IFormula_domain_simple_mem_assoc with (hds := hs2'); autos*.
        destruct pick_option eqn: Eq; simpl.
         apply pick_option_correct in Eq. erewrite <- AObject_domain_merge_total.
          rew_refl. right*.
         autos*.
        match goal with |- context [ proj1_sig ?d ] => rewrite <- (proj2_sig d) end. rew_refl~.
       fold_bool. rew_refl in M2. false M2. rewrite mem_assoc_map_fst. rewrite* map_mem.
   destruct (IFormula_domain_simple phi1) as [hs1' kes1'].
    destruct (IFormula_domain_simple phi2) as [hs2' kes2']. let_name. inverts E.
    lets (d&Md&Id): (rm I). rewrite map_mem in Md. lets (h'&I&E): (rm Md). inverts E.
    rewrite Remove_duplicates_mem in I. rewrite map_app in I. rewrite mem_app in I.
    rew_refl in I. apply locInDom_star_inv. rewrite EQmerge in Id.
    destruct pick_option eqn: Eq.
     apply pick_option_correct in Eq. erewrite <- AObject_domain_merge_total in Id; autos*.
      rew_refl in Id. inverts Id as Id.
       case_if; tryfalse~. left. rewrite~ IHphi1. eexists. splits*. apply* assoc_mem_assoc.
       case_if; tryfalse~. right. rewrite~ IHphi2. eexists. splits*. apply* assoc_mem_assoc.
      match goal with H : context [ proj1_sig ?d ] |- _ => rewrite <- (proj2_sig d) in Id end.
       rew_refl in Id. inverts Id as Id.
        case_if; tryfalse~. left. rewrite~ IHphi1. eexists. splits*. apply* assoc_mem_assoc.
        case_if; tryfalse~. right. rewrite~ IHphi2. eexists. splits*. apply* assoc_mem_assoc.
   forwards: locInDom_here I. substs. simpl. eexists. rew_refl. splits*.
    lets (o&S&I'): (rm I). repeat inverts S as S. autos*.
   lets (o&M&I'): (rm I). simpl in M. rew_refl in M. eexists. inverts M as M; tryfalse.
    inverts M. splits*. apply IFormula_sub_refl.
   false* locInDom_env.
   lets (o&M&I'): (rm I). false*.
Qed.

Instance locInDom_decidable : forall phi h f,
    Decidable (locInDom phi h f).
  introv. destruct (IFormula_domain_simple phi) eqn: E.
  applys Decidable_equiv.
   symmetry. apply* IFormula_domain_simple_locInDom.
   destruct (mem_assoc h l && in_AObject_domain f (assoc h l)) eqn: E'.
    applys Decidable_equiv True.
     iff~ I. fold_bool. rew_refl in E'. eexists. splits*. apply* assoc_mem_assoc.
     typeclass.
    applys Decidable_equiv False.
     iff I; tryfalse. lets (d&Ml&Id): (rm I). fold_bool. rew_refl in E'. false E'. splits.
      apply* mem_mem_assoc.
      erewrite* IFormula_domain_simple_mem_assoc.
     typeclass.
Defined.

Lemma IFormula_domain_locs : forall phi hds kes ks kees h d f,
  IFormula_domain phi = Some (hds, kes, ks, kees) →
  mem (h, d) hds →
  in_AObject_domain f d →
  locInDom phi h f.
Proof.
  introv E M I. destruct (IFormula_domain_simple phi) eqn: E'.
  rewrite IFormula_domain_simple_locInDom; try apply~ E'.
  forwards* (d'&Id&Ih): IFormula_domain_IFormula_domain_simple E' E M I.
Qed.

Lemma IFormula_domain_locs_empty_set : forall phi hds kes ks kees k,
  IFormula_domain phi = Some (hds, kes, ks, kees) →
  mem k ks →
  exists f, locInDom phi k f.
Proof.
  induction phi; introv E M; simpl in E.
   inverts E. false*.
   destruct (IFormula_domain phi1) as [hkes1|] eqn: Eq1; tryfalse.
    destruct hkes1 as [hkes1' kees1']. destruct hkes1' as [hkes1' ks1'].
    destruct hkes1' as [hds1' kes1'].
    destruct (IFormula_domain phi2) as [hkes2|] eqn: Eq2; tryfalse.
    destruct hkes2 as [hkes2' kees2']. destruct hkes2' as [hkes2' ks2'].
    destruct hkes2' as [hds2' kes2'].
    simpl in E. repeat let_name.
    forwards (hds1_merged_o&E3&E4): LibOption.apply_on_inv (rm E).
    forwards F: map_partial_inv (rm E3). inverts E4.
    rewrite Remove_duplicates_mem in M. repeat rewrite mem_app in M. rew_refl in M.
    inverts M as M.
     lets (n&N): mem_Nth M. apply Forall2_swap in F.
      forwards N': Forall2_Nth_nth_def (arbitrary : ALoc) F N.
      forwards Eq: Forall2_forall_Nth (rm F) N N'. destruct nth_def; inverts Eq.
      apply Nth_mem with (CA := _) in N'. rewrite EQhs_lacuna in N'. unfolds ALoc.
      rewrite map_mem with (CA := _) in N'. lets ((h&do)&Mf&Ek): (rm N'). simpl in Ek.
      subst h. rewrite filter_mem_eq in Mf. rew_refl in Mf. rewrite EQhds1_try_to_merge in Mf.
      rewrite map_mem in Mf. lets (((h&d)&Ih&Eh)&EN): (rm Mf). simpl in EN. subst do.
      inverts Eh as E. case_if as I.
      tests T: (AObject_domain_intersect d (assoc (set_loc _ k) hds2')).
       lets (f&I1&I2): (rm T). exists f. apply locInDom_star_inv.
        left. apply* IFormula_domain_locs.
       lets (d'&ME): AObject_domain_not_intersect_merge T.
        lets* (d0&Ed0): pick_option_defined. rewrite Ed0 in E. false*.
     inverts M as M.
      forwards* (f&Id): IHphi1 M. exists f. apply* locInDom_star_inv.
      forwards* (f&Id): IHphi2 M. exists f. apply* locInDom_star_inv.
   case_if as I.
    rename d0 into I. destruct a; inverts E. simpl in M. rew_refl in M.
     repeat inverts M as M. lets* (f&I'): DObject_incoherent_in_AObject_domain I.
     exists f. eexists. splits*. apply~ IFormula_sub_refl.
    rename n into I. inverts E. false*.
   inverts E. false*.
Qed.

(** Relation between [IFormula_domain_simple] and [envInDom]. **)
Lemma IFormula_domain_simple_envInDom : forall phi ke hds kes,
  IFormula_domain_simple phi = (hds, kes) →
  envInDom phi ke ↔ mem ke kes.
Proof.
  induction phi; introv E; simpl in E; inverts E as E; iff I.
   forwards*: envInDom_emp I.
   false*.
   destruct (IFormula_domain_simple phi1) as [hs1' kes1'].
    destruct (IFormula_domain_simple phi2) as [hs2' kes2']. inverts E.
    rewrite Remove_duplicates_mem. rewrite mem_app. rew_refl.
    forwards [I'|I']: envInDom_star I.
     left. forwards*: IHphi1.
     right. forwards*: IHphi2.
   destruct (IFormula_domain_simple phi1) as [hs1' kes1'].
    destruct (IFormula_domain_simple phi2) as [hs2' kes2']. inverts E.
    rewrite Remove_duplicates_mem in I. rewrite mem_app in I. rew_refl in I.
    apply envInDom_star_inv. inverts I as I.
     left. forwards*: IHphi1.
     right. forwards*: IHphi2.
   false* envInDom_loc.
   false*.
   forwards: envInDom_here I. substs. simpl. rew_refl*.
   simpl in I. rew_refl in I. eexists. inverts* I. apply IFormula_sub_refl.
Qed.

Instance envInDom_decidable : forall phi ke,
    Decidable (envInDom phi ke).
  introv. destruct (IFormula_domain_simple phi) eqn: E.
  applys Decidable_equiv.
   symmetry. apply* IFormula_domain_simple_envInDom.
   typeclass.
Defined.

Lemma IFormula_domain_envs : forall phi hds kes ks kees ke,
  IFormula_domain phi = Some (hds, kes, ks, kees) →
  mem ke kes →
  envInDom phi ke.
Proof.
  introv E M. destruct (IFormula_domain_simple phi) eqn: E'.
  rewrite IFormula_domain_simple_envInDom; try apply~ E'.
  erewrite double_incl_mem.
   rewrite mem_app. rew_refl. left*.
   apply* IFormula_domain_simple_IFormula_domain_env_loc.
Qed.

Lemma IFormula_domain_envs_empty_set : forall phi hds kes ks kees ke,
  IFormula_domain phi = Some (hds, kes, ks, kees) →
  mem ke kees →
  envInDom phi ke.
Proof.
  introv E M. destruct (IFormula_domain_simple phi) eqn: E'.
  rewrite IFormula_domain_simple_envInDom; try apply~ E'.
  erewrite double_incl_mem.
   rewrite mem_app. rew_refl. right*.
   apply* IFormula_domain_simple_IFormula_domain_env_loc.
Qed.


Lemma IFormula_domain_star_none_commut : forall phi1 phi2,
  IFormula_domain (phi1 * phi2) = None →
  IFormula_domain (phi2 * phi1) = None.
Proof.
  introv A. simpls. destruct (IFormula_domain phi2) eqn: E1; autos~.
  destruct (IFormula_domain phi1) eqn: E2; autos~. simpls.
  destruct p as [hsh1 kees1]. destruct hsh1 as [hsh1 ks1]. destruct hsh1 as [hs1 kes1].
  destruct p0 as [hsh2 kees2]. destruct hsh2 as [hsh2 ks2]. destruct hsh2 as [hs2 kes2].
  repeat let_name. apply option_map_none_inv in A. apply map_partial_inv_none in A.
  rewrite Exists_iff_exists_mem with (CA := _) in A. lets (h&Mh&Eh): (rm A).
  destruct h as [l|h]; tryfalse. rewrite~ map_partial_none.
  rewrite Exists_iff_exists_mem with (CA := _). exists (l : ALoc). splits~.
  rewrite EQhs_lacuna0 in Mh. rewrite map_mem with (CA := _) in Mh.
  lets ((h&do)&Mf&E): (rm Mh). rewrite filter_mem_eq in Mf. simpl in E. subst h.
  rew_refl in Mf. rewrite EQhds1_try_to_merge0 in Mf. rewrite map_mem with (CA := _) in Mf.
  lets (((h&d)&Ih&Ed)&E): (rm Mf). simpl in E. subst do. inverts Ed as E. cases_if as MA.
  fold_bool. rewrite mem_assoc_map_fst in MA. rewrite map_mem with (CA := _) in MA.
  lets ((h&d')&I'&E'): (rm MA). simpl in E'. subst h. rewrite EQhs_lacuna.
  rewrite map_mem with (CA := _). exists (l : ALoc, None : option AObject_domain). splits~.
  rewrite filter_mem_eq. rew_refl. splits*. rewrite EQhds1_try_to_merge. rewrite map_mem.
  exists (l : ALoc, d'). splits*. fequals. erewrite IFormula_domain_mem_assoc in E; autos*.
  cases_if as A'.
   destruct (pick_option (AObject_domain_merge d' _)) eqn: E'; autos~.
    apply pick_option_correct in E'. apply AObject_domain_merge_sym in E'.
    erewrite IFormula_domain_mem_assoc in E'; autos*. forwards* (d0&Ed0): pick_option_defined.
    rewrite Ed0 in E. inverts E.
   fold_bool. rew_refl in A'. false A'. apply* mem_mem_assoc.
Qed.

Lemma IFormula_domain_sub_none : forall phi1 phi2,
  IFormula_sub phi1 phi2 →
  IFormula_domain phi2 = None →
  IFormula_domain phi1 = None.
Proof.
  introv S E. induction~ S. forwards~ IH: IHS. inverts H; simpl; rewrite~ IH.
  destruct~ (IFormula_domain phi1).
Qed.

(** Relation between [IFormula_domain] and [locInDom]. **)
Lemma IFormula_domain_locInDom : forall phi h hds kes ks kees f,
  IFormula_domain phi = Some (hds, kes, ks, kees) →
  locInDom phi h f →
    (exists d, in_AObject_domain f d ∧ mem (h, d) hds)
  ∨ exists (k : K), h = k ∧ mem k ks.
Proof.
  introv E1 I. destruct (IFormula_domain_simple phi) eqn: E2.
  rewrite* IFormula_domain_simple_locInDom in I. lets (d&Id&Ihd): (rm I).
  applys* IFormula_domain_simple_IFormula_domain Ihd.
Qed.

Lemma IFormula_domain_locInDom_inv : forall phi h hds kes ks kees f d,
  IFormula_domain phi = Some (hds, kes, ks, kees) →
  in_AObject_domain f d ∧ mem (h, d) hds →
  locInDom phi h f.
Proof.
  introv E1 I. destruct (IFormula_domain_simple phi) eqn: E2.
  rewrite IFormula_domain_simple_locInDom; try apply E2. lets (Id&I'): (rm I).
  forwards* (d'&Id'&I''): IFormula_domain_IFormula_domain_simple I'.
Qed.

Lemma IFormula_domain_envInDom : forall phi ke hs kes ks kees,
  IFormula_domain phi = Some (hs, kes, ks, kees) →
  envInDom phi ke ↔ mem ke kes ∨ mem ke kees.
Proof.
  introv Eq. iff I.
   destruct (IFormula_domain_simple phi) eqn: E2.
    rewrite IFormula_domain_simple_envInDom in I; try apply E2.
    erewrite double_incl_mem in I; [| apply* IFormula_domain_simple_IFormula_domain_env_loc ].
    rewrite mem_app in I. rew_refl~ in I.
   inverts I as I.
    applys~ IFormula_domain_envs Eq I.
    applys~ IFormula_domain_envs_empty_set Eq I.
Qed.


(** *** Incoherent inner formulae. **)

(** There formulae have empty concretisations. **)

Inductive IFormula_incoherent : IFormula → Prop :=

  | IFormula_incoherent_singleton_loc : forall (l : L) o1 o2,
      IFormula_incoherent (l --> o1 * l --> o2)

  | IFormula_incoherent_set_loc : forall phi (k : K) o1 o2,
      IFormula_incoherent (IFormula_substs phi nil (k : ALoc)) →
      IFormula_incoherent (phi * k --> o1 * k --> o2)

  | IFormula_incoherent_set_env : forall phi (ke : KE) E1 E2,
      IFormula_incoherent (IFormula_Esubsts phi nil ke) →
      IFormula_incoherent (phi * ke ~~> E1 * ke ~~> E2)

  | IFormula_incoherent_object : forall h o,
      DObject_incoherent _ _ _ _ o →
      IFormula_incoherent (h --> o)

  | IFormula_incoherent_congruence_left : forall phi phi',
      IFormula_incoherent phi →
      IFormula_incoherent (phi' * phi)

  | IFormula_incoherent_star_closure : forall phi phi',
      IFormula_incoherent phi →
      IFormula_star_closure phi phi' →
      IFormula_incoherent phi'
  .

Lemma IFormula_incoherent_congruence_right : forall phi phi',
  IFormula_incoherent phi →
  IFormula_incoherent (phi * phi').
Proof.
  introv A. apply~ IFormula_incoherent_star_closure.
   applys~ IFormula_incoherent_congruence_left A.
   apply~ IFormula_star_commut.
Qed.

Lemma IFormula_incoherent_sub : forall phi1 phi2,
  IFormula_sub phi1 phi2 →
  IFormula_incoherent phi2 →
  IFormula_incoherent phi1.
Proof.
  introv S A. forwards (phic&T): IFormula_sub_star_closure S.
  apply~ IFormula_incoherent_star_closure.
   applys~ IFormula_incoherent_congruence_left A.
   apply~ IFormula_star_trans.
    apply~ IFormula_star_commut.
    applys~ IFormula_star_sym T.
Qed.

Lemma IFormula_incoherent_star_singleton_loc : forall phi1 phi2 (l : L) o1 o2,
  IFormula_sub phi1 (l --> o1) →
  IFormula_sub phi2 (l --> o2) →
  IFormula_incoherent (phi1 * phi2).
Proof.
  introv S1 S2. forwards (phi'1&T1): IFormula_sub_star_closure S1.
  forwards (phi'2&T2): IFormula_sub_star_closure S2. apply~ IFormula_incoherent_star_closure.
   apply IFormula_incoherent_sub with (phi2 := l --> o1 * l --> o2).
    apply rtclosure_once. apply IFormula_sub_step_star_left.
    apply IFormula_incoherent_singleton_loc.
   apply IFormula_star_sym. apply~ IFormula_star_trans.
    applys~ IFormula_star_congruence_left T1.
    apply~ IFormula_star_trans.
     apply~ IFormula_star_assoc_left.
     apply~ IFormula_star_trans.
      apply~ IFormula_star_congruence_right. applys~ IFormula_star_congruence_right T2.
      apply~ IFormula_star_trans; [| apply~ IFormula_star_assoc_right ].
       apply~ IFormula_star_congruence_right. apply~ IFormula_star_trans.
        apply~ IFormula_star_commut.
        apply~ IFormula_star_trans.
         apply~ IFormula_star_assoc_left.
         apply~ IFormula_star_congruence_right. apply~ IFormula_star_refl.
Qed.

Lemma IFormula_incoherent_star_set_loc : forall phi phi0 phi1 phi2 (k : K) o1 o2,
  IFormula_incoherent (IFormula_substs phi0 nil (k : ALoc)) →
  IFormula_sub phi phi0 →
  IFormula_sub phi1 (k --> o1) →
  IFormula_sub phi2 (k --> o2) →
  IFormula_incoherent (phi * phi1 * phi2).
Proof.
  introv IS S1 S2 S3. forwards (phi'1&T1): IFormula_sub_star_closure S1.
  forwards (phi'2&T2): IFormula_sub_star_closure S2.
  forwards (phi'3&T3): IFormula_sub_star_closure S3. apply~ IFormula_incoherent_star_closure.
   apply IFormula_incoherent_sub with (phi2 := phi0 * k --> o1 * k --> o2).
    apply rtclosure_once. apply IFormula_sub_step_star_left.
    apply~ IFormula_incoherent_set_loc.
   apply~ IFormula_star_trans.
    apply~ IFormula_star_assoc_right.
    apply~ IFormula_star_trans.
     applys~ IFormula_star_congruence_left.
      apply~ IFormula_star_trans.
       apply~ IFormula_star_assoc_left.
       applys~ IFormula_star_congruence_right.
        applys~ IFormula_star_sym T3.
     apply~ IFormula_star_trans.
      apply~ IFormula_star_assoc_left.
      apply~ IFormula_star_trans.
       applys~ IFormula_star_congruence_right IFormula_star_commut.
       apply~ IFormula_star_trans.
        apply~ IFormula_star_assoc_right.
        apply~ IFormula_star_congruence_left.
         apply~ IFormula_star_trans.
          apply~ IFormula_star_assoc_right.
          apply~ IFormula_star_trans.
           applys~ IFormula_star_congruence_left.
            apply~ IFormula_star_trans.
             apply~ IFormula_star_assoc_left.
             applys~ IFormula_star_congruence_right.
              applys~ IFormula_star_sym T2.
           applys~ IFormula_star_trans.
            apply~ IFormula_star_assoc_left.
            apply~ IFormula_star_trans.
             applys~ IFormula_star_congruence_right IFormula_star_commut.
             apply~ IFormula_star_trans.
             apply~ IFormula_star_assoc_right.
              apply~ IFormula_star_congruence_left.
              applys~ IFormula_star_sym T1.
Qed.


(** ** Membranes **)

(** *** Definition **)

Inductive IMembrane :=
  | Membrane_rewriting : ALoc → list ALoc → IMembrane
  | Membrane_new : ALoc → IMembrane
  | Membrane_Erewriting : KE → list KE → IMembrane
  | Membrane_Enew : KE → IMembrane
  .
Notation "h '--->' hs" := (Membrane_rewriting h hs)
  (at level 56) : sep_scope.
Notation "'ν' h" := (Membrane_new h)
  (at level 56) : sep_scope.
Notation "ke '~~~>' kes" := (Membrane_Erewriting ke kes)
  (at level 56) : sep_scope.
Notation "'ν~' ke" := (Membrane_Enew ke)
  (at level 56) : sep_scope.

Definition Membrane := list IMembrane.

Global Instance IMembrane_comparable : Comparable IMembrane.
  prove_comparable_simple_inductive.
Defined.

(** The invariant of membrane states that a membrane rewrites a
  location at most once, and no abstract location appears twice in a
  right hand-side of a rewriting. **)
Record Membrane_correct (M : Membrane) := {
    Membrane_unique : forall n1 n2 h hs1 hs2,
      Nth n1 M (h ---> hs1) →
      Nth n2 M (h ---> hs2) →
      n1 = n2 ;
    Membrane_unique_env : forall n1 n2 ke kes1 kes2,
      Nth n1 M (ke ~~~> kes1) →
      Nth n2 M (ke ~~~> kes2) →
      n1 = n2 ;
    Membrane_rewrites_unique : forall h hs,
      mem (h ---> hs) M →
      No_duplicates hs ;
    Membrane_rewrites_unique_env : forall ke kes,
      mem (ke ~~~> kes) M →
      No_duplicates kes
  }.

(** A dependent type for membranes.  This will only be used outside of
   this file, where the implementation of membranes should not
   matter. **)
Definition DMembrane := { M : Membrane | Membrane_correct M }.

Lemma Membrane_func : forall M h hs1 hs2,
  Membrane_correct M →
  mem (h ---> hs1) M →
  mem (h ---> hs2) M →
  hs1 = hs2.
Proof.
  introv C I1 I2. forwards (n1&N1): mem_Nth (rm I1). forwards (n2&N2): mem_Nth (rm I2).
  forwards~: Membrane_unique N1 N2. substs. forwards E: Nth_func N1 N2. inverts~ E.
Qed.

Lemma Membrane_func_env : forall M ke kes1 kes2,
  Membrane_correct M →
  mem (ke ~~~> kes1) M →
  mem (ke ~~~> kes2) M →
  kes1 = kes2.
Proof.
  introv C I1 I2. forwards (n1&N1): mem_Nth (rm I1). forwards (n2&N2): mem_Nth (rm I2).
  forwards~: Membrane_unique_env N1 N2. substs. forwards E: Nth_func N1 N2. inverts~ E.
Qed.

(** We can safely remove rewritings from a membrane to get another. **)
Lemma Membrane_correct_sub : forall M1 M2 M M' IM,
  M2 = M ++ IM :: M' →
  M1 = M ++ M' →
  Membrane_correct M2 →
  Membrane_correct M1.
Proof.
  (** This lemma does not need these hypotheses, and I clear them to avoid unwanted
    dependencies. **)
  clear ABasic_Gamma ABasic_gamma_bot ABasic_gamma_top.
  introv E1 E2 C. substs. constructors.
  (** Uniqueness **)
  introv N1 N2. forwards [N1'|(m1&E1&N1')]: Nth_app_inv (rm N1);
    forwards [N2'|(m2&E2&N2')]: Nth_app_inv (rm N2); substs;
    let Nth_app N :=
      try applys Nth_app_l N; try (applys Nth_app_r; [ applys Nth_next N |]) in
    (forwards E: Membrane_unique C; [ Nth_app N1' | Nth_app N2' |]); substs~;
    apply Nth_lt_length in N1'; apply Nth_lt_length in N2'; math.
  (** Uniqueness env **)
  introv N1 N2. forwards [N1'|(m1&E1&N1')]: Nth_app_inv (rm N1);
    forwards [N2'|(m2&E2&N2')]: Nth_app_inv (rm N2); substs;
    let Nth_app N :=
      try applys Nth_app_l N; try (applys Nth_app_r; [ applys Nth_next N |]) in
    (forwards E: Membrane_unique_env C; [ Nth_app N1' | Nth_app N2' |]); substs~;
    apply Nth_lt_length in N1'; apply Nth_lt_length in N2'; math.
  (** Rewriting uniqueness **)
  introv I. applys Membrane_rewrites_unique C.
    clear - I. rewrite mem_app in *. rew_refl in *. simpl. rew_refl. inverts I as I.
     left*.
     do 2 right~.
  (** Rewriting uniqueness env **)
  introv I. applys Membrane_rewrites_unique_env C.
    clear - I. rewrite mem_app in *. rew_refl in *. simpl. rew_refl. inverts I as I.
     left*.
     do 2 right~.
Qed.

Lemma Membrane_nil_correct : Membrane_correct nil.
Proof.
  constructors~.
  (** Uniqueness **)
  introv N1 N2. inverts N1.
  (** Uniqueness env **)
  introv N1 N2. inverts N1.
  (** Rewriting uniqueness **)
  introv I. false*.
  (** Rewriting uniqueness env **)
  introv I. false*.
Qed.

Definition DMembrane_nil : DMembrane := exist _ _ Membrane_nil_correct.


(** *** Reading **)

(** New locations allocated by the membrane. **)
Definition Mnew (M : Membrane) :=
  concat (map (fun MI =>
    match MI with
    | _ ---> _ => nil
    | ν h => [h]
    | _ ~~~> _ => nil
    | ν~ _ => nil
    end) M).

Lemma Mnew_correct : forall h M,
  mem h (Mnew M) = mem (ν h) M.
Proof.
  unfolds Mnew. extens. iff I.
   rewrite concat_mem in I. lets (hs&Mhs&Mh): (rm I).
    rewrite map_mem in Mhs. lets (IM&MIM&E): (rm Mhs).
    destruct IM; substs; tryfalse. simpl in Mh. rew_refl in Mh. inverts* Mh.
   rewrite concat_mem. eexists. splits.
    rewrite* map_mem.
    simpls. rew_refl*.
Qed.

Definition Mnew_env (M : Membrane) :=
  concat (map (fun MI =>
    match MI with
    | _ ---> _ => nil
    | ν _ => nil
    | _ ~~~> _ => nil
    | ν~ ke => [ke]
    end) M).

Lemma Mnew_env_correct : forall ke M,
  mem ke (Mnew_env M) = mem (ν~ ke) M.
Proof.
  unfolds Mnew_env. extens. iff I.
   rewrite concat_mem in I. lets (hs&Mhs&Mh): (rm I).
    rewrite map_mem in Mhs. lets (IM&MIM&E): (rm Mhs).
    destruct IM; substs; tryfalse. simpl in Mh. rew_refl in Mh. inverts* Mh.
   rewrite concat_mem. eexists. splits.
    rewrite* map_mem.
    simpls. rew_refl*.
Qed.

(** Reading how a location is transformed by a membrane. **)
Fixpoint Mread (M : Membrane) h :=
  match M with
  | nil => None
  | ν _ :: M' => Mread M' h
  | h' ---> hs :: M' =>
    ifb h = h' then Some hs else Mread M' h
  | ν~ _ :: M' => Mread M' h
  | _ ~~~> _ :: M' => Mread M' h
  end.

Lemma Mread_mem : forall M h hs,
  Membrane_correct M →
  mem (h ---> hs) M ↔
  Mread M h = Some hs.
Proof.
  (** This lemma does not need these hypotheses, and I clear them to avoid unwanted
    dependencies. **)
  clear ABasic_Gamma ABasic_gamma_bot ABasic_gamma_top.
  introv C. induction M as [|IM].
   iff; false*.
   forwards IHM': IHM.
     applys Membrane_correct_sub C; rewrite~ app_nil_l.
    simpl. destruct IM as [h' h's|h'|ke kes|ke].
     cases_if as I; fold_bool; rew_refl in I.
      substs. iff Mh.
       fequals. applys Membrane_func C Mh. simpl; rew_refl*.
       inversion Mh. substs. rew_refl*.
      iff Mh.
       rew_refl in Mh. inverts Mh as Mh; [ inverts* Mh | autos* ].
       rew_refl*.
     rew_refl. rewrite IHM'. iff~ Mh. inverts Mh as Mh; [ inverts* Mh | autos* ].
     rew_refl. rewrite IHM'. iff~ Mh. inverts Mh as Mh; [ inverts* Mh | autos* ].
     rew_refl. rewrite IHM'. iff~ Mh. inverts Mh as Mh; [ inverts* Mh | autos* ].
Qed.

Lemma Mread_No_Duplicates : forall M h hs,
  Membrane_correct M →
  Mread M h = Some hs →
  No_duplicates hs.
Proof. introv C E. rewrite~ <- Mread_mem in E. apply* Membrane_rewrites_unique. Qed.

Fixpoint Mread_env (M : Membrane) ke :=
  match M with
  | nil => None
  | ν _ :: M' => Mread_env M' ke
  | _ ---> _ :: M' => Mread_env M' ke
  | ν~ _ :: M' => Mread_env M' ke
  | ke' ~~~> kes :: M' =>
    ifb ke = ke' then Some kes else Mread_env M' ke
  end.

Lemma Mread_env_mem : forall M ke kes,
  Membrane_correct M →
  mem (ke ~~~> kes) M ↔
  Mread_env M ke = Some kes.
Proof.
  (** This lemma does not need these hypotheses, and I clear them to avoid unwanted
    dependencies. **)
  clear ABasic_Gamma ABasic_gamma_bot ABasic_gamma_top.
  introv C. induction M as [|IM].
   iff; false*.
   forwards IHM': IHM.
     applys Membrane_correct_sub C; rewrite~ app_nil_l.
    simpl. destruct IM as [h' h's|h'|ke' kes'|ke'].
     rew_refl. rewrite IHM'. iff~ Mh. inverts Mh as Mh; [ inverts* Mh | autos* ].
     rew_refl. rewrite IHM'. iff~ Mh. inverts Mh as Mh; [ inverts* Mh | autos* ].
     cases_if as I; fold_bool; rew_refl in I.
      substs. iff Mh.
       fequals. applys Membrane_func_env C Mh. simpl; rew_refl*.
       inversion Mh. substs. rew_refl*.
      iff Mh.
       rew_refl in Mh. inverts Mh as Mh; [ inverts* Mh | autos* ].
       rew_refl*.
     rew_refl. rewrite IHM'. iff~ Mh. inverts Mh as Mh; [ inverts* Mh | autos* ].
Qed.

(** Getting the list of outside locations which generates a given
  inside location.  Note that there might be some additionnal [ν h]
  construction generating [h]. **)
Fixpoint MgetDependencies (M : Membrane) h :=
  match M with
  | nil => nil
  | ν _ :: M' => MgetDependencies M' h
  | h' ---> hs :: M' =>
    let d := MgetDependencies M' h in
    ifb mem h hs ∧ ~ mem h' d then h' :: d else d
  | ν~ _ :: M' => MgetDependencies M' h
  | _ ~~~> _ :: M' => MgetDependencies M' h
  end.

Lemma MgetDependencies_correct : forall M h h0,
  Membrane_correct M →
  mem h0 (MgetDependencies M h) →
  exists hs,
    Mread M h0 = Some hs ∧ mem h hs.
Proof.
  (** This lemma does not need these hypotheses, and I clear them to avoid unwanted
    dependencies. **)
  clear ABasic_Gamma ABasic_gamma_bot ABasic_gamma_top.
  introv C I. induction M.
   false.
   forwards C': Membrane_correct_sub C; try rewrite~ app_nil_l.
    destruct a as [h' hs|h'|ke kes|ke].
     simpl in I. cases_if as I'.
      fold_bool. simpl in I. rew_refl in *. inverts I as I.
       exists hs. splits*. rewrite~ <- Mread_mem. simpl. rew_refl*.
       forwards~ (hs'&Er&Ih): IHM I. exists hs'. splits~.
        rewrite <- Mread_mem in *; autos~. simpl. rew_refl*.
      forwards~ (hs'&Er&Ih): IHM I. exists hs'. splits~.
       rewrite~ <- Mread_mem in Er. rewrite~ <- Mread_mem. simpl. rew_refl*.
     forwards~ (hs'&Er&Ih): IHM I.
     forwards~ (hs'&Er&Ih): IHM I.
     forwards~ (hs'&Er&Ih): IHM I.
Qed.

Lemma MgetDependencies_Mread : forall M h hs h',
  Membrane_correct M →
  Mread M h = Some hs →
  mem h' hs →
  mem h (MgetDependencies M h').
Proof.
  (** This lemma does not need these hypotheses, and I clear them to avoid unwanted
    dependencies. **)
  clear ABasic_Gamma ABasic_gamma_bot ABasic_gamma_top.
  introv C E I. induction M.
   false.
   forwards C': Membrane_correct_sub C; try rewrite~ app_nil_l.
    destruct a as [h'' hs'|h''|ke kes|ke].
     simpl in E. cases_if as I'.
      inverts E. fold_bool. rew_refl in I'. substs. simpl. cases_if as I'.
       simpl. rew_refl*.
       fold_bool. rew_refl in I'. rew_logic* in I'.
      simpl. cases_if as I''.
       simpl. rew_refl*.
       rewrite* IHM.
     rewrite* IHM.
     rewrite* IHM.
     rewrite* IHM.
Qed.

Lemma MgetDependencies_No_Duplicates : forall M h,
  No_duplicates (MgetDependencies M h).
Proof.
  introv. induction M as [|IM M].
   constructors~.
   simpl. destruct~ IM. cases_if as I; autos~. fold_bool. rew_refl in I. constructors~.
    erewrite* Mem_mem.
Qed.

Fixpoint MgetDependencies_env (M : Membrane) ke :=
  match M with
  | nil => nil
  | ν _ :: M' => MgetDependencies_env M' ke
  | _ ---> _ :: M' => MgetDependencies_env M' ke
  | ν~ _ :: M' => MgetDependencies_env M' ke
  | ke' ~~~> kes :: M' =>
    let d := MgetDependencies_env M' ke in
    ifb mem ke kes  ∧ ~ mem ke' d then ke' :: d else d
  end.

Lemma MgetDependencies_env_correct : forall M ke ke0,
  Membrane_correct M →
  mem ke0 (MgetDependencies_env M ke) →
  exists kes,
    Mread_env M ke0 = Some kes ∧ mem ke kes.
Proof.
  (** This lemma does not need these hypotheses, and I clear them to avoid unwanted
    dependencies. **)
  clear ABasic_Gamma ABasic_gamma_bot ABasic_gamma_top.
  introv C I. induction M.
   false.
   forwards C': Membrane_correct_sub C; try rewrite~ app_nil_l.
    destruct a as [h' hs|h'|ke' kes'|ke'].
     forwards~ (hs'&Er&Ih): IHM I.
     forwards~ (hs'&Er&Ih): IHM I.
     simpl in I. cases_if as I'.
      fold_bool. simpl in I. rew_refl in *. inverts I as I.
       exists kes'. splits*. rewrite~ <- Mread_env_mem. simpl. rew_refl*.
       forwards~ (hs'&Er&Ih): IHM I. exists hs'. splits~.
        rewrite <- Mread_env_mem in *; autos~. simpl. rew_refl*.
      forwards~ (hs'&Er&Ih): IHM I. exists hs'. splits~.
       rewrite~ <- Mread_env_mem in Er. rewrite~ <- Mread_env_mem. simpl. rew_refl*.
     forwards~ (hs'&Er&Ih): IHM I.
Qed.

Lemma MgetDependencies_Mread_env : forall M ke kes ke',
  Membrane_correct M →
  Mread_env M ke = Some kes →
  mem ke' kes →
  mem ke (MgetDependencies_env M ke').
Proof.
  (** This lemma does not need these hypotheses, and I clear them to avoid unwanted
    dependencies. **)
  clear ABasic_Gamma ABasic_gamma_bot ABasic_gamma_top.
  introv C E I. induction M.
   false.
   forwards C': Membrane_correct_sub C; try rewrite~ app_nil_l.
    destruct a as [h'' hs'|h''|ke'' kes'|ke''].
     rewrite* IHM.
     rewrite* IHM.
     simpl in E. cases_if as I'.
      inverts E. fold_bool. rew_refl in I'. substs. simpl. cases_if as I'.
       simpl. rew_refl*.
       fold_bool. rew_refl in I'. rew_logic* in I'.
      simpl. cases_if as I''.
       simpl. rew_refl*.
       rewrite* IHM.
     rewrite* IHM.
Qed.

Lemma MgetDependencies_env_No_Duplicates : forall M ke,
  No_duplicates (MgetDependencies_env M ke).
Proof.
  introv. induction M as [|IM M].
   constructors~.
   simpl. destruct~ IM. cases_if as I; autos~. fold_bool. rew_refl in I. constructors~.
    erewrite* Mem_mem.
Qed.

Lemma MgetDependencies_app : forall M1 M2 h,
  double_incl
    (MgetDependencies (M1 ++ M2) h)
    (MgetDependencies M1 h ++ MgetDependencies M2 h).
Proof.
  introv. apply mem_double_incl with (H := _). induction M1 as [|IM M1]; intro h0.
   autos*.
   simpl. destruct~ IM. fold (M1 ++ M2). rewrite IHM1.
    cases_if as I1; cases_if as I2; fold_bool; rew_refl in I1; rew_refl in I2;
      simpl; rewrite~ IHM1; extens; rew_refl; iff* I; inverts I as I; autos*.
     rew_logic in I2. inverts* I2. rewrite mem_app. rew_refl*.
     rew_logic in I1. inverts* I1.
Qed.

Lemma MgetDependencies_env_app : forall M1 M2 ke,
  double_incl
    (MgetDependencies_env (M1 ++ M2) ke)
    (MgetDependencies_env M1 ke ++ MgetDependencies_env M2 ke).
Proof.
  introv. apply mem_double_incl with (H := _). induction M1 as [|IM M1]; intro h0.
   autos*.
   simpl. destruct~ IM. fold (M1 ++ M2). rewrite IHM1.
    cases_if as I1; cases_if as I2; fold_bool; rew_refl in I1; rew_refl in I2;
      simpl; rewrite~ IHM1; extens; rew_refl; iff* I; inverts I as I; autos*.
     rew_logic in I2. inverts* I2. rewrite mem_app. rew_refl*.
     rew_logic in I1. inverts* I1.
Qed.


(** *** Structure **)

(** **** [Membrane_inf] **)

(** An auxialiary definition to compute the comparisons operations about membranes. **)
Definition Membrane_compare_rewritings F FE (M1 M2 : Membrane) :=
  Forall (fun IM =>
    match IM with
    | h ---> hs =>
      Exists (fun IM =>
        match IM with
        | h' ---> hs' =>
          h = h' ∧ F hs hs'
        | _ => False
        end) M2
    | ν _ => True
    | ke ~~~> kes =>
      Exists (fun IM =>
        match IM with
        | ke' ~~~> kes' =>
          ke = ke' ∧ FE kes kes'
        | _ => False
        end) M2
    | ν~ _ => True
    end) M1.

Local Instance Membrane_compare_rewritings_decidable : forall M1 M2 F FE,
    (forall hs1 hs2, Decidable (F hs1 hs2)) →
    (forall kes1 kes2, Decidable (FE kes1 kes2)) →
    Decidable (Membrane_compare_rewritings F FE M1 M2).
  introv D DE. apply Forall_decidable. intros [h hs|h|ke kes|ke]; try typeclass.
   apply Exists_decidable. intros [h' hs'|h'|ke' kes'|ke']; try typeclass.
   apply Exists_decidable. intros [h' hs'|h'|ke' kes'|ke']; try typeclass.
Defined.

(** The order relation over membranes. **)
Record Membrane_inf (M1 M2 : Membrane) := {
    Membrane_inf_new : double_incl (Mnew M1) (Mnew M2) ;
    Membrane_inf_new_env : double_incl (Mnew_env M1) (Mnew_env M2) ;
    Membrane_inf_read_undefined : forall h,
      Mread M2 h = None ↔ Mread M1 h = None ;
    Membrane_inf_read_env_undefined : forall ke,
      Mread_env M2 ke = None ↔ Mread_env M1 ke = None ;
    Membrane_inf_read_defined : forall h hs1 hs2,
      Mread M1 h = Some hs1 →
      Mread M2 h = Some hs2 →
      incl hs1 hs2 ;
    Membrane_inf_read_env_defined : forall ke kes1 kes2,
      Mread_env M1 ke = Some kes1 →
      Mread_env M2 ke = Some kes2 →
      incl kes1 kes2
  }.

(** An alternative definition of [Membrane_inf]. **)
Definition Membrane_inf' (M1 M2 : Membrane) :=
  Membrane_compare_rewritings (@incl _) (@incl _) M1 M2
  ∧ double_incl (Mnew M1) (Mnew M2)
  ∧ double_incl (Mnew_env M1) (Mnew_env M2)
  ∧ Membrane_compare_rewritings (const2 True) (const2 True) M2 M1.

Lemma Membrane_inf_equiv_Membrane_inf' : forall M1 M2,
  Membrane_correct M1 →
  Membrane_correct M2 →
  Membrane_inf M1 M2 ↔ Membrane_inf' M1 M2.
Proof.
  introv C1 C2. unfolds Membrane_inf', Membrane_compare_rewritings. iff I.
   splits.
    rewrite Forall_iff_forall_mem. intros MIM E1. destruct MIM as [h hs|h|ke kes|ke]; autos~.
     rewrite Exists_iff_exists_mem. rewrite~ Mread_mem in E1. destruct (Mread M2 h) eqn: E2.
      forwards~ I': Membrane_inf_read_defined E1 E2. exists (h ---> l). splits~.
       rewrite~ Mread_mem.
      rewrite* Membrane_inf_read_undefined in E2. rewrite E2 in E1. inverts E1.
     rewrite Exists_iff_exists_mem. rewrite~ Mread_env_mem in E1.
      destruct (Mread_env M2 ke) eqn: E2.
       forwards~ I': Membrane_inf_read_env_defined E1 E2. exists (ke ~~~> l). splits~.
        rewrite~ Mread_env_mem.
       rewrite* Membrane_inf_read_env_undefined in E2. rewrite E2 in E1. inverts E1.
    apply~ Membrane_inf_new.
    apply~ Membrane_inf_new_env.
    rewrite Forall_iff_forall_mem. intros MIM E2. destruct MIM as [h hs|h|ke kes|ke]; autos~.
     rewrite Exists_iff_exists_mem. rewrite~ Mread_mem in E2. destruct (Mread M1 h) eqn: E1.
      exists (h ---> l). splits~.
       rewrite~ Mread_mem.
       compute. autos~.
      rewrite* <- Membrane_inf_read_undefined in E1. rewrite E2 in E1. inverts E1.
     rewrite Exists_iff_exists_mem. rewrite~ Mread_env_mem in E2.
      destruct (Mread_env M1 ke) eqn: E1.
       exists (ke ~~~> l). splits~.
        rewrite~ Mread_env_mem.
        compute. autos~.
       rewrite* <- Membrane_inf_read_env_undefined in E1. rewrite E2 in E1. inverts E1.
   lets (F&D&DE&F'): (rm I). rewrite Forall_iff_forall_mem in F, F'. constructors.
    (** News **)
    apply~ D.
    (** News Env **)
    apply~ DE.
    (** Read Undefined **)
    introv. destruct (Mread M1 h) as [hs|] eqn: E1.
     rewrite~ <- Mread_mem in E1. forwards Ex: F E1. rewrite Exists_iff_exists_mem in Ex.
      lets (IM&MIM&DIM): (rm Ex). destruct IM as [h' hs'|h'|ke' kes'|ke']; tryfalse.
      lets (?&I): (rm DIM). substs. rewrite~ Mread_mem in MIM. rewrite MIM. iff A; inverts A.
     iff~ _. destruct (Mread M2 h) as [hs'|] eqn: E2; autos~.
      rewrite~ <- Mread_mem in E2. forwards Ex: F' E2. rewrite Exists_iff_exists_mem in Ex.
      lets (IM&MIM&DIM): (rm Ex). destruct IM as [h' hs|h'|ke' kes'|ke']; tryfalse.
      lets (?&I): (rm DIM). substs. rewrite~ Mread_mem in MIM. rewrite MIM in E1. inverts E1.
    (** Read Undefined Env **)
    introv. destruct (Mread_env M1 ke) as [kes|] eqn: E1.
     rewrite~ <- Mread_env_mem in E1. forwards Ex: F E1. rewrite Exists_iff_exists_mem in Ex.
      lets (IM&MIM&DIM): (rm Ex). destruct IM as [h' hs'|h'|ke' kes'|ke']; tryfalse.
      lets (?&I): (rm DIM). substs. rewrite~ Mread_env_mem in MIM. rewrite MIM.
      iff A; inverts A.
     iff~ _. destruct (Mread_env M2 ke) as [kes|] eqn: E2; autos~.
      rewrite~ <- Mread_env_mem in E2. forwards Ex: F' E2.
      rewrite Exists_iff_exists_mem in Ex. lets (IM&MIM&DIM): (rm Ex).
      destruct IM as [h' hs|h'|ke' kes'|ke']; tryfalse.
      lets (?&I): (rm DIM). substs. rewrite~ Mread_env_mem in MIM. rewrite MIM in E1.
      inverts E1.
    (** Read Defined **)
    introv E1 E2. rewrite <- Mread_mem in E1, E2; autos~. forwards Ex: F E1.
    rewrite Exists_iff_exists_mem in Ex. lets (IM&MIM&DIM): (rm Ex).
    destruct IM as [h' hs'|h'|ke' kes'|ke']; tryfalse. lets (?&I): (rm DIM). substs.
    lets (n1&N1): mem_Nth (rm MIM). lets (n2&N2): mem_Nth (rm E2).
    forwards~: Membrane_unique N1 N2. substs. forwards E: Nth_func N1 N2. inverts~ E.
    (** Read Defined Env **)
    introv E1 E2. rewrite <- Mread_env_mem in E1, E2; autos~. forwards Ex: F E1.
    rewrite Exists_iff_exists_mem in Ex. lets (IM&MIM&DIM): (rm Ex).
    destruct IM as [h' hs'|h'|ke' kes'|ke']; tryfalse. lets (?&I): (rm DIM). substs.
    lets (n1&N1): mem_Nth (rm MIM). lets (n2&N2): mem_Nth (rm E2).
    forwards~: Membrane_unique_env N1 N2. substs. forwards E: Nth_func N1 N2. inverts~ E.
Qed.

Local Instance Membrane_inf_decidable : forall M1 M2,
    Membrane_correct M1 →
    Membrane_correct M2 →
    Decidable (Membrane_inf M1 M2).
  introv C1 C2. apply~ Decidable_equiv.
   symmetry. apply~ Membrane_inf_equiv_Membrane_inf'.
   repeat apply and_decidable; try typeclass.
Defined.


(** **** [Membrane_incl] **)

(** An intermediary term for proofs and computation of [Membrane_eq]. **)
Record Membrane_incl (M1 M2 : Membrane) := {
    Membrane_incl_new : incl (Mnew M1) (Mnew M2) ;
    Membrane_incl_new_env : incl (Mnew_env M1) (Mnew_env M2) ;
    Membrane_incl_read_undefined : forall h,
      Mread M2 h = None → Mread M1 h = None ;
    Membrane_incl_read_env_undefined : forall ke,
      Mread_env M2 ke = None → Mread_env M1 ke = None ;
    Membrane_incl_read_defined : forall h hs1 hs2,
      Mread M1 h = Some hs1 →
      Mread M2 h = Some hs2 →
      incl hs1 hs2 ;
    Membrane_incl_read_env_defined : forall ke kes1 kes2,
      Mread_env M1 ke = Some kes1 →
      Mread_env M2 ke = Some kes2 →
      incl kes1 kes2
  }.

(** An alternative definition of [Membrane_incl]. **)
Definition Membrane_incl' (M1 M2 : Membrane) :=
  Membrane_compare_rewritings (@incl _) (@incl _) M1 M2
  ∧ incl (Mnew M1) (Mnew M2)
  ∧ incl (Mnew_env M1) (Mnew_env M2).

Lemma Membrane_incl_equiv_Membrane_incl' : forall M1 M2,
  Membrane_correct M1 →
  Membrane_correct M2 →
  Membrane_incl M1 M2 ↔ Membrane_incl' M1 M2.
Proof.
  introv C1 C2. unfolds Membrane_incl', Membrane_compare_rewritings. iff I.
   splits.
    rewrite Forall_iff_forall_mem. intros MIM E1. destruct MIM as [h hs|h|ke kes|ke]; autos~.
     rewrite Exists_iff_exists_mem. rewrite~ Mread_mem in E1. destruct (Mread M2 h) eqn: E2.
      forwards~ I': Membrane_incl_read_defined E1 E2. exists (h ---> l). splits~.
       rewrite~ Mread_mem.
      forwards~ E': Membrane_incl_read_undefined I E2. rewrite E' in E1. inverts E1.
     rewrite Exists_iff_exists_mem. rewrite~ Mread_env_mem in E1.
      destruct (Mread_env M2 ke) eqn: E2.
       forwards~ I': Membrane_incl_read_env_defined E1 E2. exists (ke ~~~> l). splits~.
        rewrite~ Mread_env_mem.
       forwards~ E': Membrane_incl_read_env_undefined I E2. rewrite E' in E1. inverts E1.
    apply~ Membrane_incl_new.
    apply~ Membrane_incl_new_env.
   lets (F&D&DE): (rm I). rewrite Forall_iff_forall_mem in F. constructors.
    (** News **)
    apply~ D.
    (** News Env **)
    apply~ DE.
    (** Read Undefined **)
    introv E2. destruct (Mread M1 h) as [hs|] eqn: E1; autos~.
    rewrite~ <- Mread_mem in E1. forwards Ex: F E1. rewrite Exists_iff_exists_mem in Ex.
    lets (IM&MIM&DIM): (rm Ex). destruct IM as [h' hs'|h'|ke' kes'|ke']; tryfalse.
    lets (?&I): (rm DIM). substs. rewrite~ Mread_mem in MIM. rewrite MIM in E2. inverts E2.
    (** Read Undefined **)
    introv E2. destruct (Mread_env M1 ke) as [hs|] eqn: E1; autos~.
    rewrite~ <- Mread_env_mem in E1. forwards Ex: F E1. rewrite Exists_iff_exists_mem in Ex.
    lets (IM&MIM&DIM): (rm Ex). destruct IM as [h' hs'|h'|ke' kes'|ke']; tryfalse.
    lets (?&I): (rm DIM). substs. rewrite~ Mread_env_mem in MIM. rewrite MIM in E2.
    inverts E2.
    (** Read Defined **)
    introv E1 E2. rewrite <- Mread_mem in E1, E2; autos~. forwards Ex: F E1.
    rewrite Exists_iff_exists_mem in Ex. lets (IM&MIM&DIM): (rm Ex).
    destruct IM as [h' hs'|h'|ke' kes'|ke']; tryfalse. lets (?&I): (rm DIM). substs.
    lets (n1&N1): mem_Nth (rm MIM). lets (n2&N2): mem_Nth (rm E2).
    forwards~: Membrane_unique N1 N2. substs. forwards E: Nth_func N1 N2. inverts~ E.
    (** Read Defined Env **)
    introv E1 E2. rewrite <- Mread_env_mem in E1, E2; autos~. forwards Ex: F E1.
    rewrite Exists_iff_exists_mem in Ex. lets (IM&MIM&DIM): (rm Ex).
    destruct IM as [h' hs'|h'|ke' kes'|ke']; tryfalse. lets (?&I): (rm DIM). substs.
    lets (n1&N1): mem_Nth (rm MIM). lets (n2&N2): mem_Nth (rm E2).
    forwards~: Membrane_unique_env N1 N2. substs. forwards E: Nth_func N1 N2. inverts~ E.
Qed.

Local Instance Membrane_incl_decidable : forall M1 M2,
    Membrane_correct M1 →
    Membrane_correct M2 →
    Decidable (Membrane_incl M1 M2).
  introv C1 C2. apply~ Decidable_equiv.
   symmetry. apply~ Membrane_incl_equiv_Membrane_incl'.
   repeat apply and_decidable; try typeclass.
Defined.


(** **** [Membrane_eq] **)

(** Membranes are considered up to reorderings. **)
Record Membrane_eq (M1 M2 : Membrane) := {
    Membrane_eq_new : double_incl (Mnew M1) (Mnew M2) ;
    Membrane_eq_new_env : double_incl (Mnew_env M1) (Mnew_env M2) ;
    Membrane_eq_read_undefined : forall h,
      Mread M1 h = None ↔ Mread M2 h = None ;
    Membrane_eq_read_env_undefined : forall ke,
      Mread_env M1 ke = None ↔ Mread_env M2 ke = None ;
    Membrane_eq_read_defined : forall h hs1 hs2,
      Mread M1 h = Some hs1 →
      Mread M2 h = Some hs2 →
      double_incl hs1 hs2 ;
    Membrane_eq_read_env_defined : forall ke kes1 kes2,
      Mread_env M1 ke = Some kes1 →
      Mread_env M2 ke = Some kes2 →
      double_incl kes1 kes2
  }.

Lemma Membrane_eq_refl : forall M,
  Membrane_eq M M.
Proof.
  introv. constructors*.
  (** News **)
  apply double_incl_refl.
  (** News Env **)
  apply double_incl_refl.
  (** Read Defined **)
  introv E1 E2. rewrite E1 in E2. inverts E2. apply double_incl_refl.
  (** Read Defined Env **)
  introv E1 E2. rewrite E1 in E2. inverts E2. apply double_incl_refl.
Qed.

Lemma Membrane_eq_sym : forall M1 M2,
  Membrane_eq M1 M2 →
  Membrane_eq M2 M1.
Proof.
  introv E. constructors~.
  (** News **)
  forwards N: Membrane_eq_new E. apply mem_double_incl with (H := _). introv.
  symmetry. applys~ double_incl_mem N.
  (** News Env **)
  forwards N: Membrane_eq_new_env E. apply mem_double_incl with (H := _). introv.
  symmetry. applys~ double_incl_mem N.
  (** Read Undefined **)
  introv. symmetry. applys~ Membrane_eq_read_undefined E.
  (** Read Undefined Env **)
  introv. symmetry. applys~ Membrane_eq_read_env_undefined E.
  (** Read Defined **)
  introv E1 E2. apply mem_double_incl with (H := _). introv.
  symmetry. apply~ double_incl_mem. applys~ Membrane_eq_read_defined E E2 E1.
  (** Read Defined Env **)
  introv E1 E2. apply mem_double_incl with (H := _). introv.
  symmetry. apply~ double_incl_mem. applys~ Membrane_eq_read_env_defined E E2 E1.
Qed.

Lemma Membrane_eq_trans : forall M1 M2 M3,
  Membrane_eq M1 M2 →
  Membrane_eq M2 M3 →
  Membrane_eq M1 M3.
Proof.
  introv E1 E2. constructors~.
  (** News **)
  eapply double_incl_trans; [ apply E1 | apply E2 ].
  (** News Env **)
  eapply double_incl_trans; [ apply E1 | apply E2 ].
  (** Read Undefined **)
  introv. etransitivity; [ apply E1 | apply E2 ].
  (** Read Undefined Env **)
  introv. etransitivity; [ apply E1 | apply E2 ].
  (** Read Defined **)
  introv Eq1 Eq3. destruct (Mread M2 h) eqn: Eq2.
   eapply double_incl_trans; [ apply* E1 | apply* E2 ].
   apply E1 in Eq2. rewrite Eq2 in Eq1. inverts Eq1.
  (** Read Defined Env **)
  introv Eq1 Eq3. destruct (Mread_env M2 ke) eqn: Eq2.
   eapply double_incl_trans; [ apply* E1 | apply* E2 ].
   apply E1 in Eq2. rewrite Eq2 in Eq1. inverts Eq1.
Qed.

Lemma Membrane_eq_Membrane_inf : forall M1 M2,
  Membrane_eq M1 M2 →
  Membrane_inf M1 M2.
Proof.
  introv E. constructors~.
  (** News **)
  applys~ Membrane_eq_new E.
  (** News Env **)
  applys~ Membrane_eq_new_env E.
  (** Read Undefined **)
  introv. rewrite* <- Membrane_eq_read_undefined.
  (** Read Undefined Env **)
  introv. rewrite* <- Membrane_eq_read_env_undefined.
  (** Read Defined **)
  introv E1 E2. forwards~ (?&?): Membrane_eq_read_defined E1 E2.
  (** Read Defined Env **)
  introv E1 E2. forwards~ (?&?): Membrane_eq_read_env_defined E1 E2.
Qed.

Lemma Membrane_inf_Membrane_incl : forall M1 M2,
  Membrane_inf M1 M2 →
  Membrane_incl M1 M2.
Proof.
  introv E. constructors~.
  (** News **)
  applys~ Membrane_inf_new E.
  (** News Env **)
  applys~ Membrane_inf_new_env E.
  (** Read Undefined **)
  introv N. rewrite* <- Membrane_inf_read_undefined.
  (** Read Undefined Env **)
  introv N. rewrite* <- Membrane_inf_read_env_undefined.
  (** Read Defined **)
  introv E1 E2. applys~ Membrane_inf_read_defined E E1.
  (** Read Defined Env **)
  introv E1 E2. applys~ Membrane_inf_read_env_defined E E1.
Qed.

Lemma Membrane_eq_Membrane_incl : forall M1 M2,
  Membrane_eq M1 M2 →
  Membrane_incl M1 M2.
Proof. introv E. apply~ Membrane_inf_Membrane_incl. applys~ Membrane_eq_Membrane_inf E. Qed.

Lemma Membrane_incl_Membrane_eq : forall M1 M2,
  Membrane_incl M1 M2 →
  Membrane_incl M2 M1 →
  Membrane_eq M1 M2.
Proof.
  introv I1 I2. constructors~.
  (** News **)
  splits; applys~ Membrane_incl_new.
  (** News **)
  splits; applys~ Membrane_incl_new_env.
  (** Read Undefined **)
  introv. iff I; applys~ Membrane_incl_read_undefined I.
  (** Read Undefined Env **)
  introv. iff I; applys~ Membrane_incl_read_env_undefined I.
  (** Read Defined **)
  introv E1 E2. splits.
   applys~ Membrane_incl_read_defined I1 E1 E2.
   applys~ Membrane_incl_read_defined I2 E2 E1.
  (** Read Defined Env **)
  introv E1 E2. splits.
   applys~ Membrane_incl_read_env_defined I1 E1 E2.
   applys~ Membrane_incl_read_env_defined I2 E2 E1.
Qed.

Instance Membrane_eq_decidable : forall M1 M2,
    Membrane_correct M1 →
    Membrane_correct M2 →
    Decidable (Membrane_eq M1 M2).
  introv C1 C2. applys Decidable_equiv (Membrane_incl M1 M2 ∧ Membrane_incl M2 M1).
   iff I.
    apply* Membrane_incl_Membrane_eq.
    splits; apply~ Membrane_eq_Membrane_incl. applys~ Membrane_eq_sym I.
   typeclass.
Defined.


(** *** Interfaces **)

(** Definition of the outer interface of a membrane. **)
Definition MInOuterInterface h (M : Membrane) :=
  exists hs, mem (h ---> hs) M.

Definition MEInOuterInterface ke (M : Membrane) :=
  exists kes, mem (ke ~~~> kes) M.

(** Definition of the inner interface of a membrane. **)
Inductive MInInnerInterface h (M : Membrane) : Prop :=
  | MInInnerInterface_rewriting : forall ho hs,
      mem h hs →
      mem (ho ---> hs) M →
      MInInnerInterface h M
  | MInInnerInterface_new :
      mem (ν h) M →
      MInInnerInterface h M
  .

Inductive MEInInnerInterface ke (M : Membrane) : Prop :=
  | MEInInnerInterface_rewriting : forall keo kes,
      mem ke kes →
      mem (keo ~~~> kes) M →
      MEInInnerInterface ke M
  | MEInInnerInterface_new :
      mem (ν~ ke) M →
      MEInInnerInterface ke M
  .

Global Instance MInOuterInterfaceDecidable : forall h M,
    Decidable (MInOuterInterface h M).
  (** This lemma does not need these hypotheses, and I clear them to avoid unwanted
    dependencies. **)
  clear ABasic_Gamma ABasic_gamma_bot ABasic_gamma_top L_infinite K_infinite KE_infinite.
  introv. induction M as [|[h0 hs|h0|ke0 kes0|ke0] M].
   applys Decidable_equiv false_decidable. iff I; inverts* I.
   asserts E: (Decidable (h = h0)). typeclass.
    destruct E as [() E]; fold_bool; rew_refl in E.
     substs. applys decidable_make true. fold_bool. rew_refl.
      exists hs. rewrite~ mem_cons_eq.
     applys decidable_make (@decide _ IHM). extens. rew_refl. iff (hs'&I).
      exists hs'. simpl. rew_refl. right~.
      exists hs'. unfolds in I. fold mem in I. rew_refl in I. repeat inverts I as I; autos*.
   applys decidable_make (@decide _ IHM). extens. rew_refl. iff (hs'&I).
    exists hs'. unfolds. fold mem. rew_refl. right~.
    exists hs'. unfolds in I. fold mem in I. rew_refl in I. repeat inverts I as I; autos*.
   applys decidable_make (@decide _ IHM). extens. rew_refl. iff (hs'&I).
    exists hs'. unfolds. fold mem. rew_refl. right~.
    exists hs'. unfolds in I. fold mem in I. rew_refl in I. repeat inverts I as I; autos*.
   applys decidable_make (@decide _ IHM). extens. rew_refl. iff (hs'&I).
    exists hs'. unfolds. fold mem. rew_refl. right~.
    exists hs'. unfolds in I. fold mem in I. rew_refl in I. repeat inverts I as I; autos*.
Defined.

Global Instance MEInOuterInterfaceDecidable : forall ke M,
    Decidable (MEInOuterInterface ke M).
  (** This lemma does not need these hypotheses, and I clear them to avoid unwanted
    dependencies. **)
  clear ABasic_Gamma ABasic_gamma_bot ABasic_gamma_top L_infinite K_infinite KE_infinite.
  introv. induction M as [|[h0 hs|h0|ke0 kes|ke0] M].
   applys Decidable_equiv false_decidable. iff I; inverts* I.
   applys decidable_make (@decide _ IHM). extens. rew_refl. iff (hs'&I).
    exists hs'. unfolds. fold mem. rew_refl. right~.
    exists hs'. unfolds in I. fold mem in I. rew_refl in I. repeat inverts I as I; autos*.
   applys decidable_make (@decide _ IHM). extens. rew_refl. iff (hs'&I).
    exists hs'. unfolds. fold mem. rew_refl. right~.
    exists hs'. unfolds in I. fold mem in I. rew_refl in I. repeat inverts I as I; autos*.
   asserts E: (Decidable (ke = ke0)). typeclass.
    destruct E as [() E]; fold_bool; rew_refl in E.
     substs. applys decidable_make true. fold_bool. rew_refl.
      exists kes. rewrite~ mem_cons_eq.
     applys decidable_make (@decide _ IHM). extens. rew_refl. iff (kes'&I).
      exists kes'. simpl. rew_refl. right~.
      exists kes'. unfolds in I. fold mem in I. rew_refl in I. repeat inverts I as I; autos*.
   applys decidable_make (@decide _ IHM). extens. rew_refl. iff (kes'&I).
    exists kes'. unfolds. fold mem. rew_refl. right~.
    exists kes'. unfolds in I. fold mem in I. rew_refl in I. repeat inverts I as I; autos*.
Defined.

(** Adding a rewriting from a membrane only increases its interface. **)
Lemma MInInnerInterface_monotone : forall h IM M,
  MInInnerInterface h M → MInInnerInterface h (IM :: M).
Proof.
  introv MI. inverts MI.
   apply* MInInnerInterface_rewriting. simpl. rew_refl. right*.
   apply* MInInnerInterface_new. simpl. rew_refl. right*.
Qed.

Lemma MEInInnerInterface_monotone : forall ke IM M,
  MEInInnerInterface ke M → MEInInnerInterface ke (IM :: M).
Proof.
  introv MI. inverts MI.
   apply* MEInInnerInterface_rewriting. simpl. rew_refl. right*.
   apply* MEInInnerInterface_new. simpl. rew_refl. right*.
Qed.

Lemma MInInnerInterface_concat : forall h M1 M2,
  MInInnerInterface h (M1 ++ M2) ↔ MInInnerInterface h M1 ∨ MInInnerInterface h M2.
Proof.
  iff I.
   inverts I as I1 I2.
    rewrite mem_app in I2. rew_refl in I2.
     inverts I2; [ left | right ]; apply* MInInnerInterface_rewriting.
    rewrite mem_app in I1. rew_refl in I1.
     inverts I1; [ left | right ]; apply* MInInnerInterface_new.
   inverts I as I; inverts I as I1 I2.
    applys~ MInInnerInterface_rewriting I1. rewrite mem_app. rew_refl*.
    applys~ MInInnerInterface_new. rewrite mem_app. rew_refl*.
    applys~ MInInnerInterface_rewriting I1. rewrite mem_app. rew_refl*.
    applys~ MInInnerInterface_new. rewrite mem_app. rew_refl*.
Qed.

Lemma MEInInnerInterface_concat : forall ke M1 M2,
  MEInInnerInterface ke (M1 ++ M2) ↔ MEInInnerInterface ke M1 ∨ MEInInnerInterface ke M2.
Proof.
  iff I.
   inverts I as I1 I2.
    rewrite mem_app in I2. rew_refl in I2.
     inverts I2; [ left | right ]; apply* MEInInnerInterface_rewriting.
    rewrite mem_app in I1. rew_refl in I1.
     inverts I1; [ left | right ]; apply* MEInInnerInterface_new.
   inverts I as I; inverts I as I1 I2.
    applys~ MEInInnerInterface_rewriting I1. rewrite mem_app. rew_refl*.
    applys~ MEInInnerInterface_new. rewrite mem_app. rew_refl*.
    applys~ MEInInnerInterface_rewriting I1. rewrite mem_app. rew_refl*.
    applys~ MEInInnerInterface_new. rewrite mem_app. rew_refl*.
Qed.

Global Instance MInInnerInterfaceDecidable : forall h M,
    Decidable (MInInnerInterface h M).
  (** This lemma does not need these hypotheses, and I clear them to avoid unwanted
    dependencies. **)
  clear ABasic_Gamma ABasic_gamma_bot ABasic_gamma_top L_infinite K_infinite KE_infinite.
  introv. induction M as [|[h0 hs|h0|ke0 kes|ke0] M].
   applys Decidable_equiv false_decidable. iff I; inverts* I.
   destruct (mem h hs) eqn: E; fold_bool.
    applys decidable_make true. fold_bool. rew_refl.
     applys MInInnerInterface_rewriting E. rewrite~ mem_cons_eq.
    applys decidable_make (@decide _ IHM). extens. rew_refl. iff I.
     apply~ MInInnerInterface_monotone.
     inverts I as I1 I2.
      apply* MInInnerInterface_rewriting. simpl in I2. rew_refl in *.
       repeat inverts I2 as I2; fold_bool; autos*.
      apply* MInInnerInterface_new. unfolds in I1. fold mem in I1. rew_refl in *.
       repeat inverts I1 as I1; autos*.
   asserts E: (Decidable (h = h0)). typeclass.
    destruct E as [() E]; fold_bool; rew_refl in E.
     substs. applys decidable_make true. fold_bool. rew_refl.
      applys~ MInInnerInterface_new. rewrite~ mem_cons_eq.
     applys decidable_make (@decide _ IHM). extens. rew_refl. iff I.
      apply~ MInInnerInterface_monotone.
      inverts I as I1 I2.
       apply* MInInnerInterface_rewriting. unfolds in I2. fold mem in I2. rew_refl in *.
        repeat inverts I2 as I2; fold_bool; autos*.
       apply* MInInnerInterface_new. simpl in I1. rew_refl in *.
        repeat inverts I1 as I1; autos*.
   applys decidable_make (@decide _ IHM). extens. rew_refl. iff I.
    apply~ MInInnerInterface_monotone.
    inverts I as I1 I2.
     apply* MInInnerInterface_rewriting. unfolds in I2. fold mem in I2. rew_refl in *.
      repeat inverts I2 as I2; fold_bool; autos*.
     apply* MInInnerInterface_new. unfolds in I1. fold mem in I1. rew_refl in *.
      repeat inverts I1 as I1; autos*.
   applys decidable_make (@decide _ IHM). extens. rew_refl. iff I.
    apply~ MInInnerInterface_monotone.
    inverts I as I1 I2.
     apply* MInInnerInterface_rewriting. unfolds in I2. fold mem in I2. rew_refl in *.
      repeat inverts I2 as I2; fold_bool; autos*.
     apply* MInInnerInterface_new. unfolds in I1. fold mem in I1. rew_refl in *.
      repeat inverts I1 as I1; autos*.
Defined.

Global Instance MEInInnerInterfaceDecidable : forall ke M,
    Decidable (MEInInnerInterface ke M).
  (** This lemma does not need these hypotheses, and I clear them to avoid unwanted
    dependencies. **)
  clear ABasic_Gamma ABasic_gamma_bot ABasic_gamma_top L_infinite K_infinite KE_infinite.
  introv. induction M as [|[h0 hs|h0|ke0 kes|ke0] M].
   applys Decidable_equiv false_decidable. iff I; inverts* I.
   applys decidable_make (@decide _ IHM). extens. rew_refl. iff I.
    apply~ MEInInnerInterface_monotone.
    inverts I as I1 I2.
     apply* MEInInnerInterface_rewriting. unfolds in I2. fold mem in I2. rew_refl in *.
      repeat inverts I2 as I2; fold_bool; autos*.
     apply* MEInInnerInterface_new. unfolds in I1. fold mem in I1. rew_refl in *.
      repeat inverts I1 as I1; autos*.
   applys decidable_make (@decide _ IHM). extens. rew_refl. iff I.
    apply~ MEInInnerInterface_monotone.
    inverts I as I1 I2.
     apply* MEInInnerInterface_rewriting. unfolds in I2. fold mem in I2. rew_refl in *.
      repeat inverts I2 as I2; fold_bool; autos*.
     apply* MEInInnerInterface_new. unfolds in I1. fold mem in I1. rew_refl in *.
      repeat inverts I1 as I1; autos*.
   destruct (mem ke kes) eqn: E; fold_bool.
    applys decidable_make true. fold_bool. rew_refl.
     applys MEInInnerInterface_rewriting E. rewrite~ mem_cons_eq.
    applys decidable_make (@decide _ IHM). extens. rew_refl. iff I.
     apply~ MEInInnerInterface_monotone.
     inverts I as I1 I2.
      apply* MEInInnerInterface_rewriting. simpl in I2. rew_refl in *.
       repeat inverts I2 as I2; fold_bool; autos*.
      apply* MEInInnerInterface_new. unfolds in I1. fold mem in I1. rew_refl in *.
       repeat inverts I1 as I1; autos*.
   asserts E: (Decidable (ke = ke0)). typeclass.
    destruct E as [() E]; fold_bool; rew_refl in E.
     substs. applys decidable_make true. fold_bool. rew_refl.
      applys~ MEInInnerInterface_new. rewrite~ mem_cons_eq.
     applys decidable_make (@decide _ IHM). extens. rew_refl. iff I.
      apply~ MEInInnerInterface_monotone.
      inverts I as I1 I2.
       apply* MEInInnerInterface_rewriting. unfolds in I2. fold mem in I2. rew_refl in *.
        repeat inverts I2 as I2; fold_bool; autos*.
       apply* MEInInnerInterface_new. simpl in I1. rew_refl in *.
        repeat inverts I1 as I1; autos*.
Defined.

(** Computable version of the outer interface of a membrane. **)
Definition MOuterInterface (M : Membrane) :=
  concat (map (fun MI =>
    match MI with
    | h ---> _ => [h]
    | ν _ => []
    | _ ~~~> _ => []
    | ν~ _ => []
    end) M).

Definition MEOuterInterface (M : Membrane) :=
  concat (map (fun MI =>
    match MI with
    | _ ---> _ => []
    | ν _ => []
    | ke ~~~> _ => [ke]
    | ν~ _ => []
    end) M).

(** Computable version of the inner interface of a membrane. **)
Definition MInnerInterface (M : Membrane) :=
  concat (map (fun MI =>
    match MI with
    | _ ---> hs => hs
    | ν h => [h]
    | _ ~~~> _ => []
    | ν~ _ => []
    end) M).

Definition MEInnerInterface (M : Membrane) :=
  concat (map (fun MI =>
    match MI with
    | _ ---> _ => []
    | ν _ => []
    | _ ~~~> kes => kes
    | ν~ ke => [ke]
    end) M).

Lemma Mnew_MInnerInterface : forall M,
  incl (Mnew M) (MInnerInterface M).
Proof.
  introv. apply mem_incl with (H := _). introv I. apply concat_mem in I.
  lets (hs&Ihs&Mhs): (rm I). rewrite map_mem in Ihs. lets (IM&MIM&E): (rm Ihs).
  apply concat_mem. eexists. splits.
   rewrite map_mem. eexists. splits*.
   substs. destruct~ IM; tryfalse.
Qed.

Lemma Mnew_env_MEInnerInterface : forall M,
  incl (Mnew_env M) (MEInnerInterface M).
Proof.
  introv. apply mem_incl with (H := _). introv I. apply concat_mem in I.
  lets (hs&Ihs&Mhs): (rm I). rewrite map_mem in Ihs. lets (IM&MIM&E): (rm Ihs).
  apply concat_mem. eexists. splits.
   rewrite map_mem. eexists. splits*.
   substs. destruct~ IM; tryfalse.
Qed.

Lemma MOuterInterface_correct : forall h M,
  mem h (MOuterInterface M) = decide (MInOuterInterface h M).
Proof.
  extens. rew_refl. iff I.
   apply concat_mem in I. lets (hs&I'): (rm I).
    rewrite map_mem in I'. lets ((IM&MIM&E)&I): (rm I').
    destruct IM; substs; inverts I as I. fold_bool. rew_refl in I.
    inverts* I. constructors*.
   lets (hs&I'): (rm I). apply concat_mem. exists [h]. splits.
    apply map_mem with (CA := _). eexists. splits*.
    simpl. rew_refl*.
Qed.

Lemma MEOuterInterface_correct : forall ke M,
  mem ke (MEOuterInterface M) = decide (MEInOuterInterface ke M).
Proof.
  extens. rew_refl. iff I.
   apply concat_mem in I. lets (hs&I'): (rm I).
    rewrite map_mem in I'. lets ((IM&MIM&E)&I): (rm I').
    destruct IM; substs; inverts I as I. fold_bool. rew_refl in I.
    inverts* I. constructors*.
   lets (hs&I'): (rm I). apply concat_mem. exists [ke]. splits.
    apply map_mem with (CA := _). eexists. splits*.
    simpl. rew_refl*.
Qed.

Lemma MOuterInterface_mem : forall M h hs,
   mem (h ---> hs) M →
   mem h (MOuterInterface M).
Proof.
  introv C. unfolds MOuterInterface. rewrite concat_mem. exists [h]. splits.
   rewrite map_mem. eexists. splits*.
   simpl. rew_refl*.
Qed.

Lemma MEOuterInterface_mem : forall M ke kes,
   mem (ke ~~~> kes) M →
   mem ke (MEOuterInterface M).
Proof.
  introv C. unfolds MEOuterInterface. rewrite concat_mem. exists [ke]. splits.
   rewrite map_mem. eexists. splits*.
   simpl. rew_refl*.
Qed.

Lemma MOuterInterface_mem_inv : forall M h,
   mem h (MOuterInterface M) →
   exists hs, mem (h ---> hs) M.
Proof.
  introv C. unfolds MOuterInterface. rewrite concat_mem in C.
  lets (hs&Mhs&Mh): (rm C). rewrite map_mem in Mhs. lets (IM&MIM&EIM): (rm Mhs).
  destruct IM; substs; simpl in Mh; rew_refl in Mh; inverts* Mh.
Qed.

Lemma MEOuterInterface_mem_inv : forall M ke,
   mem ke (MEOuterInterface M) →
   exists kes, mem (ke ~~~> kes) M.
Proof.
  introv C. unfolds MEOuterInterface. rewrite concat_mem in C.
  lets (hs&Mhs&Mh): (rm C). rewrite map_mem in Mhs. lets (IM&MIM&EIM): (rm Mhs).
  destruct IM; substs; simpl in Mh; rew_refl in Mh; inverts* Mh.
Qed.

Lemma MInnerInterface_correct : forall h M,
  mem h (MInnerInterface M) = decide (MInInnerInterface h M).
Proof.
  extens. rew_refl. iff I.
   apply concat_mem in I. lets (hs&I'): (rm I).
    rewrite map_mem in I'. lets ((IM&MIM&E)&I): (rm I').
    destruct IM; substs; inverts I as I; fold_bool.
     applys* MInInnerInterface_rewriting.
     applys* MInInnerInterface_new. rew_refl in I. inverts* I.
   apply concat_mem. inverts I as I1 I2.
    exists hs. rewrite map_mem. splits*.
    exists [h]. rewrite map_mem. splits*. simpl. rew_refl*.
Qed.

Lemma MEInnerInterface_correct : forall ke M,
  mem ke (MEInnerInterface M) = decide (MEInInnerInterface ke M).
Proof.
  extens. rew_refl. iff I.
   apply concat_mem in I. lets (hs&I'): (rm I).
    rewrite map_mem in I'. lets ((IM&MIM&E)&I): (rm I').
    destruct IM; substs; inverts I as I; fold_bool.
     applys* MEInInnerInterface_rewriting.
     applys* MEInInnerInterface_new. rew_refl in I. inverts* I.
   apply concat_mem. inverts I as I1 I2.
    exists kes. rewrite map_mem. splits*.
    exists [ke]. rewrite map_mem. splits*. simpl. rew_refl*.
Qed.

Lemma MInInnerInterface_Mread : forall M h h' hs,
  Membrane_correct M →
  Mread M h = Some hs →
  mem h' hs →
  MInInnerInterface h' M.
Proof. introv C E I. apply* MInInnerInterface_rewriting. rewrite* Mread_mem. Qed.

Lemma MEInInnerInterface_Mread_env : forall M ke ke' kes,
  Membrane_correct M →
  Mread_env M ke = Some kes →
  mem ke' kes →
  MEInInnerInterface ke' M.
Proof. introv C E I. apply* MEInInnerInterface_rewriting. rewrite* Mread_env_mem. Qed.

Lemma MOuterInterface_app : forall M1 M2,
  MOuterInterface (M1 ++ M2) =
    MOuterInterface M1 ++ MOuterInterface M2.
Proof. introv. unfolds MOuterInterface. rewrite map_app. rewrite~ concat_app. Qed.

Lemma MEOuterInterface_app : forall M1 M2,
  MEOuterInterface (M1 ++ M2) =
    MEOuterInterface M1 ++ MEOuterInterface M2.
Proof. introv. unfolds MEOuterInterface. rewrite map_app. rewrite~ concat_app. Qed.

Lemma MInnerInterface_app : forall M1 M2,
  MInnerInterface (M1 ++ M2) =
    MInnerInterface M1 ++ MInnerInterface M2.
Proof. introv. unfolds MInnerInterface. rewrite map_app. rewrite~ concat_app. Qed.

Lemma MEInnerInterface_app : forall M1 M2,
  MEInnerInterface (M1 ++ M2) =
    MEInnerInterface M1 ++ MEInnerInterface M2.
Proof. introv. unfolds MEInnerInterface. rewrite map_app. rewrite~ concat_app. Qed.

Lemma MInOuterInterface_monotone : forall h IM M,
  MInOuterInterface h M → MInOuterInterface h (IM :: M).
Proof. introv MI. inverts MI. eexists. simpl. rew_refl. right*. Qed.

Lemma MEInOuterInterface_monotone : forall ke IM M,
  MEInOuterInterface ke M → MEInOuterInterface ke (IM :: M).
Proof. introv MI. inverts MI. eexists. simpl. rew_refl. right*. Qed.

Lemma Mread_outer_interface : forall M h,
  MInOuterInterface h M → exists hs, Mread M h = Some hs.
Proof.
  introv. induction M as [|[h0 hs|h0|ke0 kes|ke0] M].
   introv I. repeat inverts I as I.
   simpl. introv (hs'&I'). cases_if as I; fold_bool; rew_refl in I.
    substs. eexists. autos*.
    apply IHM. unfolds in I'. fold mem in I'. rew_refl in I'.
     repeat inverts I' as I'; tryfalse*. exists~ hs'.
   simpl. introv I. apply IHM. inverts I as I. unfolds in I. fold mem in I. rew_refl in I.
    inverts I; tryfalse. eexists. autos*.
   simpl. introv I. apply IHM. inverts I as I. unfolds in I. fold mem in I. rew_refl in I.
    inverts I; tryfalse. eexists. autos*.
   simpl. introv I. apply IHM. inverts I as I. unfolds in I. fold mem in I. rew_refl in I.
    inverts I; tryfalse. eexists. autos*.
Qed.

Lemma Mread_env_outer_interface : forall M ke,
  MEInOuterInterface ke M → exists kes, Mread_env M ke = Some kes.
Proof.
  introv. induction M as [|[h0 hs|h0|ke0 kes|ke0] M].
   introv I. repeat inverts I as I.
   simpl. introv I. apply IHM. inverts I as I. unfolds in I. fold mem in I. rew_refl in I.
    inverts I; tryfalse. eexists. autos*.
   simpl. introv I. apply IHM. inverts I as I. unfolds in I. fold mem in I. rew_refl in I.
    inverts I; tryfalse. eexists. autos*.
   simpl. introv (hs'&I'). cases_if as I; fold_bool; rew_refl in I.
    substs. eexists. autos*.
    apply IHM. unfolds in I'. fold mem in I'. rew_refl in I'.
     repeat inverts I' as I'; tryfalse*. exists~ hs'.
   simpl. introv I. apply IHM. inverts I as I. unfolds in I. fold mem in I. rew_refl in I.
    inverts I; tryfalse. eexists. autos*.
Qed.

Lemma Mread_defined_inv : forall M h hs,
  Mread M h = Some hs →
  MInOuterInterface h M.
Proof.
  introv I. induction M as [|[h' hs'|h'|ke' kes'|ke'] M].
   repeat inverts I as I.
   simpls. cases_if as I'; fold_bool; rew_refl in I'.
    inverts I. substs. eexists. simpl. rew_refl. left*.
    apply MInOuterInterface_monotone. apply~ IHM.
   apply MInOuterInterface_monotone. apply~ IHM.
   apply MInOuterInterface_monotone. apply~ IHM.
   apply MInOuterInterface_monotone. apply~ IHM.
Qed.

Lemma Mread_env_defined_inv : forall M ke kes,
  Mread_env M ke = Some kes →
  MEInOuterInterface ke M.
Proof.
  introv I. induction M as [|[h' hs'|h'|ke' kes'|ke'] M].
   repeat inverts I as I.
   apply MEInOuterInterface_monotone. apply~ IHM.
   apply MEInOuterInterface_monotone. apply~ IHM.
   simpls. cases_if as I'; fold_bool; rew_refl in I'.
    inverts I. substs. eexists. simpl. rew_refl. left*.
    apply MEInOuterInterface_monotone. apply~ IHM.
   apply MEInOuterInterface_monotone. apply~ IHM.
Qed.

Lemma Mread_None : forall M h,
  Mread M h = None ↔ ~ MInOuterInterface h M.
Proof.
  introv. iff I.
   introv E. forwards (hs&Ehs): Mread_outer_interface E. rewrite Ehs in I. false*.
   destruct (Mread M h) eqn: E; autos~. false I. applys~ Mread_defined_inv E.
Qed.

Lemma Mread_env_None : forall M ke,
  Mread_env M ke = None ↔ ~ MEInOuterInterface ke M.
Proof.
  introv. iff I.
   introv E. forwards (hs&Ehs): Mread_env_outer_interface E. rewrite Ehs in I. false*.
   destruct (Mread_env M ke) eqn: E; autos~. false I. applys~ Mread_env_defined_inv E.
Qed.

Lemma Membrane_eq_outer_interface : forall M M' h,
  Membrane_eq M M' →
  MInOuterInterface h M →
  MInOuterInterface h M'.
Proof.
  introv E I. lets (hs&Ehs): Mread_outer_interface (rm I).
  destruct (Mread M' h) as [hs'|] eqn: E'.
   applys~ Mread_defined_inv E'.
   rewrite* <- Membrane_eq_read_undefined in E'. rewrite Ehs in E'. inverts E'.
Qed.

Lemma Membrane_eq_outer_interface_env : forall M M' ke,
  Membrane_eq M M' →
  MEInOuterInterface ke M →
  MEInOuterInterface ke M'.
Proof.
  introv E I. lets (hs&Ehs): Mread_env_outer_interface (rm I).
  destruct (Mread_env M' ke) as [hs'|] eqn: E'.
   applys~ Mread_env_defined_inv E'.
   rewrite* <- Membrane_eq_read_env_undefined in E'. rewrite Ehs in E'. inverts E'.
Qed.

Lemma Membrane_eq_inner_interface : forall M M' h,
  Membrane_correct M →
  Membrane_correct M' →
  Membrane_eq M M' →
  MInInnerInterface h M →
  MInInnerInterface h M'.
Proof.
  introv C C' E I. inverts I as I1 I2.
   rewrite~ Mread_mem in I2. destruct (Mread M' ho) as [hs'|] eqn: E'.
    applys~ MInInnerInterface_rewriting hs'.
     forwards~ (I&_): Membrane_eq_read_defined I2 E'. rewrite <- In_mem in *. autos~.
     rewrite* Mread_mem.
    rewrite* <- Membrane_eq_read_undefined in E'. rewrite E' in I2. inverts I2.
   rewrite <- Mnew_correct in I1. rewrite (double_incl_mem _ (Membrane_eq_new E)) in I1.
    rewrite Mnew_correct in I1. applys~ MInInnerInterface_new I1.
Qed.

Lemma Membrane_eq_inner_interface_env : forall M M' ke,
  Membrane_correct M →
  Membrane_correct M' →
  Membrane_eq M M' →
  MEInInnerInterface ke M →
  MEInInnerInterface ke M'.
Proof.
  introv C C' E I. inverts I as I1 I2.
   rewrite~ Mread_env_mem in I2. destruct (Mread_env M' keo) as [kes'|] eqn: E'.
    applys~ MEInInnerInterface_rewriting kes'.
     forwards~ (I&_): Membrane_eq_read_env_defined I2 E'. rewrite <- In_mem in *. autos~.
     rewrite* Mread_env_mem.
    rewrite* <- Membrane_eq_read_env_undefined in E'. rewrite E' in I2. inverts I2.
   rewrite <- Mnew_env_correct in I1.
    rewrite (double_incl_mem _ (Membrane_eq_new_env E)) in I1.
    rewrite Mnew_env_correct in I1. applys~ MEInInnerInterface_new I1.
Qed.


(** ** Wraping Up States Formulae **)

(** *** Definition **)

Section ExtType.

(** Formulae are extended to carry additional value.  These operations
  do not change depending on such an additional value; we just
  parametrise this section by any additional carried value. **)

Variable Ext : Type.

Hypothesis ExtPoset : PosetDec.t Ext.

Variable Ext_substs : Ext → list ALoc → ALoc → Ext.
Local Notation "ext '~~[[' hs '//' h ']]'" := (Ext_substs ext hs h)
  (at level 25, no associativity).
Variable Ext_Esubsts : Ext → list KE → KE → Ext.
Local Notation "ext '~~|[[' kes '//' ke ']]'" := (Ext_Esubsts ext kes ke)
  (at level 25, no associativity).


(** [Ext_substs_basic] and [Ext_subst_nosubsts] would make sense for
  the type [Ext] as there is no coercion from [ALoc] or [KE] to
  [Ext] in general. **)
Hypothesis Ext_substs_congruence : forall hs h ext1 ext2,
  ext1 ⊑♯ ext2 →
  ext1 ~~[[ hs // h ]] ⊑♯ ext2 ~~[[ hs // h ]].
Hypothesis Ext_Esubsts_congruence : forall kes ke ext1 ext2,
  ext1 ⊑♯ ext2 →
  ext1 ~~|[[ kes // ke ]] ⊑♯ ext2 ~~|[[ kes // ke ]].

Variable ExtNames : Ext → list ALoc.
Variable ExtENames : Ext → list KE.

Hypothesis ExtNames_substs : forall ext h hs h0,
  mem h0 (ExtNames (ext ~~[[ hs // h ]]))
    = mem h (ExtNames ext) &&& mem h0 hs
    ||| mem h0 (ExtNames ext) &&& decide (h ≠ h0).
Hypothesis ExtENames_Esubsts : forall ext ke kes ke0,
  mem ke0 (ExtENames (ext ~~|[[ kes // ke ]]))
    = mem ke (ExtENames ext) &&& mem ke0 kes
    ||| mem ke0 (ExtENames ext) &&& decide (ke ≠ ke0).
Hypothesis ExtNames_Esubsts : forall ext ke kes h,
  mem h (ExtNames (ext ~~|[[ kes // ke ]]))
  = mem h (ExtNames ext).
Hypothesis ExtENames_substs : forall ext h hs ke,
  mem ke (ExtENames (ext ~~[[ hs // h ]]))
  = mem ke (ExtENames ext).


(** Extended state formulae are a couple of formulae and some
  additionnal arguments protected by a membrane. **)
Record EFormula : Type := EFormula_make {
    EFormulaFormula : IFormula ;
    EFormulaExt : Ext ;
    EFormulaMembrane : Membrane
  }.

Record EFormula_correct (Phi : EFormula) := {
    (** The membrane follows the invariants of membranes **)
    EFormula_membrane_correct : Membrane_correct (EFormulaMembrane Phi) ;
    (** The invariant of formulae states that every inner name is caught
      by the membrane. **)
    EFormula_naming : forall h,
      mem h (IFNames (EFormulaFormula Phi)) →
      MInInnerInterface h (EFormulaMembrane Phi) ;
    EFormula_naming_env : forall ke,
      mem ke (IFENames (EFormulaFormula Phi)) →
      MEInInnerInterface ke (EFormulaMembrane Phi) ;
    (** The same for the carried value. **)
    EFormula_naming_ext : forall h,
      mem h (ExtNames (EFormulaExt Phi)) →
      MInInnerInterface h (EFormulaMembrane Phi) ;
    EFormula_naming_ext_env : forall ke,
      mem ke (ExtENames (EFormulaExt Phi)) →
      MEInInnerInterface ke (EFormulaMembrane Phi) ;
    (** Every rewrited abstract location is either not present in the
      domain of the inner formula, or each of its parts are present. **)
    EFormula_rewriting_complete : forall h hs h1 h2 f,
      Mread (EFormulaMembrane Phi) h = Some hs →
      mem h1 hs →
      mem h2 hs →
      locInDom (EFormulaFormula Phi) h1 f →
      locInDom (EFormulaFormula Phi) h2 f ;
    EFormula_rewriting_complete_env : forall ke kes ke1 ke2,
      Mread_env (EFormulaMembrane Phi) ke = Some kes →
      mem ke1 kes →
      mem ke2 kes →
      envInDom (EFormulaFormula Phi) ke1 →
      envInDom (EFormulaFormula Phi) ke2
  }.


(** A dependent type for formulae.  This will only be used outside of this file,
   where the implementation of formulae should not matter. **)
Definition DFormula := { Phi : EFormula | EFormula_correct Phi }.

Definition EFormula_with_Formula (Phi : EFormula) (phi : IFormula) :=
  EFormula_make phi (EFormulaExt Phi) (EFormulaMembrane Phi).

Definition EFormula_with_Ext (Phi : EFormula) (ext : Ext) :=
  EFormula_make (EFormulaFormula Phi) ext (EFormulaMembrane Phi).

Definition EFormula_with_Membrane (Phi : EFormula) (M : Membrane) :=
  EFormula_make (EFormulaFormula Phi) (EFormulaExt Phi) M.

Global Instance EFormula_Inhab : Inhab Ext → Inhab EFormula.
  introv I. apply prove_Inhab. constructors; exact arbitrary.
Defined.

(** To ease the proof, we request that an [ext] exists such that it
  respect every property which we need among those respected by the
  domains which we will use in practise. **)
Lemma DFormula_Inhab : forall ext,
  ExtNames ext = nil →
  ExtENames ext = nil →
  Inhab DFormula.
Proof.
  introv N1 N2. apply prove_Inhab.
  refine (exist _ (EFormula_make emp ext _) _). constructors~.
   apply Membrane_nil_correct.
   introv I. false*.
   introv I. false*.
   introv I. simpl in I. rewrite N1 in I. false*.
   introv I. simpl in I. rewrite N2 in I. false*.
   introv I. false*.
   introv I. false*.
Defined.


(** *** Domain **)

(** The interface of an extended formula is the set of abstract
  locations accessible from the outside of the formula: as formulae
  are considered modulo alpha-conversion, the inner identifiers means
  nothing in the outside. **)
Definition EFormulaInterface Phi :=
  MOuterInterface (EFormulaMembrane Phi).

Definition EFormulaEInterface Phi :=
  MEOuterInterface (EFormulaMembrane Phi).

(** The domain of an extended formula is the set of abstract locations
  accessible from the outside of the formula which represent left-hand
  sides of subformulae in the formula.  The concretisations of the
  domain of an extended formula is exactly the domain of the formula’s
  concretisations.  As a special case, if [k ---> nil] is present in
  the membrane, both interpretations (whether [k] is in the domain of
  the formula) would make sense.  However, as [k ---> nil] is the same
  (with respect to concretisations) as [k ---> k' | k' --> { _ : bot
  }], it is natural to consider [k] such that [k ---> nil] as part of
  the domain.  This allows to later make the simplification
  “incoherence found for [k]” into “[k ---> nil] in the membrane”. **)
Record EFormulaInDomain Phi h := {
    EFormulaInDomain_InInterface : mem h (EFormulaInterface Phi) ;
    EFormulaInDomain_LeftHandSides : forall hs h',
      Mread (EFormulaMembrane Phi) h = Some hs →
      mem h' hs →
      exists f, locInDom (EFormulaFormula Phi) h' f
  }.

Record EFormulaInEDomain Phi ke := {
    EFormulaInEDomain_InInterface : mem ke (EFormulaEInterface Phi) ;
    EFormulaInEDomain_LeftHandSides : forall kes ke',
      Mread_env (EFormulaMembrane Phi) ke = Some kes →
      mem ke' kes →
      envInDom (EFormulaFormula Phi) ke'
  }.

Local Instance LookForALocLeftHandSide : forall phi h,
    Decidable (exists f, locInDom phi h f).
  introv. induction phi.
   applys decidable_make false. fold_bool. rew_refl. introv (f&I). applys locInDom_emp I.
   destruct IHphi1 as [d1 D1]. destruct IHphi2 as [d2 D2]. applys decidable_make (d1 || d2).
    clear - D1 D2. destruct d1; destruct d2; rew_bool; fold_bool; rew_refl in *;
      try match goal with D : exists f, _ |- _ =>
        lets (f&D'): D; exists f; apply* locInDom_star_inv end.
    introv (f&I). apply locInDom_star in I. inverts* I.
   applys decidable_make (decide (a = h ∧ DObject_dom d ≠ AObject_domain_empty)).
    extens. rew_refl. iff I.
     lets (E&D): (rm I). lets (f&I): AObject_dom_non_empty D. exists f.
      substs. eexists. splits*. constructors~.
     lets (f&o&S&ID): (rm I). repeat inverts S as S. splits~. introv E. rewrite E in ID.
      simpl in ID. rew_refl~ in ID.
   applys decidable_make false. fold_bool. rew_refl. introv (f&o&S&D). repeat inverts S as S.
Defined.

Global Instance EFormulaInDomain_decidable : forall Phi h,
    EFormula_correct Phi →
    Decidable (EFormulaInDomain Phi h).
  introv C. destruct (Mread (EFormulaMembrane Phi) h) as [hs|] eqn: ER.
   forwards IO: Mread_defined_inv ER. rewrite <- istrue_decide in IO.
    rewrite <- MOuterInterface_correct in IO. destruct hs as [|h' hs].
     applys decidable_make true. fold_bool. rew_refl. constructors~.
      introv ER' M'. rewrite ER in ER'. inverts ER'. false*.
     applys decidable_make (decide (exists f, locInDom (EFormulaFormula Phi) h' f)).
      extens. rew_refl. iff I.
       constructors~. introv ER' M'. rewrite ER in ER'. inverts ER'. lets (f&II): (rm I).
        exists f. applys~ EFormula_rewriting_complete ER II. simpl. rew_refl*.
       applys~ EFormulaInDomain_LeftHandSides I ER. simpl. rew_refl*.
   applys decidable_make false. fold_bool. rew_refl. introv (II&ID).
    rewrite Mread_None in ER. unfolds EFormulaInterface.
    rewrite MOuterInterface_correct in II. rew_refl* in II.
Defined.

Global Instance EFormulaInEDomain_decidable : forall Phi ke,
    EFormula_correct Phi →
    Decidable (EFormulaInEDomain Phi ke).
  introv C. destruct (Mread_env (EFormulaMembrane Phi) ke) as [kes|] eqn: ER.
   forwards IO: Mread_env_defined_inv ER. rewrite <- istrue_decide in IO.
    rewrite <- MEOuterInterface_correct in IO. destruct kes as [|ke' kes].
     applys decidable_make true. fold_bool. rew_refl. constructors~.
      introv ER' M'. rewrite ER in ER'. inverts ER'. false*.
     applys decidable_make (decide (envInDom (EFormulaFormula Phi) ke')).
      rew_refl. fequals. extens. iff I.
       constructors~. introv ER' M'. rewrite ER in ER'. inverts ER'.
        applys~ EFormula_rewriting_complete_env ER I. simpl. rew_refl*.
       lets (II&ID): (rm I). apply* ID. simpl. rew_refl*.
   applys decidable_make false. fold_bool. rew_refl. introv (II&ID).
    rewrite Mread_env_None in ER. unfolds EFormulaEInterface.
    rewrite MEOuterInterface_correct in II. rew_refl* in II.
Defined.

Definition DFormulaDomain (Phi : DFormula) :=
  let (Phi, C) := Phi in
  filter (fun h =>
    let _ := EFormulaInDomain_decidable h C in
    decide (EFormulaInDomain Phi h)) (EFormulaInterface Phi).

Definition DFormulaEDomain (Phi : DFormula) :=
  let (Phi, C) := Phi in
  filter (fun h =>
    let _ := EFormulaInEDomain_decidable h C in
    decide (EFormulaInEDomain Phi h)) (EFormulaEInterface Phi).

Lemma DFormulaDomain_correct : forall Phi h,
  mem h (DFormulaDomain Phi) ↔ EFormulaInDomain (proj1_sig Phi) h.
Proof.
  introv. destruct Phi as (Phi&C). unfolds DFormulaDomain. simpl.
  rewrite filter_mem_eq. rew_refl. iff* I. splits~. inverts~ I.
Qed.

Lemma DFormulaEDomain_correct : forall Phi ke,
  mem ke (DFormulaEDomain Phi) ↔ EFormulaInEDomain (proj1_sig Phi) ke.
Proof.
  introv. destruct Phi as (Phi&C). unfolds DFormulaEDomain. simpl.
  rewrite filter_mem_eq. rew_refl. iff* I. splits~. inverts~ I.
Qed.

Definition locDomain : forall phi h,
    { d | forall f, in_AObject_domain f d ↔ locInDom phi h f }.
  introv. destruct (IFormula_domain_simple phi) as [hds kes] eqn: E.
  destruct (mem_assoc h hds) eqn: E'; fold_bool.
   exists (assoc h hds). introv. rewrite* IFormula_domain_simple_locInDom. iff I.
    exists (assoc h hds). splits~. apply~ assoc_mem_assoc.
    lets (d&I1&I2): (rm I). erewrite* IFormula_domain_simple_mem_assoc.
   rew_refl in E'. exists AObject_domain_empty. introv.
    rewrite* IFormula_domain_simple_locInDom. iff I; inverts I as (I1&I2).
    false E'. apply* mem_mem_assoc.
Defined.

Lemma EFormula_correct_AObject_domain_compare : forall Phi h hs h1 h2,
  EFormula_correct Phi →
  Mread (EFormulaMembrane Phi) h = Some hs →
  mem h1 hs →
  mem h2 hs →
  AObject_domain_compare (proj1_sig (locDomain (EFormulaFormula Phi) h1))
    (proj1_sig (locDomain (EFormulaFormula Phi) h2)).
Proof.
  introv C E I1 I2. introv. extens.
  rewrite (proj2_sig (locDomain (EFormulaFormula Phi) h1) f).
  rewrite (proj2_sig (locDomain (EFormulaFormula Phi) h2) f).
  iff I; applys* EFormula_rewriting_complete I.
Qed.


(** *** Structure **)

(** We equip formulae with a basic equivalence relation.  Note that it
  will later be extended to an order relation compatible (that is,
  functionnal) with this equivalence relation, but not
  antisymmetrical, as the order relation will allow much more
  operations to be performed on the formulae than this relation.  This
  equivalence relation only take into account the basic equalities on
  the lower levels, as well as renamings. **)

Inductive EFormula_eq_step : binary EFormula :=

  | EFormula_eq_step_star_closure : forall phi phi' ext M,
      IFormula_star_closure phi phi' →
      EFormula_eq_step (EFormula_make phi ext M) (EFormula_make phi' ext M)

  | EFormula_eq_step_membrane_eq : forall phi ext M M',
      Membrane_eq M M' →
      EFormula_eq_step (EFormula_make phi ext M) (EFormula_make phi ext M')

  | EFormula_eq_step_ext_eq : forall phi ext ext' M,
      ext =♯ ext' →
      EFormula_eq_step (EFormula_make phi ext M) (EFormula_make phi ext' M)
  .

Definition EFormula_eq := rstclosure EFormula_eq_step.

Global Instance EFormula_eq_decidable : forall Phi1 Phi2,
    EFormula_correct Phi1 →
    EFormula_correct Phi2 →
    Decidable (EFormula_eq Phi1 Phi2).
  introv C1 C2. destruct Phi1 as [phi1 ext1 M1]. destruct Phi2 as [phi2 ext2 M2].
  applys Decidable_equiv (IFormula_star_closure phi1 phi2 ∧ Membrane_eq M1 M2 ∧ ext1 =♯ ext2).
   clear C1 C2. iff E.
    lets (Ephi&EM&Eext): (rm E).
     eapply rstclosure_trans.
      apply rstclosure_step. apply* EFormula_eq_step_star_closure.
      eapply rstclosure_trans.
       apply rstclosure_step. apply* EFormula_eq_step_membrane_eq.
       apply rstclosure_step. apply* EFormula_eq_step_ext_eq.
    sets_eq Phi1: (EFormula_make phi1 ext1 M1). sets_eq Phi2: (EFormula_make phi2 ext2 M2).
     gen phi1 phi2 ext1 ext2 M1 M2. induction E; introv EQPhi1 EQPhi2; substs.
      inverts H as E; splits~; try apply Membrane_eq_refl; try apply IFormula_star_refl.
      inverts EQPhi2. splits~.
       apply IFormula_star_refl.
       apply Membrane_eq_refl.
      forwards~ (Ephi&EM&Eext): IHE. splits~.
       apply~ IFormula_star_sym.
       apply~ Membrane_eq_sym.
      destruct y as [phi M ext]. forwards~ (Ephi2&EM2&Eext2): IHE2.
       forwards~ (Ephi1&EM1&Eext1): IHE1. splits.
        apply* IFormula_star_trans.
        apply* Membrane_eq_trans.
        autos*.
   forwards MC1: EFormula_membrane_correct C1. forwards MC2: EFormula_membrane_correct C2.
    typeclass.
Defined.

Definition DFormula_eq (Phi1 Phi2 : DFormula) :=
  EFormula_eq (proj1_sig Phi1) (proj1_sig Phi2).

Global Instance DFormula_eq_decidable : forall Phi1 Phi2,
    Decidable (DFormula_eq Phi1 Phi2).
  introv. destruct Phi1 as [Phi1 OKPhi1]. destruct Phi2 as [Phi2 OKPhi2]. typeclass.
Defined.

End ExtType.


(** *** Formulae Without Extensions **)

Definition SFormula : Type := EFormula unit.

Definition SFormula_make phi M : SFormula :=
  EFormula_make phi tt M.

Definition SFormula_correct (Phi : SFormula) :=
  EFormula_correct (LibFunc.const nil) (LibFunc.const nil) Phi.

Lemma EFormula_sub_correct : forall Ext ExtNames ExtENames (Phi : EFormula Ext),
  EFormula_correct ExtNames ExtENames Phi →
  SFormula_correct (SFormula_make (EFormulaFormula Phi) (EFormulaMembrane Phi)).
Proof.
  introv C. destruct Phi as [phi0 ext M]. simpls. constructors; try apply~ C; intros; false~.
Qed.

End General.

