
(** This file contains the structure to manipulate separation-logic
  formulae used in the abstract semantics of the example language. **)

Set Implicit Arguments.

Require Export MembranesSep.

Section General.

(** [L] represents singleton locations, [K] set locations and [KE]
  environments, which behave like set locations. **)
Variable L K KE : Type. (** Typically, set of integers **)
Hypothesis L_comparable : Comparable L.
Hypothesis K_comparable : Comparable K.
Hypothesis KE_comparable : Comparable KE.

(** These sets have to be infinite. **)
Hypothesis L_infinite : forall ls,
  exists (l : L), ~ mem l ls.
Hypothesis K_infinite : forall ks,
  exists (k : K), ~ mem k ks.
Hypothesis KE_infinite : forall ks,
  exists (k : KE), ~ mem k ks.
Hypothesis L_infinite_pick : forall ls,
  Pickable (fun l : L => ~ mem l ls).
Hypothesis K_infinite_pick : forall ks,
  Pickable (fun k : K => ~ mem k ks).
Hypothesis KE_infinite_pick : forall ks,
  Pickable (fun k : KE => ~ mem k ks).

Variable ABasic : Type. (** Abstract basic values **)
Hypothesis ABasic_comparable : Comparable ABasic.
Hypothesis ABasicLat : LatticeDec.t ABasic.
Hypothesis ABasicTop : TopDec.t ABasic.
Variable ABasic_gamma : ABasic → ℘ basic_val.
Hypothesis ABasic_Gamma : Gamma.t _ _ ABasic_gamma.
Hypothesis ABasic_gamma_bot : forall b,
  ~ ABasic_gamma (⊥♯) b.
Hypothesis ABasic_gamma_top : forall b,
  ABasic_gamma (⊤♯) b.

Let ALoc := ALoc L K.
Let d_singleton_loc := singleton_loc _ : L → ALoc.
Let d_set_loc := set_loc _ (K := K) : K → ALoc.
Coercion d_singleton_loc : L >-> ALoc.
Coercion d_set_loc : K >-> ALoc.

Let ALoc_Equiv : EquivDec.t ALoc :=
  ALoc_Equiv _ _.

Let ALocS := ALocS L K.

Local Instance Local_Inhab_ALoc : Inhab ALoc.
  apply~ ALoc_inhab.
Defined.

Local Instance Local_Comparable_ALoc : Comparable ALoc.
  apply~ ALoc_comparable.
Defined.

Let AClosure := AClosure KE.
Let AValue := AValue L K KE ABasic.
Let ASValue := ASValue L K KE ABasic.
Let DObject := DObject L K KE ABasic.
Let AEnv := AEnv L K KE ABasic.

Variable Ext : Type.
Variable Ext_substs : Ext → list ALoc → ALoc → Ext.
Variable Ext_Esubsts : Ext → list KE → KE → Ext.
Variable ExtNames : Ext → list ALoc.
Variable ExtENames : Ext → list KE.

Let IFormula := IFormula L K KE ABasic.
Let EFormula := EFormula L K KE ABasic Ext.
Let IMembrane := IMembrane L K KE.
Let Membrane := Membrane L K KE.
Let DFormula := DFormula _ _ _ ABasic ExtNames ExtENames.

Notation "'emp'" := (IFormula_emp _ _ _ _) : sep_scope.
Notation "phi1 '*' phi2" := (IFormula_star phi1 phi2) : sep_scope.
Notation "h '-->' o" := (IFormula_loc h o)
  (at level 39, no associativity) : sep_scope.
Notation "ke '~~>' E" := (IFormula_env ke E)
  (at level 39, no associativity) : sep_scope.

Notation "h '--->' hs" := (Membrane_rewriting _ h hs)
  (at level 56) : sep_scope.
Notation "'ν' h" := (Membrane_new _ h)
  (at level 56) : sep_scope.
Notation "ke '~~~>' kes" := (Membrane_Erewriting _ _ ke kes)
  (at level 56) : sep_scope.
Notation "'ν~' ke" := (Membrane_Enew _ _ ke)
  (at level 56) : sep_scope.

Open Scope sep_scope.

Instance IFormula_Inhab : Inhab IFormula.
  applys prove_Inhab (emp : IFormula).
Defined.

Let Mread := Mread _ _ : Membrane → ALoc → option (list ALoc).
Let Mread_env := Mread_env _ : Membrane → KE → option (list KE).

Let MInnerInterface_correct :=
  MInnerInterface_correct _ _ _ : forall (h : ALoc) (M : Membrane), _.
Let MInInnerInterfaceDecidable :=
  MInInnerInterfaceDecidable _ _ _ : forall (h : ALoc) (M : Membrane), _.
Let MEInnerInterface_correct :=
  MEInnerInterface_correct _ _ _ : forall (ke : KE) (M : Membrane), _.
Let MEInInnerInterfaceDecidable :=
  MEInInnerInterfaceDecidable _ _ _ : forall (ke : KE) (M : Membrane), _.

Let EFormula_correct :=
  EFormula_correct _ _ _ ExtNames ExtENames : EFormula → Prop.

Instance Local_AClosure_LatDec : LatticeDec.t AClosure.
  apply~ AClosure_LatDec.
Defined.

Local Instance Local_AEnv_Inhab : Inhab AEnv.
  applys prove_Inhab. constructors.
Defined.

Local Instance Local_DObject_Inhab : Inhab DObject.
  apply Local_DObject_Inhab.
Defined.

Let read_DObject : DObject → property → option ASValue :=
  read_DObject _ _ _ _.
Let read_AEnv : AEnv → var → option AValue :=
  read_AEnv _ _ _ _.

Local Instance Local_locInDom_decidable : forall (phi : IFormula) (h : ALoc) f,
    Decidable (locInDom phi h f).
  apply locInDom_decidable; typeclass.
Defined.

Local Instance Local_envInDom_decidable : forall (phi : IFormula) ke,
    Decidable (envInDom phi ke).
  apply envInDom_decidable; typeclass.
Defined.

Let passThroughMembrane : IFormula → Membrane → option IFormula :=
  passThroughMembrane _ _ _ _ _.

Let MembraneCompose : Membrane → Membrane → Membrane :=
  MembraneCompose _ _ _.
Let MembraneCheckCompose : Membrane → Membrane → option Membrane :=
  MembraneCheckCompose _ _ _.


(** *** Renaming Formulae **)

(** We would like to consider formulae up to renamings.  This section
  presents some operations for this end. **)

(** An auxiliary function for [getRenamingMembrane].  This function
  has been defined in the general case [H] (meaning either [L] or [K])
  coercible to [ALoc] through [ALocH]. **)
Definition getRenamingMembrane_aux (H : Type) `{Comparable H}
    (infinite_pick : forall hs : list H, Pickable (fun h => ~ mem h hs))
    ALocH init names_to_change : list H * Membrane * list H :=
  fold_left (fun h hsMa =>
      let '(avoid, M, already_seen) := hsMa in
      if mem h already_seen then
        (avoid, M, already_seen)
      else
        let h' := @pick _ _ (infinite_pick avoid) in
        (h' :: avoid, ALocH h ---> [ALocH h'] :: M, h :: already_seen))
    init names_to_change.

Definition getRenamingMembrane_env_aux init names_to_change : list KE * Membrane * list KE :=
  fold_left (fun ke kesMa =>
      let '(avoid, M, already_seen) := kesMa in
      if mem ke already_seen then
        (avoid, M, already_seen)
      else
        let ke' := @pick _ _ (KE_infinite_pick avoid) in
        (ke' :: avoid, ke ~~~> [ke'] :: M, ke :: already_seen))
    init names_to_change.

(** From a list [avoid] of identifiers to avoid, creates a membrane
  renaming the identifiers [names_to_change]: an application of this
  membrane to the problematic (extended) formula returns an (extended)
  formula equivalent to the initial one, but without the problematic
  identifiers. **)
Definition getRenamingMembrane H `{Comparable H} infinite_pick (ALocH : H → _) avoid names_to_change :=
  let '(_, M, _) :=
    getRenamingMembrane_aux infinite_pick ALocH (avoid, nil, nil) names_to_change in M.

Definition getRenamingMembrane_env avoid names_to_change :=
  let '(_, M, _) :=
    getRenamingMembrane_env_aux (avoid, nil, nil) names_to_change in M.

(** A renaming membrane only has constructions of the type [h ---> h']
  where [h] and [h'] have the same type, as well as no merging. **)
Record renamingMembraneProp (M : Membrane) := {
    renamingMembraneProp_kind :
      Forall (fun IM =>
        match IM with
        | h ---> [h'] => ALoc_same_type h h'
        | _ => False
        end) M ;
    renamingMembraneProp_noDuplicates :
      forall n1 n2 h1 h2 h1' h2',
        Nth n1 M (h1 ---> [h1']) →
        Nth n2 M (h2 ---> [h2']) →
        h1' = h2' →
        n1 = n2
  }.

Record renamingMembraneProp_env (M : Membrane) := {
    renamingMembraneProp_env_kind :
      Forall (fun IM =>
        match IM with
        | ke ~~~> [ke'] => True
        | _ => False
        end) M ;
    renamingMembraneProp_env_noDuplicates :
      forall n1 n2 ke1 ke2 ke1' ke2',
        Nth n1 M (ke1 ~~~> [ke1']) →
        Nth n2 M (ke2 ~~~> [ke2']) →
        ke1' = ke2' →
        n1 = n2
  }.

Lemma renamingMembraneProp_noNewMembrane : forall M,
  renamingMembraneProp M →
  noNewMembrane M.
Proof.
  introv RP. apply renamingMembraneProp_kind in RP.
  unfolds. rewrite Forall_iff_forall_mem in *.
  introv I. forwards: RP I. destruct~ x.
Qed.

Lemma renamingMembraneProp_env_noNewMembrane : forall M,
  renamingMembraneProp_env M →
  noNewMembrane M.
Proof.
  introv RP. apply renamingMembraneProp_env_kind in RP.
  unfolds. rewrite Forall_iff_forall_mem in *.
  introv I. forwards: RP I. destruct~ x.
Qed.

Lemma renamingMembraneProp_noMergingMembrane : forall M,
  Membrane_correct _ _ _ M →
  renamingMembraneProp M →
  noMergingMembrane _ _ _ M.
Proof.
  (** This lemma does not need these hypotheses, and I clear them to avoid unwanted
    dependencies. **)
  clear ABasic_Gamma ABasic_gamma_bot ABasic_gamma_top.
  introv C RP. constructors; introv.
   asserts (h'&F): (exists h', forall h'', mem h'' (MgetDependencies _ _ M h) → h'' = h').
     tests T: (exists h' hs, mem (h' ---> hs) M ∧ mem h hs).
      lets (h'&hs&Mh'&Mhs): (rm T). exists h'.
       introv Mh''. forwards (hs'&Rhs'&Mhs'): MgetDependencies_correct Mh''; try typeclass.
       rewrite <- Mread_mem in Rhs'; autos*. forwards F: renamingMembraneProp_kind RP.
       rewrite Forall_iff_forall_mem in F. forwards E1: F Mh'. forwards E2': F Rhs'.
       do 2 (destruct hs as [|? hs]; tryfalse). do 2 (destruct hs' as [|? hs']; tryfalse).
       lets (n1&N1): mem_Nth (rm Mh'). lets (n2&N2): mem_Nth (rm Rhs').
       simpl in Mhs. rew_refl in Mhs. inverts Mhs; tryfalse.
       simpl in Mhs'. rew_refl in Mhs'. inverts Mhs'; tryfalse.
       forwards~: renamingMembraneProp_noDuplicates N1 N2. substs.
       forwards Eq: Nth_func N1 N2. inverts~ Eq.
      exists h. introv Mh''. false T.
       forwards~ (hs&Rhs&Mhs): MgetDependencies_correct Mh'';
         try solve [ rewrite <- Mread_mem in Rhs; autos* ]; typeclass.
    forwards ND: MgetDependencies_No_Duplicates L_comparable K_comparable M h.
    destruct MgetDependencies as [|h0 [|h1 l]]; try solve [ compute; math ].
    inverts ND as NM. false NM. asserts: (h0 = h1).
     transitivity h'; [| symmetry ]; apply F; simpl; rew_refl*.
    substs. constructors~.
   asserts Eq: (MgetDependencies_env _ M ke = nil).
     apply nil_mem with (CA := _). intros ke' Mke'.
     forwards (kes&Rkes&Mkes): MgetDependencies_env_correct Mke'; try typeclass.
     rewrite <- Mread_env_mem in Rkes; autos*. forwards F: renamingMembraneProp_kind RP.
     rewrite Forall_iff_forall_mem in F. forwards* E: F Rkes.
    rewrite Eq. rewrite length_nil. math.
Qed.

Lemma renamingMembraneProp_env_noMergingMembrane : forall M,
  Membrane_correct _ _ _ M →
  renamingMembraneProp_env M →
  noMergingMembrane _ _ _ M.
Proof.
  (** This lemma does not need these hypotheses, and I clear them to avoid unwanted
    dependencies. **)
  clear ABasic_Gamma ABasic_gamma_bot ABasic_gamma_top.
  introv C RP. constructors; introv.
   asserts Eq: (MgetDependencies _ _ M h = nil).
     apply nil_mem with (CA := _). intros h' Mh'.
     forwards (hs&Rhs&Mhs): MgetDependencies_correct Mh'; try typeclass.
     rewrite <- Mread_mem in Rhs; autos*. forwards F: renamingMembraneProp_env_kind RP.
     rewrite Forall_iff_forall_mem in F. forwards* E: F Rhs.
    rewrite Eq. rewrite length_nil. math.
   asserts (ke'&F): (exists ke', forall ke'', mem ke'' (MgetDependencies_env _ M ke) → ke'' = ke').
     tests T: (exists ke' kes, mem (ke' ~~~> kes) M ∧ mem ke kes).
      lets (ke'&kes&Mke'&Mkes): (rm T). exists ke'.
       introv Mke''. forwards (kes'&Rkes'&Mkes'): MgetDependencies_env_correct Mke''; try typeclass.
       rewrite <- Mread_env_mem in Rkes'; autos*. forwards F: renamingMembraneProp_env_kind RP.
       rewrite Forall_iff_forall_mem in F. forwards E1: F Mke'. forwards E2': F Rkes'.
       do 2 (destruct kes as [|? kes]; tryfalse). do 2 (destruct kes' as [|? kes']; tryfalse).
       lets (n1&N1): mem_Nth (rm Mke'). lets (n2&N2): mem_Nth (rm Rkes').
       simpl in Mkes. rew_refl in Mkes. inverts Mkes; tryfalse.
       simpl in Mkes'. rew_refl in Mkes'. inverts Mkes'; tryfalse.
       forwards~: renamingMembraneProp_env_noDuplicates N1 N2. substs.
       forwards Eq: Nth_func N1 N2. inverts~ Eq.
      exists ke. introv Mke''. false T.
       forwards~ (kes&Rkes&Mkes): MgetDependencies_env_correct Mke'';
         try solve [ rewrite <- Mread_env_mem in Rkes; autos* ]; typeclass.
    forwards ND: MgetDependencies_env_No_Duplicates KE_comparable M ke.
    destruct MgetDependencies_env as [|ke0 [|ke1 l]]; try solve [ compute; math ].
    inverts ND as NM. false NM. asserts: (ke0 = ke1).
     transitivity ke'; [| symmetry ]; apply F; simpl; rew_refl*.
    substs. constructors~.
Qed.

Lemma getRenamingMembrane_correct : forall H C infinite_pick ALocH avoid names_to_change,
  (forall hs : list H, exists h, ~ mem h hs) →
  (forall h h', ALoc_same_type (ALocH h) (ALocH h')) →
  (forall h1 h2, ALocH h1 = ALocH h2 → h1 = h2) →
  renamingMembraneProp (@getRenamingMembrane H C infinite_pick ALocH avoid names_to_change).
Proof.
  introv infinite ST Inv. unfolds getRenamingMembrane, getRenamingMembrane_aux.
  fold IMembrane. sets_eq p: (avoid, @nil IMembrane, @nil H).
  asserts OKp: (renamingMembraneProp (snd (fst p))).
   substs. constructors.
    constructors.
    introv N1. inverts N1.
  asserts Invp: (forall h0 h hs,
    mem (ALocH h) hs → mem (h0 ---> hs) (snd (fst p)) → mem h (fst (fst p))).
   introv M1 M2. substs. false*.
  clear EQp. gen p. induction names_to_change; introv OKp Invp; destruct~ p as [pM ?].
  simpl. destruct pM as [? M]. apply IHnames_to_change.
   cases_if as I; fold_bool; rew_refl~ in I. constructors.
    apply~ Forall_cons. applys~ renamingMembraneProp_kind OKp.
    simpl. introv N1 N2 E. forwards [(M1'&E1&?)|(IM1&M1'&m1&E1&N1'&?)]: Nth_cons_inv (rm N1);
      forwards [(M2'&E2&?)|(IM2&M2'&m2&E2&N2'&?)]: Nth_cons_inv (rm N2); substs~;
      inverts E1; inverts E2.
     simpls. apply Nth_mem with (CA := _) in N2'. forwards A: Invp N2'.
      simpl. rew_refl*.
      forwards A': pick_spec infinite. false* A'.
     simpls. apply Nth_mem with (CA := _) in N1'. forwards A: Invp N1'.
      simpl. rew_refl*.
      forwards A': pick_spec infinite. false* A'.
     fequals. applys~ renamingMembraneProp_noDuplicates OKp N1' N2'.
   introv I1 I2. cases_if as I; fold_bool; rew_refl~ in I; simpls.
    apply* Invp.
    rew_refl in I2. rew_refl. inverts I2 as I2.
     left. lets (?&?): (rm I2). substs. simpl in I1. rew_refl in I1.
      inverts I1 as E; tryfalse. forwards: Inv E. substs~.
     right. apply* Invp.
Qed.

Lemma getRenamingMembrane_env_correct : forall avoid names_to_change,
  renamingMembraneProp_env (@getRenamingMembrane_env avoid names_to_change).
Proof.
  introv. unfolds getRenamingMembrane_env, getRenamingMembrane_env_aux.
  fold IMembrane. sets_eq p: (avoid, @nil IMembrane, @nil KE).
  asserts OKp: (renamingMembraneProp_env (snd (fst p))).
   substs. constructors.
    constructors.
    introv N1. inverts N1.
  asserts Invp: (forall ke0 ke kes,
    mem ke kes → mem (ke0 ~~~> kes) (snd (fst p)) → mem ke (fst (fst p))).
   introv M1 M2. substs. false*.
  clear EQp. gen p. induction names_to_change; introv OKp Invp; destruct~ p as [pM ?].
  simpl. destruct pM as [? M]. apply IHnames_to_change.
   cases_if as I; fold_bool; rew_refl~ in I. constructors.
    apply~ Forall_cons. applys~ renamingMembraneProp_env_kind OKp.
    simpl. introv N1 N2 E. forwards [(M1'&E1&?)|(IM1&M1'&m1&E1&N1'&?)]: Nth_cons_inv (rm N1);
      forwards [(M2'&E2&?)|(IM2&M2'&m2&E2&N2'&?)]: Nth_cons_inv (rm N2); substs~;
      inverts E1; inverts E2.
     simpls. apply Nth_mem with (CA := _) in N2'. forwards A: Invp N2'.
      simpl. rew_refl*.
      forwards A': pick_spec KE_infinite. false* A'.
     simpls. apply Nth_mem with (CA := _) in N1'. forwards A: Invp N1'.
      simpl. rew_refl*.
      forwards A': pick_spec KE_infinite. false* A'.
     fequals. applys~ renamingMembraneProp_env_noDuplicates OKp N1' N2'.
   introv I1 I2. cases_if as I; fold_bool; rew_refl~ in I; simpls.
    apply* Invp.
    rew_refl in I2. rew_refl. inverts I2 as I2.
     left. lets (?&?): (rm I2). substs. simpl in I1. rew_refl in I1.
      inverts I1; tryfalse. substs~.
     right. apply* Invp.
Qed.

Lemma getRenamingMembrane_incl_membrane_new : forall H `{Comparable H}
    infinite_pick ALocH avoid (hs : list H) M M0 alr alr0 names h hsh,
  (forall h1 h2, ALocH h1 = ALocH h2 → h1 = h2) →
  getRenamingMembrane_aux infinite_pick ALocH (avoid, M0, alr0) names = (hs, M, alr) →
  mem h alr0 → mem (ALocH h ---> hsh) M → mem (ALocH h ---> hsh) M0.
Proof.
  introv Inv E. gen avoid M0 alr0 hs M alr h hsh. induction names; introv E.
   inverts~ E.
   unfolds in E. simpls. cases_if as I; fold_bool; rew_refl in I.
    applys~ IHnames E.
    lets~ New: IHnames E.
     introv I1 I2. forwards~ I': New I2.
      simpl. rew_refl~.
      unfolds in I'. fold mem in I'. rew_refl in I'. inverts I' as I'; autos~.
       inverts I' as I'. apply Inv in I'. substs. false~ I.
Qed.

Lemma getRenamingMembrane_env_incl_membrane_new : forall avoid kes M M0 alr alr0 names ke kes',
  getRenamingMembrane_env_aux (avoid, M0, alr0) names = (kes, M, alr) →
  mem ke alr0 → mem (ke ~~~> kes') M → mem (ke ~~~> kes') M0.
Proof.
  introv E. gen avoid M0 alr0 kes M alr ke kes'. induction names; introv E.
   inverts~ E.
   unfolds in E. simpls. cases_if as I; fold_bool; rew_refl in I.
    applys~ IHnames E.
    lets~ New: IHnames E.
     introv I1 I2. forwards~ I': New I2.
      simpl. rew_refl~.
      unfolds in I'. fold mem in I'. rew_refl in I'. inverts I' as I'; autos~.
       inverts I' as I'. false* I.
Qed.

Lemma getRenamingMembrane_incl_already_gen : forall H `{Comparable H}
    infinite_pick ALocH avoid names_to (hs : list H) M M0 alr alr0 names,
  getRenamingMembrane_aux infinite_pick ALocH (avoid, M0, alr0) names = (hs, M, alr) →
  (forall x, mem x names_to = mem x alr0 ||| mem x names) →
  incl alr names_to ∧ incl names_to alr.
Proof.
  (** This lemma does not need these hypotheses, and I clear them to avoid unwanted
    dependencies. **)
  clear ABasic_Gamma ABasic_gamma_bot ABasic_gamma_top L_infinite K_infinite KE_infinite L_infinite_pick K_infinite_pick KE_infinite_pick Ext_substs Ext_Esubsts ExtNames ExtENames DFormula EFormula_correct.
  introv E Inv. unfolds getRenamingMembrane_aux.
  gen hs M M0 alr alr0 avoid. induction names; introv Inv E.
   simpls. inverts E. splits; apply mem_incl with (H := _); introv I.
    rewrite Inv. rew_refl*.
    rewrite Inv in I. rew_refl in I. inverts~ I. false*.
   simpl in E. cases_if as I; fold_bool.
    forwards~: IHnames E. clear E IHnames. simpls. introv. rewrite Inv. extens. rew_refl.
     iff Hyp; repeat inverts Hyp as Hyp; autos~.
    rew_refl in I. forwards~: IHnames E. clear E IHnames. simpls. introv. rewrite Inv.
     extens. rew_refl*.
Qed.

Lemma getRenamingMembrane_env_incl_already_gen : forall avoid names_to
    kes M M0 alr alr0 names,
  getRenamingMembrane_env_aux (avoid, M0, alr0) names = (kes, M, alr) →
  (forall x, mem x names_to = mem x alr0 ||| mem x names) →
  incl alr names_to ∧ incl names_to alr.
Proof.
  (** This lemma does not need these hypotheses, and I clear them to avoid unwanted
    dependencies. **)
  clear ABasic_Gamma ABasic_gamma_bot ABasic_gamma_top L_infinite K_infinite KE_infinite L_infinite_pick K_infinite_pick Ext_substs Ext_Esubsts ExtNames ExtENames DFormula EFormula_correct.
  introv E Inv. unfolds getRenamingMembrane_env_aux.
  gen kes M M0 alr alr0 avoid. induction names; introv Inv E.
   simpls. inverts E. splits; apply mem_incl with (H := _); introv I.
    rewrite~ Inv. rew_refl*.
    rewrite~ Inv in I. rew_refl in I. inverts~ I. false*.
   simpl in E. cases_if as I; fold_bool.
    forwards~: IHnames E. clear E IHnames. simpls. introv. rewrite Inv. extens. rew_refl.
     iff Hyp; repeat inverts Hyp as Hyp; autos~.
    rew_refl in I. forwards~: IHnames E. clear E IHnames. simpls. introv. rewrite Inv.
     extens. rew_refl*.
Qed.

Lemma getRenamingMembrane_incl_already : forall H `{Comparable H}
    infinite_pick ALocH avoid names_to_change (hs : list H) M alr,
  getRenamingMembrane_aux infinite_pick ALocH (avoid, nil, nil) names_to_change
    = (hs, M, alr) →
  incl alr names_to_change ∧ incl names_to_change alr.
Proof.
  introv E. applys~ getRenamingMembrane_incl_already_gen E.
  introv. clear. simpl. destruct (mem x _); autos~.
Qed.

Lemma getRenamingMembrane_env_incl_already : forall avoid names_to_change kes M alr,
  getRenamingMembrane_env_aux (avoid, nil, nil) names_to_change = (kes, M, alr) →
  incl alr names_to_change ∧ incl names_to_change alr.
Proof.
  introv E. applys~ getRenamingMembrane_env_incl_already_gen E.
  introv. clear. simpl. destruct (mem x _); autos~.
Qed.

Lemma getRenamingMembrane_incl_membrane_incl : forall H `{Comparable H}
    infinite_pick ALocH avoid (hs : list H) M M0 alr alr0 names,
  getRenamingMembrane_aux infinite_pick ALocH (avoid, M0, alr0) names = (hs, M, alr) →
  exists M', M = M' ++ M0.
Proof.
  introv E. gen avoid M0 alr0 hs M alr. induction names; introv E.
   inverts E. exists~ (@nil IMembrane).
   unfolds in E. simpls. cases_if as I; fold_bool; rew_refl in I.
    applys~ IHnames E.
    forwards~ (M'&EM): IHnames E. substs. eexists. rewrite~ app_last.
Qed.

Lemma getRenamingMembrane_env_incl_membrane_incl : forall avoid
    kes M M0 alr alr0 names,
  getRenamingMembrane_env_aux (avoid, M0, alr0) names = (kes, M, alr) →
  exists M', M = M' ++ M0.
Proof.
  introv E. gen avoid M0 alr0 kes M alr. induction names; introv E.
   inverts E. exists~ (@nil IMembrane).
   unfolds in E. simpls. cases_if as I; fold_bool; rew_refl in I.
    applys~ IHnames E.
    forwards~ (M'&EM): IHnames E. substs. eexists. rewrite~ app_last.
Qed.

Lemma getRenamingMembrane_cancel_app : forall H `{Comparable H}
    infinite_pick ALocH avoid (hs : list H) M M0 M1 alr alr0 names,
  getRenamingMembrane_aux infinite_pick ALocH (avoid, M0 ++ M1, alr0) names
    = (hs, M ++ M1, alr) →
  getRenamingMembrane_aux infinite_pick ALocH (avoid, M0, alr0) names = (hs, M, alr).
Proof.
  introv E. gen avoid M0 alr0 hs M alr. induction names; introv E.
   inverts E as E. apply app_cancel_r in E. substs~.
   unfolds getRenamingMembrane_aux. simpl. simpl in E.
    cases_if as I; apply~ IHnames.
Qed.

Lemma getRenamingMembrane_env_cancel_app : forall avoid kes M M0 M1 alr alr0 names,
  getRenamingMembrane_env_aux (avoid, M0 ++ M1, alr0) names = (kes, M ++ M1, alr) →
  getRenamingMembrane_env_aux (avoid, M0, alr0) names = (kes, M, alr).
Proof.
  introv E. gen avoid M0 alr0 kes M alr. induction names; introv E.
   inverts E as E. apply app_cancel_r in E. substs~.
   unfolds getRenamingMembrane_env_aux. simpl. simpl in E.
    cases_if as I; apply~ IHnames.
Qed.

Lemma getRenamingMembrane_exists_rew : forall H `{Comparable H}
    infinite_pick ALocH avoid M0 alr0 names (hs : list H) M alr h hs',
  getRenamingMembrane_aux infinite_pick ALocH (avoid, M0, alr0) names = (hs, M, alr) →
  ~ Exists (fun IM => exists hs, IM = h ---> hs) M0 →
  mem (h ---> hs') M →
  exists h', h = ALocH h' ∧ mem h' names.
Proof.
  introv E Np I. unfolds getRenamingMembrane_aux.
  gen M0 avoid alr0 hs M alr. induction names; introv Np I Efold; simpl in Efold.
    inverts Efold. false Np. rewrite* Exists_iff_exists_mem.
    cases_if as II; fold_bool; rew_refl in II.
     forwards* (h'&Eh&Mh): IHnames. exists h'. splits~. simpl. rew_refl*.
     tests Eq: (ALocH a = h).
      exists a. splits~. simpl. rew_refl*.
     forwards* (h'&Eh&Mh): IHnames Efold.
      introv A. apply Np. clear Np. simpls.
       repeat rewrite Exists_iff_exists_mem with (CA := _) in *.
       lets (IM&MIM&Ex): (rm A). exists IM. splits~. simpl in MIM. rew_refl in MIM.
       inverts~ MIM. lets (hs''&E): (rm Ex). inverts E. false~ Eq.
      eexists. splits*. simpl. rew_refl*.
Qed.

Lemma getRenamingMembrane_env_exists_rew : forall avoid M0 alr0 names
    kes M alr ke kes',
  getRenamingMembrane_env_aux (avoid, M0, alr0) names = (kes, M, alr) →
  ~ Exists (fun IM => exists kes, IM = ke ~~~> kes) M0 →
  mem (ke ~~~> kes') M →
  mem ke names.
Proof.
  introv E Np I. unfolds getRenamingMembrane_env_aux.
  gen M0 avoid alr0 kes M alr. induction names; introv Np I Efold; simpl in Efold.
    inverts Efold. false Np. rewrite* Exists_iff_exists_mem.
    cases_if as II; fold_bool; rew_refl in II.
     forwards* Mke: IHnames. simpl. rew_refl*.
     tests Eq: (ke = a).
      simpl. rew_refl*.
     forwards* Mke: IHnames Efold.
      introv A. apply Np. clear Np. simpls.
       repeat rewrite Exists_iff_exists_mem with (CA := _) in *.
       lets (IM&MIM&Ex): (rm A). exists IM. splits~. simpl in MIM. rew_refl in MIM.
       inverts~ MIM. lets (hs&E): (rm Ex). inverts E. false~ Eq.
      simpl. rew_refl*.
Qed.

Lemma getRenamingMembrane_exists_avoid : forall H `{Comparable H}
    infinite_pick ALocH avoid M0 alr0 names (hs : list H) M alr h,
  (forall hs : list H, exists h, ~ mem h hs) →
  (forall h1 h2, ALocH h1 = ALocH h2 → h1 = h2) →
  getRenamingMembrane_aux infinite_pick ALocH (avoid, M0, alr0) names = (hs, M, alr) →
  ~ MInInnerInterface _ _ _ h M0 →
  MInInnerInterface _ _ _ h M →
  exists h', h = ALocH h' ∧ ~ mem h' avoid.
Proof.
  introv H_infinite InvertALocH E Np I. unfolds getRenamingMembrane_aux.
  gen M0 avoid alr0 hs M alr. induction names; introv Np I Efold; simpl in Efold.
    inverts Efold. false~ Np.
    cases_if as II; fold_bool; rew_refl in II.
     apply* IHnames.
     tests Eq: (h = ALocH (pick (λ h : H, ¬ mem h avoid))).
      eexists. splits*. introv Ma. forwards* (h'&Eh'&NMh'): (rm IHnames) (rm Efold).
       introv I'. inverts I' as I1 I2.
        simpl in I2. rew_refl in I2. inverts I2 as I2.
         inverts I2. simpl in I1. rew_refl in I1. inverts I1 as E; tryfalse.
          forwards: InvertALocH (rm E). substs.
          applys~ pick_spec H_infinite Ma.
         false Np. apply* MInInnerInterface_rewriting.
        unfolds in I1. fold mem in I1. rew_refl in I1. inverts I1 as I1; tryfalse.
         false Np. constructors*.
       forwards: InvertALocH Eh'. substs. false NMh'. simpl. rew_refl*.
      forwards* (h'&Eh'&NMh'): (rm IHnames) Efold.
       introv Ex. inverts Ex as Ex Ex'.
         simpl in Ex'. rew_refl in Ex'. inverts Ex' as Ex'.
          inverts Ex'. simpl in Ex. rew_refl in Ex. repeat inverts Ex as Ex; tryfalse*.
          false Np. apply* MInInnerInterface_rewriting.
         unfolds in Ex. fold mem in Ex. rew_refl in Ex. inverts Ex as Ex; tryfalse.
          false Np. constructors*.
       substs. eexists. splits*. introv Mh'. apply NMh'. simpl. rew_refl*.
Qed.

Lemma getRenamingMembrane_env_exists_avoid : forall avoid M0 alr0 names kes M alr ke,
  getRenamingMembrane_env_aux (avoid, M0, alr0) names = (kes, M, alr) →
  ~ MEInInnerInterface _ _ _ ke M0 →
  MEInInnerInterface _ _ _ ke M →
  ~ mem ke avoid.
Proof.
  introv E Np I Ike. unfolds getRenamingMembrane_env_aux.
  gen M0 avoid alr0 kes M alr. induction names; introv Np Ike I Efold; simpl in Efold.
    inverts Efold. false~ Np.
    cases_if as II; fold_bool; rew_refl in II.
     apply* IHnames.
     tests Eq: (ke = pick (λ ke, ¬ mem ke avoid)).
      forwards*: pick_spec Ike.
      forwards* A: (rm IHnames) (rm Efold).
       introv Ex. inverts Ex as Ex Ex'.
         simpl in Ex'. rew_refl in Ex'. inverts Ex' as Ex'.
          inverts Ex'. simpl in Ex. rew_refl in Ex. repeat inverts Ex as Ex; tryfalse*.
          false Np. apply* MEInInnerInterface_rewriting.
         simpl in Ex. rew_refl in Ex. inverts Ex as Ex; tryfalse. false Np. constructors*.
          cases_if~.
       simpl. rew_refl*.
Qed.

Lemma getRenamingMembrane_MembraneCorrect : forall H `{Comparable H}
    infinite_pick ALocH avoid names_to_change,
  (forall hs : list H, exists h, ~ mem h hs) →
  (forall h1 h2, ALocH h1 = ALocH h2 → h1 = h2) →
  (forall h1 h2, ALoc_same_type (ALocH h1) (ALocH h2)) →
  No_duplicates names_to_change →
  Membrane_correct _ _ _ (@getRenamingMembrane H _ infinite_pick ALocH avoid names_to_change).
Proof.
  (** This lemma does not need these hypotheses, and I clear them to avoid unwanted
    dependencies. **)
  clear ABasic_Gamma ABasic_gamma_bot ABasic_gamma_top L_infinite K_infinite KE_infinite L_infinite_pick K_infinite_pick KE_infinite_pick Ext_substs Ext_Esubsts ExtNames ExtENames DFormula EFormula_correct.
  introv infinite InvertALocH ALocHSameType ND. constructors.
  (** Uniqueness **)
  introv I1 I2. sets_eq Fo: (getRenamingMembrane infinite_pick ALocH avoid names_to_change).
  unfolds getRenamingMembrane.
  destruct (getRenamingMembrane_aux _ _ _ _) as [pM alr] eqn: Efold. destruct pM as [hs M].
  subst Fo. forwards~ I1': Nth_mem I1.
  forwards (hf&Eh&Ih): getRenamingMembrane_exists_rew Efold.
   introv A. inverts A.
   applys~ Nth_mem I1.
  forwards (names_prev&names&Enames&F): mem_split Ih. subst names_to_change.
  unfolds getRenamingMembrane_aux. rewrite fold_left_app in Efold. simpl in Efold.
  match type of Efold with context [ fold_left ?f ?i names_prev ] =>
    sets_eq <- F0: (fold_left f i names_prev) end. destruct F0 as ((hs0&M0)&alr0).
  cases_if as I; fold_bool.
   forwards (Incl1&Incl2): getRenamingMembrane_incl_already EQF0.
    false F. applys* incl_mem Incl1.
   rew_refl in I. forwards (M'&EM'): getRenamingMembrane_incl_membrane_incl Efold. substs.
    asserts NI: (Forall (λ x : H, ALocH x ≠ ALocH hf) names).
      rewrite Forall_iff_forall_mem. introv Ix Ex. forwards~: InvertALocH Ex. substs.
      forwards (_&ND'&_): No_duplicates_inv_app ND. inverts ND' as ND1 ND2.
      false ND1. erewrite* Mem_mem.
    asserts NM0: (forall hs, ~ mem (ALocH hf ---> hs) M0).
      introv Ihs3. forwards (hf'&E&Mhs): getRenamingMembrane_exists_rew EQF0 Ihs3.
       introv Ex. inverts Ex.
       forwards: InvertALocH E. substs. false* F.
    asserts NM': (forall hs, ~ mem (ALocH hf ---> hs) M').
      introv Ihs3. rewrite <- app_nil_l with
        (l := ALocH hf ---> [ALocH (pick (λ h : H, ¬mem h hs0))] :: M0) in Efold.
      forwards E': getRenamingMembrane_cancel_app Efold.
      forwards~ Ex: getRenamingMembrane_incl_membrane_new E' Ihs3.
       simpl. rew_refl. left~.
    forwards [N1|(m1&Em1&N1)]: Nth_app_inv I1.
     false NM'. apply* Nth_mem.
     forwards [(M1&EM1&Eq)|(x&M1&m&EM1&Nm&Em)]: Nth_cons_inv N1.
      inverts EM1. forwards [N2|(m2&Em2&N2)]: Nth_app_inv I2.
       false NM'. apply* Nth_mem.
       forwards [(M2&EM2&Eq')|(x'&M2&m&EM2&Nm'&Em')]: Nth_cons_inv N2.
        inverts EM2. substs~.
        inverts EM2. false NM0. apply* Nth_mem.
      inverts EM1. false NM0. apply* Nth_mem.
  (** Uniqueness env **)
  introv I1 I2. sets_eq Fo: (getRenamingMembrane infinite_pick ALocH avoid names_to_change).
  forwards~ RP: getRenamingMembrane_correct; try rewrite <- EQFo in RP.
  apply renamingMembraneProp_kind in RP. rewrite Forall_iff_forall_mem in RP.
  apply Nth_mem with (CA := _) in I1. forwards*: RP I1.
  (** Rewriting uniqueness **)
  introv I. sets_eq Fo: (getRenamingMembrane infinite_pick ALocH avoid names_to_change).
  unfolds getRenamingMembrane.
  destruct (getRenamingMembrane_aux _ _ _ _) as [pM alr] eqn: Efold. destruct pM as [hs' M'].
  subst Fo. unfolds getRenamingMembrane_aux. rewrite fold_left_eq_fold_right in Efold.
  gen hs' M' alr. induction (rev names_to_change); introv I E.
   inverts E. false*.
   simpl in E. match type of E with context [fold_right ?f ?i ?l ] =>
     destruct (fold_right f i l) as ((hs0&M0)&alr0) eqn: E' end. cases_if~; inverts* E.
    simpl in I. rew_refl in I. inverts I as I.
     inverts I. constructors~. introv A. inverts A.
     apply* IHl.
  (** Rewriting uniqueness env **)
  introv I. sets_eq Fo: (getRenamingMembrane infinite_pick ALocH avoid names_to_change).
  forwards~ RP: getRenamingMembrane_correct; try rewrite <- EQFo in RP.
  apply renamingMembraneProp_kind in RP. rewrite Forall_iff_forall_mem in RP. forwards*: RP I.
Qed.

Lemma getRenamingMembrane_env_MembraneCorrect : forall avoid names_to_change,
  No_duplicates names_to_change →
  Membrane_correct _ _ _ (getRenamingMembrane_env avoid names_to_change).
Proof.
  (** This lemma does not need these hypotheses, and I clear them to avoid unwanted
    dependencies. **)
  clear ABasic_Gamma ABasic_gamma_bot ABasic_gamma_top L_infinite K_infinite L_infinite_pick K_infinite_pick Ext_substs Ext_Esubsts ExtNames ExtENames DFormula EFormula_correct.
  introv ND. constructors.
  (** Uniqueness **)
  introv I1 I2. forwards~ RP: getRenamingMembrane_env_correct.
  apply renamingMembraneProp_env_kind in RP. rewrite Forall_iff_forall_mem in RP.
  apply Nth_mem with (CA := _) in I1. forwards*: RP I1.
  (** Uniqueness env **)
  introv I1 I2. sets_eq Fo: (getRenamingMembrane_env avoid names_to_change).
  unfolds getRenamingMembrane_env.
  destruct (getRenamingMembrane_env_aux _ _) as [pM alr] eqn: Efold. destruct pM as [hs M].
  subst Fo. forwards~ I1': Nth_mem I1. forwards I: getRenamingMembrane_env_exists_rew Efold.
   introv A. inverts A.
   applys~ Nth_mem I1.
  forwards (names_prev&names&Enames&F): mem_split I. subst names_to_change.
  unfolds getRenamingMembrane_env_aux. rewrite fold_left_app in Efold. simpl in Efold.
  match type of Efold with context [ fold_left ?f ?i names_prev ] =>
    sets_eq <- F0: (fold_left f i names_prev) end. destruct F0 as ((hs0&M0)&alr0).
  cases_if as I'; fold_bool.
   forwards (Incl1&Incl2): getRenamingMembrane_env_incl_already EQF0.
    false F. applys* incl_mem Incl1.
   rew_refl in I. forwards (M'&EM'): getRenamingMembrane_env_incl_membrane_incl Efold.
    substs. asserts NM0: (forall kes, ~ mem (ke ~~~> kes) M0).
      introv Ihs3. forwards I'': getRenamingMembrane_env_exists_rew EQF0 Ihs3.
       introv Ex. inverts Ex.
       false* F.
    asserts NM': (forall kes, ~ mem (ke ~~~> kes) M').
      introv Ihs3. rewrite <- app_nil_l with
        (l := ke ~~~> [pick (λ ke, ~ mem ke hs0)] :: M0) in Efold.
      forwards E': getRenamingMembrane_env_cancel_app Efold.
      forwards~ Ex: getRenamingMembrane_env_incl_membrane_new E' Ihs3.
       simpl. rew_refl. left~.
    forwards [N1|(m1&Em1&N1)]: Nth_app_inv I1.
     false NM'. apply* Nth_mem.
     forwards [(M1&EM1&Eq)|(x&M1&m&EM1&Nm&Em)]: Nth_cons_inv N1.
      inverts EM1. forwards [N2|(m2&Em2&N2)]: Nth_app_inv I2.
       false NM'. apply* Nth_mem.
       forwards [(M2&EM2&Eq')|(x'&M2&m&EM2&Nm'&Em')]: Nth_cons_inv N2.
        inverts EM2. substs~.
        inverts EM2. false NM0. apply* Nth_mem.
      inverts EM1. false NM0. apply* Nth_mem.
  (** Rewriting uniqueness **)
  introv I. sets_eq Fo: (getRenamingMembrane_env avoid names_to_change).
  forwards~ RP: getRenamingMembrane_env_correct; try rewrite <- EQFo in RP.
  apply renamingMembraneProp_env_kind in RP. rewrite Forall_iff_forall_mem in RP.
  forwards*: RP I.
  (** Rewriting uniqueness env **)
  introv I. sets_eq Fo: (getRenamingMembrane_env avoid names_to_change).
  unfolds getRenamingMembrane_env.
  destruct (getRenamingMembrane_env_aux _ _) as [pM alr] eqn: Efold. destruct pM as [hs' M'].
  subst Fo. unfolds getRenamingMembrane_env_aux. rewrite fold_left_eq_fold_right in Efold.
  gen hs' M' alr. induction (rev names_to_change); introv I E.
   inverts E. false*.
   simpl in E. match type of E with context [fold_right ?f ?i ?l ] =>
     destruct (fold_right f i l) as ((hs0&M0)&alr0) eqn: E' end. cases_if~; inverts* E.
    simpl in I. rew_refl in I. inverts I as I.
     inverts I. constructors~. introv A. inverts A.
     apply* IHl.
Qed.


Lemma getRenamingMembrane_noMergingMembrane : forall H `{Comparable H}
    infinite_pick (ALocH : H → ALoc) avoid names,
  (forall hs : list H, exists h, ~ mem h hs) →
  (forall h h', ALoc_same_type (ALocH h) (ALocH h')) →
  (forall h1 h2, ALocH h1 = ALocH h2 → h1 = h2) →
  No_duplicates names →
  noMergingMembrane _ _ _ (getRenamingMembrane infinite_pick ALocH avoid names).
Proof.
  (** This lemma does not need these hypotheses, and I clear them to avoid unwanted
    dependencies. **)
  clear ABasic_Gamma ABasic_gamma_bot ABasic_gamma_top.
  introv infinite ST Inv ND. apply~ renamingMembraneProp_noMergingMembrane.
   apply~ getRenamingMembrane_MembraneCorrect.
   apply~ getRenamingMembrane_correct.
Qed.

Lemma getRenamingMembrane_env_noMergingMembrane : forall avoid names,
  No_duplicates names →
  noMergingMembrane _ _ _ (getRenamingMembrane_env avoid names).
Proof.
  (** This lemma does not need these hypotheses, and I clear them to avoid unwanted
    dependencies. **)
  clear ABasic_Gamma ABasic_gamma_bot ABasic_gamma_top.
  introv ND. apply~ renamingMembraneProp_env_noMergingMembrane.
   apply~ getRenamingMembrane_env_MembraneCorrect.
   apply~ getRenamingMembrane_env_correct.
Qed.

Lemma getRenamingMembrane_check_outer : forall H `{Comparable H}
    infinite_pick ALocH avoid names,
  Forall (fun h : ALoc => exists h', h = ALocH h' ∧ mem h' names)
    (MOuterInterface (@getRenamingMembrane H _ infinite_pick ALocH avoid names)).
Proof.
  introv. unfolds getRenamingMembrane. rewrite Forall_iff_forall_mem. introv I.
  destruct (getRenamingMembrane_aux _ _ _ _) as [pM alr] eqn: E. destruct pM as [p M].
  unfolds MOuterInterface. unfolds ALoc. rewrite concat_mem in I. lets (hs&Mhs&Mh): (rm I).
  rewrite map_mem in Mhs. lets (IM&MIM&EIM): (rm Mhs). substs. destruct IM; tryfalse.
  forwards~ (h'&Eh'&Mh'): getRenamingMembrane_exists_rew E MIM.
   introv A. inverts A.
   simpl in Mh. rew_refl in Mh. inverts* Mh.
Qed.

Lemma getRenamingMembrane_env_check_outer : forall avoid names,
  incl (MEOuterInterface (getRenamingMembrane_env avoid names)) names.
Proof.
  introv. apply mem_incl with (H := _). introv I. unfolds getRenamingMembrane_env.
  destruct (getRenamingMembrane_env_aux _ _) as [pM alr] eqn: E. destruct pM as [p M].
  unfolds MEOuterInterface. rewrite concat_mem in I. lets (hs&Mhs&Mh): (rm I).
  rewrite map_mem in Mhs. lets (IM&MIM&EIM): (rm Mhs). substs. destruct IM; tryfalse.
  simpl in Mh. rew_refl in Mh. inverts* Mh.
  applys~ getRenamingMembrane_env_exists_rew E MIM.
  introv A. inverts A.
Qed.

Lemma getRenamingMembrane_check_inner : forall H `{Comparable H}
    infinite_pick ALocH avoid names,
  (forall hs : list H, exists h, ~ mem h hs) →
  (forall h1 h2, ALocH h1 = ALocH h2 → h1 = h2) →
  (forall h h', ALoc_same_type (ALocH h) (ALocH h')) →
  Forall (fun h : ALoc => exists h', h = ALocH h' ∧ ~ mem h' avoid)
    (MInnerInterface (@getRenamingMembrane H _ infinite_pick ALocH avoid names)).
Proof.
  introv H_infinite InvertALocH ST. unfolds getRenamingMembrane.
  rewrite Forall_iff_forall_mem. introv I.
  destruct (getRenamingMembrane_aux _ _ _ _) as [pM alr] eqn: E. destruct pM as [p M].
  unfolds MInnerInterface. unfolds ALoc. rewrite concat_mem in I. lets (hs&Mhs&Mh): (rm I).
  rewrite map_mem in Mhs. lets (IM&MIM&EIM): (rm Mhs). substs.
  destruct IM as [h hs|h|ke kes|ke]; tryfalse.
   applys* getRenamingMembrane_exists_avoid E.
    introv A. inverts A; false*.
    applys~ MInInnerInterface_rewriting MIM. apply* Mh.
   simpl in Mh. rew_refl in Mh. inverts Mh; tryfalse.
    forwards~ F: getRenamingMembrane_correct ST. unfolds getRenamingMembrane.
    rewrite E in F. apply renamingMembraneProp_kind in F.
    rewrite Forall_iff_forall_mem in F. forwards*: F MIM.
Qed.

Lemma getRenamingMembrane_env_check_Einner : forall avoid names,
  Forall (fun ke => ~ mem ke avoid)
    (MEInnerInterface (@getRenamingMembrane_env avoid names)).
Proof.
  introv. unfolds getRenamingMembrane_env. rewrite Forall_iff_forall_mem. introv I Ia.
  destruct (getRenamingMembrane_env_aux _ _) as [pM alr] eqn: E. destruct pM as [p M].
  unfolds MEInnerInterface. rewrite concat_mem in I. lets (hs&Mhs&Mh): (rm I).
  rewrite map_mem in Mhs. lets (IM&MIM&EIM): (rm Mhs). substs.
  destruct IM as [h hs|h|ke kes|ke]; tryfalse.
   applys* getRenamingMembrane_env_exists_avoid E.
    introv A. inverts A; false*.
    applys~ MEInInnerInterface_rewriting MIM. apply* Mh.
   simpl in Mh. rew_refl in Mh. inverts Mh; tryfalse.
    forwards* F: getRenamingMembrane_env_correct. unfolds getRenamingMembrane_env.
    rewrite E in F. apply renamingMembraneProp_env_kind in F.
    rewrite Forall_iff_forall_mem in F. forwards*: F MIM.
Qed.

Lemma getRenamingMembrane_aux_all : forall H `{Comparable H}
    infinite_pick ALocH avoid M0 alr0 names (hs : list H) M alr h,
  getRenamingMembrane_aux infinite_pick ALocH (avoid, M0, alr0) names = (hs, M, alr) →
  ~ mem h alr0 →
  mem h names →
  MInOuterInterface _ _ _ (ALocH h) M.
Proof.
  introv Nalr0 E Mh. unfolds getRenamingMembrane_aux.
  gen avoid M0 alr0 hs M alr. induction names; simpl; introv Nalr0 E.
   false*.
   cases_if as II; fold_bool; rew_refl in II.
    apply* IHnames. simpl in Mh. rew_refl in Mh. inverts* Mh.
    tests Eq: (h = a).
     forwards~ (M'&EM'): getRenamingMembrane_incl_membrane_incl E. substs.
      constructors~. rewrite mem_app. simpl. rew_refl*.
     simpl in Mh. rew_refl in Mh. inverts Mh; tryfalse~.
     applys~ IHnames E. simpl. rew_refl. introv [A|A]; tryfalse~.
Qed.

Lemma getRenamingMembrane_all : forall H `{Comparable H}
    infinite_pick ALocH avoid names (h : H),
  mem h names →
  mem (ALocH h) (MOuterInterface (getRenamingMembrane infinite_pick ALocH avoid names)).
Proof.
  introv Mh. unfolds getRenamingMembrane.
  destruct (getRenamingMembrane_aux _ _ _ _) as [pM alr] eqn: E. destruct pM as [p M].
  forwards*: getRenamingMembrane_aux_all E.
  unfolds Local_Comparable_ALoc. erewrite MOuterInterface_correct. rew_refl~.
Qed.


(** Computing the membrane needed for [renameEFormula_aux]. **)
Definition renameEFormula_membrane hs_avoid names : Membrane :=
  'let l_names := filter_l names in
  'let k_names := filter_k names in
  'let names_to_change := filter (fun h => mem h hs_avoid) names in
  'let l_names_to_change := filter_l names_to_change in
  'let k_names_to_change := filter_k names_to_change in
  'let membrane_l :=
    getRenamingMembrane _ (id : L → ALoc)
      (filter_l hs_avoid ++ l_names) l_names_to_change in
  'let membrane_k :=
    getRenamingMembrane _ (id : K → ALoc)
      (filter_k hs_avoid ++ k_names) k_names_to_change in
   membrane_l ++ membrane_k.

Definition renameEFormula_membrane_env kes_avoid names : Membrane :=
  'let names_to_change := filter (fun ke => mem ke kes_avoid) names in
  getRenamingMembrane_env (kes_avoid ++ names) names_to_change.

Definition renameEFormula_membrane_complete hs_avoid names kes_avoid Enames : Membrane :=
  renameEFormula_membrane hs_avoid names
  ++ renameEFormula_membrane_env kes_avoid Enames.

Lemma renameEFormula_membrane_renamingMembraneProp : forall hs_avoid names,
  No_duplicates names →
  renamingMembraneProp (renameEFormula_membrane hs_avoid names).
Proof.
  introv ND. unfolds renameEFormula_membrane. repeat let_name.
  rewrite EQmembrane_l, EQmembrane_k. constructors.
   apply Forall_app; apply* getRenamingMembrane_correct.
    introv. compute. autos*.
    introv. compute. autos*.
    introv E. inverts~ E.
    introv. exact I.
    introv E. inverts~ E.
   introv N1 N2 E. forwards [N1'|(m1&E1&N1')]: Nth_app_inv (rm N1); forwards I1: Nth_mem N1'.
    forwards~ F: getRenamingMembrane_check_inner (id : L → ALoc).
      introv E0. inverts~ E0.
      introv. exact I.
     rewrite Forall_iff_forall_mem in F. forwards (ha'&Ea'&Ia'): (rm F); [| clear F ].
      rewrite MInnerInterface_correct. rew_refl. applys~ MInInnerInterface_rewriting I1.
       simpl. rew_refl*.
      forwards~ F: getRenamingMembrane_check_outer (id : L → ALoc).
       rewrite Forall_iff_forall_mem in F. forwards (ha&Ea&Ia): (rm F); [| clear F ].
        lets R: MOuterInterface_correct. unfolds ALoc, Local_ALoc_Comparable, ALoc_comparable. erewrite R. clear R.
         rew_refl. eexists. apply* I1.
        forwards [N2'|(m2&E2&N2')]: Nth_app_inv (rm N2); forwards I2: Nth_mem N2'.
         forwards~ F: getRenamingMembrane_check_inner (id : L → ALoc).
           introv E0. inverts~ E0.
           introv. exact I.
          rewrite Forall_iff_forall_mem in F. forwards (hb'&Eb'&Ib'): (rm F); [| clear F ].
           rewrite MInnerInterface_correct. rew_refl. applys~ MInInnerInterface_rewriting I2.
            simpl. rew_refl*.
           forwards~ F: getRenamingMembrane_check_outer (id : L → ALoc).
            rewrite Forall_iff_forall_mem in F. forwards (hb&Eb&Ib): (rm F); [| clear F ].
             lets R: MOuterInterface_correct. unfolds ALoc, Local_ALoc_Comparable, ALoc_comparable. erewrite R. clear R.
              rew_refl. eexists. apply* I2.
             unfold id in *. substs. applys noMergingMembrane_unique_obj N1' N2'; autos.
              apply* getRenamingMembrane_MembraneCorrect.
               introv E0. inverts~ E0.
               introv. exact I.
               apply~ filter_l_No_duplicates. apply~ filter_No_duplicates. typeclass.
              applys~ getRenamingMembrane_noMergingMembrane (id : L → ALoc).
               introv. exact I.
               introv E0. inverts~ E0.
               apply~ filter_l_No_duplicates. apply~ filter_No_duplicates. typeclass.
              discriminate.
         forwards~ F: getRenamingMembrane_check_inner (id : K → ALoc).
           introv E0. inverts~ E0.
           introv. exact I.
          rewrite Forall_iff_forall_mem in F. forwards (hb'&Eb'&Ib'): (rm F); [| clear F ].
           rewrite MInnerInterface_correct. rew_refl. applys~ MInInnerInterface_rewriting I2.
            simpl. rew_refl*.
            substs. inverts Eb'.
    forwards~ F: getRenamingMembrane_check_inner (id : K → ALoc).
      introv E0. inverts~ E0.
      introv. exact I.
     rewrite Forall_iff_forall_mem in F. forwards (ha'&Ea'&Ia'): (rm F); [| clear F ].
      rewrite MInnerInterface_correct. rew_refl. applys~ MInInnerInterface_rewriting I1.
       simpl. rew_refl*.
      forwards~ F: getRenamingMembrane_check_outer (id : K → ALoc).
       rewrite Forall_iff_forall_mem in F. forwards (ha&Ea&Ia): (rm F); [| clear F ].
        lets R: MOuterInterface_correct. unfolds ALoc, Local_ALoc_Comparable, ALoc_comparable. erewrite R. clear R.
         rew_refl. eexists. apply* I1.
        forwards [N2'|(m2&E2&N2')]: Nth_app_inv (rm N2); forwards I2: Nth_mem N2'.
         forwards~ F: getRenamingMembrane_check_inner (id : L → ALoc).
           introv E0. inverts~ E0.
           introv. exact I.
          rewrite Forall_iff_forall_mem in F. forwards (hb'&Eb'&Ib'): (rm F); [| clear F ].
           rewrite MInnerInterface_correct. rew_refl. applys~ MInInnerInterface_rewriting I2.
            simpl. rew_refl*.
            substs. inverts Eb'.
         forwards~ F: getRenamingMembrane_check_inner (id : K → ALoc).
           introv E0. inverts~ E0.
           introv. exact I.
          rewrite Forall_iff_forall_mem in F. forwards (hb'&Eb'&Ib'): (rm F); [| clear F ].
           rewrite MInnerInterface_correct. rew_refl. applys~ MInInnerInterface_rewriting I2.
            simpl. rew_refl*.
           forwards~ F: getRenamingMembrane_check_outer (id : K → ALoc).
            rewrite Forall_iff_forall_mem in F. forwards (hb&Eb&Ib): (rm F); [| clear F ].
             lets R: MOuterInterface_correct. unfolds ALoc, Local_ALoc_Comparable, ALoc_comparable. erewrite R. clear R.
              rew_refl. eexists. apply* I2.
             unfold id in *. substs. fequals. applys noMergingMembrane_unique_obj N1' N2'; autos.
              apply* getRenamingMembrane_MembraneCorrect.
               introv E0. inverts~ E0.
               introv. exact I.
               apply~ filter_k_No_duplicates. apply~ filter_No_duplicates. typeclass.
              applys~ getRenamingMembrane_noMergingMembrane (id : K → ALoc).
               introv. exact I.
               introv E0. inverts~ E0.
               apply~ filter_k_No_duplicates. apply~ filter_No_duplicates. typeclass.
              discriminate.
Qed.

Lemma renameEFormula_membrane_renamingMembraneProp_env : forall hs_avoid names,
  renamingMembraneProp_env (renameEFormula_membrane_env hs_avoid names).
Proof.
  introv. unfolds renameEFormula_membrane_env. repeat let_name.
  apply getRenamingMembrane_env_correct.
Qed.

Lemma renameEFormula_membrane_complete_noNewMembrane : forall hs_avoid names kes_avoid Enames,
  No_duplicates names →
  noNewMembrane (renameEFormula_membrane_complete hs_avoid names kes_avoid Enames).
Proof.
  introv ND. unfolds renameEFormula_membrane_complete. apply Forall_app.
   apply renamingMembraneProp_noNewMembrane.
    apply~ renameEFormula_membrane_renamingMembraneProp.
   apply renamingMembraneProp_env_noNewMembrane.
    apply renameEFormula_membrane_renamingMembraneProp_env.
Qed.

Lemma renameEFormula_membrane_correct : forall hs_avoid names,
  No_duplicates names →
  Membrane_correct _ _ _ (renameEFormula_membrane hs_avoid names).
Proof.
  (** This lemma does not need these hypotheses, and I clear them to avoid unwanted
    dependencies. **)
  clear ABasic_Gamma ABasic_gamma_bot ABasic_gamma_top KE_infinite_pick Ext_substs Ext_Esubsts ExtNames ExtENames DFormula EFormula_correct.
  introv ND. unfolds renameEFormula_membrane. repeat let_name.
  asserts ND': (No_duplicates names_to_change).
    subst names_to_change. apply~ filter_No_duplicates. typeclass.
  asserts NDl: (No_duplicates l_names_to_change).
    subst l_names_to_change. apply~ filter_l_No_duplicates.
  asserts NDk: (No_duplicates k_names_to_change).
    subst k_names_to_change. apply~ filter_k_No_duplicates.
  constructors.
  (** Uniqueness **)
  introv I1 I2. destruct h as [l|k].
   forwards [N1|(m1&Em1&N1)]: Nth_app_inv I1.
    forwards [N2|(m2&Em2&N2)]: Nth_app_inv I2.
     subst membrane_l. forwards~ MC: getRenamingMembrane_MembraneCorrect (id : L → ALoc) NDl.
       introv I. inverts~ I.
       introv. constructors~.
      applys~ Membrane_unique MC N1 N2.
     rewrite EQmembrane_k in N2. unfold getRenamingMembrane in N2.
      destruct (getRenamingMembrane_aux _ _ _ _) as [pM ?] eqn: E. destruct pM as [? M].
      forwards (?&A&?): getRenamingMembrane_exists_rew E.
        introv A. inverts A.
        applys~ Nth_mem N2.
      inverts A.
    rewrite EQmembrane_k in N1. unfold getRenamingMembrane in N1.
     destruct (getRenamingMembrane_aux _ _ _ _) as [pM ?] eqn: E. destruct pM as [? M].
     forwards (?&A&?): getRenamingMembrane_exists_rew E.
       introv A. inverts A.
       applys~ Nth_mem N1.
     inverts A.
   forwards [N1|(m1&Em1&N1)]: Nth_app_inv I1.
    rewrite EQmembrane_l in N1. unfold getRenamingMembrane in N1.
     destruct (getRenamingMembrane_aux _ _ _ _) as [pM ?] eqn: E. destruct pM as [? M].
     forwards (?&A&?): getRenamingMembrane_exists_rew E.
       introv A. inverts A.
       applys~ Nth_mem N1.
     inverts A.
    forwards [N2|(m2&Em2&N2)]: Nth_app_inv I2.
     rewrite EQmembrane_l in N2. unfold getRenamingMembrane in N2.
      destruct (getRenamingMembrane_aux _ _ _ _) as [pM ?] eqn: E. destruct pM as [? M].
      forwards (?&A&?): getRenamingMembrane_exists_rew E.
        introv A. inverts A.
        applys~ Nth_mem N2.
      inverts A.
     subst membrane_k. forwards~ MC: getRenamingMembrane_MembraneCorrect (id : K → ALoc) NDk.
       introv I. inverts~ I.
       introv. constructors~.
      forwards~: Membrane_unique MC N1 N2. substs~.
  (** Uniqueness env **)
  introv I1 I2. apply Nth_mem with (CA := _) in I1. rewrite mem_app in I1. rew_refl in I1.
  inverts I1 as I1.
   lets~ RP0: getRenamingMembrane_correct.
    forwards~ RP: RP0; try rewrite <- EQmembrane_l in RP.
     introv. exact I.
     introv E0. inverts~ E0.
     apply renamingMembraneProp_kind in RP. rewrite Forall_iff_forall_mem in RP. forwards*: RP I1.
   lets~ RP0: getRenamingMembrane_correct.
    forwards~ RP: RP0 (id : K → ALoc); try rewrite <- EQmembrane_k in RP.
     introv. exact I.
     introv E0. inverts~ E0.
     apply renamingMembraneProp_kind in RP. rewrite Forall_iff_forall_mem in RP. forwards*: RP I1.
  (** Rewriting uniqueness **)
  introv I. rewrite mem_app in I. rew_refl in I. inverts I as I.
   forwards~ MC: getRenamingMembrane_MembraneCorrect (id : L → ALoc) NDl.
     introv I'. inverts~ I'.
     introv. constructors~.
    subst membrane_l. applys~ Membrane_rewrites_unique MC I.
   forwards~ MC: getRenamingMembrane_MembraneCorrect (id : K → ALoc) NDk.
     introv I'. inverts~ I'.
     introv. constructors~.
    subst membrane_k. applys~ Membrane_rewrites_unique MC I.
  (** Rewriting uniqueness env **)
  introv I. rewrite mem_app in I. rew_refl in I. inverts I as I.
   forwards~ RP: getRenamingMembrane_correct; try rewrite <- EQmembrane_l in RP.
    introv. constructors~.
    introv E0. inverts~ E0.
    apply renamingMembraneProp_kind in RP. rewrite Forall_iff_forall_mem in RP. forwards*: RP I.
   forwards~ RP: getRenamingMembrane_correct (id : K → ALoc); try rewrite <- EQmembrane_k in RP.
    introv. constructors~.
    introv E0. inverts~ E0.
    apply renamingMembraneProp_kind in RP. rewrite Forall_iff_forall_mem in RP. forwards*: RP I.
Qed.

Lemma renameEFormula_membrane_env_correct : forall kes_avoid names,
  No_duplicates names →
  Membrane_correct _ _ _ (renameEFormula_membrane_env kes_avoid names).
Proof.
  (** This lemma does not need these hypotheses, and I clear them to avoid unwanted
    dependencies. **)
  clear ABasic_Gamma ABasic_gamma_bot ABasic_gamma_top L_infinite_pick K_infinite_pick Ext_substs Ext_Esubsts ExtNames ExtENames DFormula EFormula_correct.
  introv ND. unfolds renameEFormula_membrane_env. repeat let_name.
  asserts ND': (No_duplicates names_to_change).
    subst names_to_change. apply~ filter_No_duplicates.
  constructors.
  (** Uniqueness **)
  introv I1 I2. apply Nth_mem with (CA := _) in I1.
  forwards~ RP: getRenamingMembrane_env_correct. apply renamingMembraneProp_env_kind in RP.
  rewrite Forall_iff_forall_mem in RP. forwards*: RP I1.
  (** Uniqueness env **)
  introv I1 I2. forwards MC: getRenamingMembrane_env_MembraneCorrect ND'.
  applys~ Membrane_unique_env MC I1 I2.
  (** Rewriting uniqueness **)
  introv I. forwards MC: getRenamingMembrane_env_MembraneCorrect ND'.
  applys~ Membrane_rewrites_unique MC I.
  (** Rewriting uniqueness env **)
  introv I. forwards MC: getRenamingMembrane_env_MembraneCorrect ND'.
  applys~ Membrane_rewrites_unique_env MC I.
Qed.

Lemma renameEFormula_membrane_complete_correct : forall hs_avoid names kes_avoid Enames,
  No_duplicates names →
  No_duplicates Enames →
  Membrane_correct _ _ _ (renameEFormula_membrane_complete hs_avoid names kes_avoid Enames).
Proof.
  introv ND1 ND2. unfolds renameEFormula_membrane_complete. constructors.
  (** Uniqueness **)
  introv I1 I2. forwards [N1|(m1&Em1&N1)]: Nth_app_inv I1.
   forwards [N2|(m2&Em2&N2)]: Nth_app_inv I2.
    applys Membrane_unique N1 N2. apply~ renameEFormula_membrane_correct.
    apply Nth_mem with (CA := _) in N2. forwards F: renameEFormula_membrane_renamingMembraneProp_env.
     apply renamingMembraneProp_env_kind in F. rewrite Forall_iff_forall_mem in F. forwards*: F N2.
   apply Nth_mem with (CA := _) in N1. forwards F: renameEFormula_membrane_renamingMembraneProp_env.
    apply renamingMembraneProp_env_kind in F. rewrite Forall_iff_forall_mem in F. forwards*: F N1.
  (** Uniqueness env **)
  introv I1 I2. forwards [N1|(m1&Em1&N1)]: Nth_app_inv I1.
   apply Nth_mem with (CA := _) in N1. forwards F: renameEFormula_membrane_renamingMembraneProp.
    eassumption.
    apply renamingMembraneProp_kind in F. rewrite Forall_iff_forall_mem in F. forwards*: F N1.
   forwards [N2|(m2&Em2&N2)]: Nth_app_inv I2.
    apply Nth_mem with (CA := _) in N2. forwards F: renameEFormula_membrane_renamingMembraneProp.
     eassumption.
     apply renamingMembraneProp_kind in F. rewrite Forall_iff_forall_mem in F. forwards*: F N2.
    forwards: Membrane_unique_env N1 N2.
     apply~ renameEFormula_membrane_env_correct.
     substs~.
  (** Rewriting uniqueness **)
  introv I. rewrite mem_app in I. rew_refl in I. inverts I as I.
   applys Membrane_rewrites_unique I. apply~ renameEFormula_membrane_correct.
   forwards F: renameEFormula_membrane_renamingMembraneProp_env.
    apply renamingMembraneProp_env_kind in F. rewrite Forall_iff_forall_mem in F. forwards*: F I.
  (** Rewriting uniqueness env **)
  introv I. rewrite mem_app in I. rew_refl in I. inverts I as I.
   forwards F: renameEFormula_membrane_renamingMembraneProp; try eassumption.
    apply renamingMembraneProp_kind in F. rewrite Forall_iff_forall_mem in F. forwards*: F I.
   applys Membrane_rewrites_unique_env I. apply~ renameEFormula_membrane_env_correct.
Qed.

Lemma renameEFormula_membrane_check : forall hs_avoid names,
  incl (MOuterInterface (renameEFormula_membrane hs_avoid names)) names.
Proof.
  introv. apply mem_incl with (H := _). introv I. unfolds renameEFormula_membrane.
  repeat let_name. substs. rewrite MOuterInterface_app in I.
  rewrite mem_app in I. rew_refl in I. inverts I as I.
   forwards F: getRenamingMembrane_check_outer. rewrite Forall_iff_forall_mem in F.
    forwards~ (h'&Eh&Mh): F I. substs. rewrite filter_l_correct with (K_comparable := _) in Mh.
    rewrite filter_mem_eq in Mh. rew_refl* in Mh.
   forwards F: getRenamingMembrane_check_outer. rewrite Forall_iff_forall_mem in F.
    forwards~ (h'&Eh&Mh): F I. substs. rewrite filter_k_correct with (L_comparable := _) in Mh.
    rewrite filter_mem_eq in Mh. rew_refl* in Mh.
Qed.

Lemma renameEFormula_membrane_env_check : forall kes_avoid names,
  incl (MEOuterInterface (renameEFormula_membrane_env kes_avoid names)) names.
Proof.
  introv. unfolds renameEFormula_membrane_env. repeat let_name. substs.
  applys incl_tran getRenamingMembrane_env_check_outer. apply mem_incl with (H := _). introv I.
  rewrite filter_mem_eq in I. rew_refl* in I.
Qed.

Lemma renameEFormula_membrane_Eouter : forall hs_avoid names,
  No_duplicates names →
  MEOuterInterface (renameEFormula_membrane hs_avoid names) = nil.
Proof.
  introv ND. apply nil_mem with (CA := _). introv I. forwards (kes&Mkes): MEOuterInterface_mem_inv I.
  forwards F: renameEFormula_membrane_renamingMembraneProp ND.
  apply renamingMembraneProp_kind in F. rewrite Forall_iff_forall_mem in F. forwards~: F Mkes.
Qed.

Lemma renameEFormula_membrane_env_outer : forall hs_avoid names,
  MOuterInterface (renameEFormula_membrane_env hs_avoid names) = nil.
Proof.
  introv. apply nil_mem with (CA := _). introv I. forwards (kes&Mkes): MOuterInterface_mem_inv I.
  forwards F: renameEFormula_membrane_renamingMembraneProp_env.
  apply renamingMembraneProp_env_kind in F. rewrite Forall_iff_forall_mem in F. forwards~: F Mkes.
Qed.

Lemma renameEFormula_membrane_Einner : forall hs_avoid names,
  No_duplicates names →
  MEInnerInterface (renameEFormula_membrane hs_avoid names) = nil.
Proof.
  introv ND. apply nil_mem with (CA := _). introv I. rewrite MEInnerInterface_correct in I. rew_refl in I.
  inverts I as I1 I2.
   forwards F: renameEFormula_membrane_renamingMembraneProp ND.
    apply renamingMembraneProp_kind in F. rewrite Forall_iff_forall_mem in F. forwards~: F I2.
   forwards F: renameEFormula_membrane_renamingMembraneProp ND.
    apply renamingMembraneProp_kind in F. rewrite Forall_iff_forall_mem in F. forwards~: F I1.
Qed.

Lemma renameEFormula_membrane_env_inner : forall hs_avoid names,
  MInnerInterface (renameEFormula_membrane_env hs_avoid names) = nil.
Proof.
  introv. apply nil_mem with (CA := _). introv I. rewrite MInnerInterface_correct in I. rew_refl in I.
  inverts I as I1 I2.
   forwards F: renameEFormula_membrane_renamingMembraneProp_env.
    apply renamingMembraneProp_env_kind in F. rewrite Forall_iff_forall_mem in F. forwards~: F I2.
   forwards F: renameEFormula_membrane_renamingMembraneProp_env.
    apply renamingMembraneProp_env_kind in F. rewrite Forall_iff_forall_mem in F. forwards~: F I1.
Qed.

Lemma renameEFormula_membrane_inner : forall hs_avoid names h,
  MInInnerInterface _ _ _ h (renameEFormula_membrane hs_avoid names) →
  ~ mem h hs_avoid ∧ ~ mem h names.
Proof.
  introv I. unfolds renameEFormula_membrane. repeat let_name. substs.
  set (MID := MInInnerInterfaceDecidable).
  rewrite <- istrue_decide in I. rewrite <- MInnerInterface_correct in I.
  rewrite MInnerInterface_app in I. rewrite mem_app in I. rew_refl in I. inverts I as I.
   forwards~ F: getRenamingMembrane_check_inner (id : L → ALoc).
    introv E. inverts~ E.
    introv. constructors~.
    rewrite Forall_iff_forall_mem in F. forwards~ (h'&Eh&Mh): F I. substs.
     rewrite mem_app in Mh. rew_refl in Mh. rew_logic in Mh.
     repeat rewrite filter_l_correct with (K_comparable := _) in Mh. autos*.
   forwards~ F: getRenamingMembrane_check_inner (id : K → ALoc).
    introv E. inverts~ E.
    introv. constructors~.
    rewrite Forall_iff_forall_mem in F. forwards~ (h'&Eh&Mh): F I. substs.
     rewrite mem_app in Mh. rew_refl in Mh. rew_logic in Mh.
     repeat rewrite filter_k_correct with (L_comparable := _) in Mh. autos*.
Qed.

Lemma renameEFormula_membrane_env_Einner : forall kes_avoid names ke,
  MEInInnerInterface _ _ _ ke (renameEFormula_membrane_env kes_avoid names) →
  ~ mem ke kes_avoid ∧ ~ mem ke names.
Proof.
  introv I. unfolds renameEFormula_membrane_env. repeat let_name. substs.
  set (MID := MEInInnerInterfaceDecidable).
  rewrite <- istrue_decide in I. rewrite <- MEInnerInterface_correct in I.
  forwards~ F: getRenamingMembrane_env_check_Einner.
  rewrite Forall_iff_forall_mem in F. forwards~ NM: F I.
  splits; introv A; false NM; rewrite mem_app; rew_refl*.
Qed.

Lemma renameEFormula_membrane_all : forall hs_avoid names h,
  mem h hs_avoid →
  mem h names →
  mem h (MOuterInterface (renameEFormula_membrane hs_avoid names)).
Proof.
  introv Mavoid Mnames. unfolds renameEFormula_membrane. repeat let_name. substs.
  rewrite MOuterInterface_app. rewrite mem_app. rew_refl. destruct h as [l|k].
   left. apply~ getRenamingMembrane_all. rewrite filter_l_correct with (K_comparable := _).
    rewrite filter_mem_eq. rew_refl*.
   right. apply~ getRenamingMembrane_all. rewrite filter_k_correct with (L_comparable := _).
    rewrite filter_mem_eq. rew_refl*.
Qed.

Lemma MembraneCompose_renameEFormula_membrane : forall hs_avoid kes_avoid M0 M,
  M = renameEFormula_membrane_complete
        hs_avoid (Remove_duplicates (MInnerInterface M0))
        kes_avoid (Remove_duplicates (MEInnerInterface M0)) →
  MembraneCheckCompose M M0 = Some (MembraneCompose M M0).
Proof.
  introv E. substs. do 2 unfolds. unfolds renameEFormula_membrane_complete.
  cases_if as I; autos~; fold_bool; rew_refl in I. false (rm I). splits.
   apply mem_incl with (H := _). introv I'. rewrite MOuterInterface_app in I'.
    rewrite mem_app in I'. rew_refl in I'. inverts I' as I'.
     forwards I: incl_mem renameEFormula_membrane_check (rm I').
      rewrite~ Remove_duplicates_mem in I.
     rewrite renameEFormula_membrane_env_outer in I'. false*.
   introv Ex. rewrite Exists_iff_exists_mem in Ex. lets (h&M1&M2&NM): (rm Ex).
    rewrite MInnerInterface_app in M1. rewrite mem_app in M1. rew_refl in M1. inverts M1 as M1.
     forwards (NM1&NM2): renameEFormula_membrane_inner.
       set (MID := MInInnerInterfaceDecidable).
       rewrite <- istrue_decide. rewrite* <- MInnerInterface_correct.
      false NM2. rewrite~ Remove_duplicates_mem.
     rewrite renameEFormula_membrane_env_inner in M1. false*.
   apply mem_incl with (H := _). introv I. rewrite MEOuterInterface_app in I.
    rewrite mem_app in I. rew_refl in I. inverts I as I.
     rewrite renameEFormula_membrane_Eouter in I; tryfalse*.
      apply~ Remove_duplicates_spec.
     forwards I': incl_mem renameEFormula_membrane_env_check (rm I).
      rewrite~ Remove_duplicates_mem in I'.
   introv Ex. rewrite Exists_iff_exists_mem in Ex. lets (ke&Mke&Mke'&NMke): (rm Ex).
    rewrite MEInnerInterface_app in Mke. rewrite mem_app in Mke. rew_refl in Mke. inverts Mke as Mke.
     rewrite renameEFormula_membrane_Einner in Mke; tryfalse*.
      apply~ Remove_duplicates_spec.
     rewrite MEInnerInterface_correct in Mke. rew_refl in Mke.
     forwards (NM1&NM2): renameEFormula_membrane_env_Einner Mke.
     false NM2. rewrite~ Remove_duplicates_mem.
Qed.

Lemma MgetDependencies_renameEFormula_membrane_env : forall hs_avoid names h,
  No_duplicates names →
  MgetDependencies _ _ (renameEFormula_membrane_env hs_avoid names) h = nil.
Proof.
  introv ND. apply nil_mem with (CA := _). introv I.
  forwards~ (hs&R&I'): MgetDependencies_correct I.
   apply getRenamingMembrane_env_MembraneCorrect. apply~ filter_No_duplicates.
   rewrite <- Mread_mem in R; autos~.
    forwards F: renameEFormula_membrane_renamingMembraneProp_env.
     apply renamingMembraneProp_env_kind in F. rewrite Forall_iff_forall_mem in F.
     applys~ F R.
    apply getRenamingMembrane_env_MembraneCorrect. apply~ filter_No_duplicates.
Qed.

Lemma MgetDependencies_env_renameEFormula_membrane : forall hs_avoid names ke,
  No_duplicates names →
  MgetDependencies_env _ (renameEFormula_membrane hs_avoid names) ke = nil.
Proof.
  introv ND. apply nil_mem with (CA := _). introv I.
  forwards~ (hs&R&I'): MgetDependencies_env_correct I.
   apply~ renameEFormula_membrane_correct.
   rewrite <- Mread_env_mem in R; autos~.
    forwards F: renameEFormula_membrane_renamingMembraneProp ND.
     apply renamingMembraneProp_kind in F. rewrite Forall_iff_forall_mem in F.
     applys~ F R.
    apply~ renameEFormula_membrane_correct.
Qed.

Lemma renameEFormula_membrane_complete_noMergingMembrane : forall hs_avoid names kes_avoid Enames,
  No_duplicates names →
  No_duplicates Enames →
  noMergingMembrane _ _ _ (renameEFormula_membrane_complete hs_avoid names kes_avoid Enames).
Proof.
  introv ND END. unfolds renameEFormula_membrane_complete. constructors.
   introv. etransitivity.
    apply No_duplicates_double_incl_length.
     typeclass.
     apply MgetDependencies_No_Duplicates.
     apply double_incl_sym. apply MgetDependencies_app.
    rewrite~ MgetDependencies_renameEFormula_membrane_env. rewrite app_nil_r.
     apply~ renamingMembraneProp_noMergingMembrane.
      apply~ renameEFormula_membrane_correct.
      apply~ renameEFormula_membrane_renamingMembraneProp.
   introv. etransitivity.
    apply No_duplicates_double_incl_length.
     typeclass.
     apply MgetDependencies_env_No_Duplicates.
     apply double_incl_sym. apply MgetDependencies_env_app.
    rewrite~ MgetDependencies_env_renameEFormula_membrane. rewrite app_nil_l.
     apply~ renamingMembraneProp_env_noMergingMembrane.
      apply~ renameEFormula_membrane_env_correct.
      apply~ renameEFormula_membrane_renamingMembraneProp_env.
Qed.


(** Renaming the innner identifiers of a formula by avoiding a list of
  such identifiers. **)
Definition renameEFormula_aux Phi hs_avoid kes_avoid : option EFormula :=
  let phi := EFormulaFormula Phi in
  let ext := EFormulaExt Phi in
  let M := EFormulaMembrane Phi in
  'let membrane :=
    renameEFormula_membrane_complete hs_avoid (Remove_duplicates (MInnerInterface M))
      kes_avoid (Remove_duplicates (MEInnerInterface M)) in
  option_case None (fun phi' =>
    LibOption.map (fun ext' =>
      let M' := MembraneCompose membrane M in
        EFormula_make phi' ext' M'
      ) (applyMembrane _ _ _ Ext_substs ExtNames Ext_Esubsts ExtENames membrane ext)
    ) (passThroughMembrane phi membrane).

Lemma renameEFormula_aux_defined : forall Phi hs_avoid kes_avoid,
  renameEFormula_aux Phi hs_avoid kes_avoid ≠ None.
Proof.
  introv. unfolds renameEFormula_aux. repeat let_name.
  destruct Phi as [phi ext M]. simpl. destruct (passThroughMembrane _ _) eqn: E.
   destruct applyMembrane eqn: E'.
    discriminate.
    forwards~: noNewMembrane_applyMembrane E'.
     rewrite EQmembrane. apply renameEFormula_membrane_complete_noNewMembrane.
     apply~ Remove_duplicates_spec.
   rewrite EQmembrane in E. unfolds passThroughMembrane.
    eapply passThroughMembrane_defined in E; try typeclass.
     apply~ renameEFormula_membrane_complete_noNewMembrane.
      apply~ Remove_duplicates_spec.
     apply~ renameEFormula_membrane_complete_noMergingMembrane.
      apply~ Remove_duplicates_spec.
      apply~ Remove_duplicates_spec.
     apply~ renameEFormula_membrane_complete_correct.
      apply~ Remove_duplicates_spec.
      apply~ Remove_duplicates_spec.
Qed.

Definition renameEFormula (Phi : EFormula) (hs_avoid : list ALoc) (kes_avoid : list KE)
    : EFormula.
  destruct (renameEFormula_aux Phi hs_avoid kes_avoid) as [Phi'|] eqn: E.
   exact Phi'.
   false* renameEFormula_aux_defined.
Defined.

Lemma renameEFormula_aux_renameEFormula : forall Phi hs_avoid kes_avoid Phi',
  renameEFormula_aux Phi hs_avoid kes_avoid = Some Phi' →
  renameEFormula Phi hs_avoid kes_avoid = Phi'.
Proof.
  introv E. unfolds. asserts Hyp: (forall A a c b (r : A) F,
      a = Some r →
      match a as a0 return (c = a0 → _) with
      | Some r => fun _ => r
      | None => F
      end b = r).
    clear. introv Eq. substs~.
  apply~ Hyp.
Qed.

Lemma renameEFormula_renameEFormula_aux : forall Phi hs_avoid kes_avoid Phi',
  renameEFormula Phi hs_avoid kes_avoid = Phi' →
  renameEFormula_aux Phi hs_avoid kes_avoid = Some Phi'.
Proof.
  introv E. unfolds in E.
  set_eq <- R: (renameEFormula_aux Phi hs_avoid kes_avoid). destruct R.
   asserts Hyp: (forall A a c b (r : A) F,
       a = Some r →
       match a as a0 return (c = a0 → _) with
       | Some r => fun _ => r
       | None => F
       end b = r).
      clear. introv Eq. substs~.
    fequals. erewrite Hyp in E; autos*.
   false* renameEFormula_aux_defined.
Qed.

Lemma renameEFormula_aux_correct : forall Phi hs_avoid kes_avoid Phi',
  EFormula_correct Phi →
  renameEFormula_aux Phi hs_avoid kes_avoid = Some Phi' →
  EFormula_correct Phi'.
Proof.
  introv C Eq. destruct Phi as [phi ext M]. unfolds in Eq. repeat let_name.
  destruct (passThroughMembrane _ _) as [phi'|] eqn: E1;
    destruct (applyMembrane _ _ _ _) as [ext'|] eqn: E2; inverts Eq. simpls.
  forwards Eq: MembraneCompose_renameEFormula_membrane EQmembrane. constructors.
  (** Membrane **)
  apply MembraneCompose_correct; try typeclass.
   substs. apply~ renameEFormula_membrane_complete_correct; apply~ Remove_duplicates_spec.
   applys EFormula_membrane_correct C.
  (** Naming **)
  simpl. introv I. applys~ MembraneCheckCompose_inner_interface_inner Eq.
   substs. apply~ renameEFormula_membrane_complete_correct; apply~ Remove_duplicates_spec.
   applys~ EFormula_membrane_correct C.
   skip. (* Admitted *)
  (** Naming Env **)
  skip. (* Admitted *)
  (** Naming Ext **)
  simpl. introv I. skip. (* Admitted *)
  (** Naming Ext Env **)
  skip. (* Admitted *)
  (** Rewriting Complete **)
  simpl. introv E I1 I2 S. skip. (* Admitted *)
  (** Rewriting Complete Env **)
  skip. (* Admitted *)
Qed.

Lemma renameEFormula_correct : forall Phi hs_avoid kes_avoid,
  EFormula_correct Phi →
  EFormula_correct (renameEFormula Phi hs_avoid kes_avoid).
Proof.
  introv C. destruct (renameEFormula_aux Phi hs_avoid kes_avoid) eqn: E.
   forwards~: renameEFormula_aux_correct E.
    forwards Eq: renameEFormula_aux_renameEFormula E. rewrite~ Eq.
   forwards*: renameEFormula_aux_defined E.
Qed.

Lemma renameEFormulaInterface : forall Phi hs_avoid kes_avoid Phi',
  Phi' = renameEFormula Phi hs_avoid kes_avoid →
  EFormulaInterface Phi = EFormulaInterface Phi'.
Proof.
  introv E. skip. (* Admitted *)
Qed.

Lemma renameEFormulaInnerInterface : forall Phi hs_avoid kes_avoid Phi' h,
  Phi' = renameEFormula Phi hs_avoid kes_avoid →
  mem h hs_avoid →
  ~ mem h (MInnerInterface (EFormulaMembrane Phi')).
Proof.
  introv E I II. symmetry in E. forwards E': renameEFormula_renameEFormula_aux (rm E).
  unfolds in E'. let_name. forwards [(E1&E2)|(Phi0&E1&E2)]: option_case_out E'; tryfalse.
  symmetry in E2. forwards [(Ea&Eb)|(Phi1&Ea&Eb)]: option_case_out E2; tryfalse.
  inverts Eb. simpl in II. rewrite MInnerInterface_correct in II. rew_refl in II.
  skip. (* Admitted *)
Qed.


(** *** Framing Membranes **)

(** This operation immerses a formula [Phi] in a context given by a
  membrane.  The completeness of the rewriting (i.e., the fact that
  the new formula will respect the properties
  [EFormula_rewriting_complete] and [EFormula_rewriting_complete_env])
  has to be checked to be propagated. **)
Definition frameMembrane (M : Membrane) (Phi : EFormula) : option EFormula :=
  'let locDom := fun h =>
    proj1_sig (locDomain _ _ (EFormulaFormula Phi) h)
  in let domCompare h1 h2 :=
    decide (AObject_domain_compare (locDom h1) (locDom h2))
  in let getInner neutral h cont :=
    match Mread (EFormulaMembrane Phi) h with
    | None => cont h
    | Some nil => neutral
    | Some (h :: _) =>
      (** Because of the invariant [EFormula_rewriting_complete], only
        one check is enough. **)
      cont h
    end
  in let locDom_forall hs :=
    decide (Forall (fun h1 =>
      Forall (fun h2 =>
        getInner true h1 (fun h1' =>
          getInner true h2 (fun h2' =>
            domCompare h1' h2'))) hs) hs)
  in let env_in ke :=
    decide (envInDom (EFormulaFormula Phi) ke)
  in let in_dom_env neutral ke :=
    match Mread_env (EFormulaMembrane Phi) ke with
    | None => env_in ke
    | Some nil => neutral
    | Some (ke :: _) =>
      (** Because of the invariant [EFormula_rewriting_complete_env],
        only one check is enough. **)
      env_in ke
    end
  in let in_dom_exists_env kes :=
    decide (Exists (in_dom_env false) kes)
  in let in_dom_forall_env kes :=
    decide (Forall (in_dom_env true) kes)
  in ifb Forall (fun h =>
         match Mread M h return bool with
         | None => arbitrary
         | Some hs => locDom_forall hs
         end) (MOuterInterface M) then
       ifb Forall (fun ke =>
           match Mread_env M ke return bool with
           | None => arbitrary
           | Some kes =>
             decide (~ in_dom_exists_env kes ∨ in_dom_forall_env kes)
           end) (MEOuterInterface M) then
         LibOption.map (EFormula_with_Membrane Phi)
           (MembraneCheckCompose (EFormulaMembrane Phi) M)
       else None
     else None.

Lemma frameMembrane_correct : forall Phi Phi' M,
  EFormula_correct Phi →
  Membrane_correct _ _ _ M →
  frameMembrane M Phi = Some Phi' →
  EFormula_correct Phi'.
Proof.
  introv C1 C2 E. unfolds in E. let_name. cases_if as F. cases_if as F'.
  fold_bool. rew_refl in F. rew_refl in F'. rewrite Forall_iff_forall_mem in F, F'.
  forwards (M_new&E'&EM_new): LibOption.map_inv (rm E). subst Phi'.
  lets C: E'. eapply MembraneCheckCompose_correct in C; try typeclass;
    try applys~ EFormula_membrane_correct C1.
  forwards E: MembraneCheckCompose_MembraneCompose E'. constructors~.
  (** Naming **)
  simpl. introv I. forwards~ I': EFormula_naming C1 I.
  eapply MembraneCheckCompose_inner_interface_inner; try apply E'; try apply C1; typeclass.
  (** Naming Env **)
  simpl. introv I. forwards~ I': EFormula_naming_env C1 I.
  eapply MembraneCheckCompose_inner_interface_inner_env; try apply E'; try apply C1; typeclass.
  (** Naming Ext **)
  simpl. introv I. forwards~ I': EFormula_naming_ext C1 I.
  eapply MembraneCheckCompose_inner_interface_inner; try apply E'; try apply I; try typeclass.
   apply C1.
  (** Naming Ext Env **)
  simpl. introv I. forwards~ I': EFormula_naming_ext_env C1 I.
  eapply MembraneCheckCompose_inner_interface_inner_env; try apply E'; try apply I'; try apply C1; try typeclass.
  (** Rewriting Complete **)
  simpl. introv Eh I1 I2 S. rewrite <- E in Eh. forwards~ Eq: MembraneCompose_Mread_inv Eh.
   applys~ EFormula_membrane_correct C1.
  rewrites (rm Eq) in I1. forwards~ Eq: MembraneCompose_Mread_inv Eh.
   applys~ EFormula_membrane_correct C1.
  rewrites (rm Eq) in I2. asserts Sh: (AObject_domain_compare (locDom h1) (locDom h2)).
   unfolds ALoc, Mread, Local_Comparable_ALoc.
    lets [(hs1&h1'&hs1'&E1a&I1a&E1b&I1b)|(hs1&E1&I1'&N1)]: (rm I1).
     forwards P: F.
      erewrite MOuterInterface_correct. rew_refl. applys~ Mread_defined_inv E1a.
      rewrite E1a in P. rew_refl in P. rewrite Forall_iff_forall_mem in P.
      forwards F1: P I1a. rewrite Forall_iff_forall_mem in F1. rewrite E1b in F1.
      lets [(hs2&h2'&hs2'&E2a&I2a&E2b&I2b)|(hs2&E2&I2'&N2)]: (rm I2).
       rewrite E1a in E2a. inverts E2a. forwards E2: F1 I2a. rewrite E2b in E2.
        destruct hs1'; tryfalse. destruct hs2'; tryfalse. subst locDom. rew_refl in E2.
        eapply AObject_domain_compare_trans; [| applys AObject_domain_compare_trans E2 ].
         applys~ EFormula_correct_AObject_domain_compare C1 E1b. simpl. rew_refl*.
         applys~ EFormula_correct_AObject_domain_compare C1 E2b. simpl. rew_refl*.
       rewrite E1a in E2. inverts E2. forwards E2: F1 I2'. rewrite <- Mread_None in N2.
        rewrite N2 in E2. destruct hs1'; tryfalse. rew_refl in E2.
        applys AObject_domain_compare_trans E2. subst locDom.
        applys~ EFormula_correct_AObject_domain_compare C1 E1b. simpl. rew_refl*.
     forwards P: F.
      erewrite MOuterInterface_correct. rew_refl. applys~ Mread_defined_inv E1.
      rewrite E1 in P. rew_refl in P. rewrite Forall_iff_forall_mem in P.
      forwards F1: P I1'. rewrite Forall_iff_forall_mem in F1. rewrite <- Mread_None in N1.
      rewrite N1 in F1. lets [(hs2&h2'&hs2'&E2a&I2a&E2b&I2b)|(hs2&E2&I2'&N2)]: (rm I2).
       rewrite E1 in E2a. inverts E2a. forwards E2: F1 I2a. rewrite E2b in E2.
        destruct hs2'; tryfalse. subst locDom. rew_refl in E2.
        applys AObject_domain_compare_trans E2.
        applys~ EFormula_correct_AObject_domain_compare C1 E2b. simpl. rew_refl*.
       rewrite E1 in E2. inverts E2. forwards E2: F1 I2'. rewrite <- Mread_None in N2.
        rewrite N2 in E2. rew_refl~ in E2.
  forwards Eq: Sh f. subst locDom.
  rewrite <- (proj2_sig (locDomain _ _ (EFormulaFormula Phi) h2) f). rewrite <- Eq.
  rewrite~ (proj2_sig (locDomain _ _ (EFormulaFormula Phi) h1) f).
  (** Rewriting Complete Env **)
  simpl. introv Eke I1 I2 S. rewrite <- E in Eke.
  asserts (kes0&Rkes0): (exists kes0, Mread_env M ke = Some kes0).
    forwards~ MR1: MembraneCompose_Mread_env_inv C1 Eke.
    rewrites (rm MR1) in I1. forwards* [(kes0&ke0&kes1&Ekes0&Mkes0&MRke0&Mke1)|(kes0&Ekes0&Mke1&NI)]: (rm I1).
  forwards F'': F'; [| rewrite Rkes0 in F''; rew_refl in F'' ].
    unfolds ALoc, Mread_env. erewrite MEOuterInterface_correct. rew_refl. applys~ Mread_env_defined_inv Rkes0.
  forwards* R: MembraneCompose_Mread_env_cancel_outer Eke Rkes0.
   apply* EFormula_membrane_correct.
   rewrites (rm R) in I1. lets [(ke0&kes1&Mke0&Rke0&Mke1)|(Mke1&Nke1)]: (rm I1); inverts F'' as F''.
    false (rm F''). rewrite Exists_iff_exists_mem. exists ke0. unfolds Mread_env. rewrite Rke0. splits*.
     destruct kes1 as [|kea kesa]; tryfalse. rew_refl.
     unfolds ALoc, Mread_env. applys* EFormula_rewriting_complete_env Rke0. simpl. rew_refl~.
    rewrite Forall_iff_forall_mem in F''.
     forwards~ R: MembraneCompose_Mread_env_cancel_outer Eke Rkes0.
      apply* EFormula_membrane_correct.
      rewrites (rm R) in I2. lets [(ke0'&kes1'&Mke0'&Rke0'&Mke1')|(Mke1'&Nke1')]: (rm I2).
       forwards~ ID': F'' Mke0'. unfolds Mread_env. rewrite Rke0' in ID'. destruct kes1' as [|kea' kesa']; tryfalse.
        rew_refl in ID'. unfolds ALoc, Mread_env. applys~ EFormula_rewriting_complete_env C1 Rke0' ID'. simpl. rew_refl~.
       forwards~ ID': F'' Mke1'. rewrite~ <- Mread_env_None in Nke1'. unfolds Mread_env. rewrite Nke1' in ID'.
        rew_refl~ in ID'.
    false (rm F''). rewrite Exists_iff_exists_mem. exists ke1. splits*.
     rewrite~ <- Mread_env_None in Nke1. unfolds Mread_env. rewrite Nke1. rew_refl~.
    rewrite Forall_iff_forall_mem in F''.
     forwards~ R: MembraneCompose_Mread_env_cancel_outer Eke Rkes0.
      apply* EFormula_membrane_correct.
      rewrites (rm R) in I2. lets [(ke0'&kes1'&Mke0'&Rke0'&Mke1')|(Mke1'&Nke1')]: (rm I2).
       forwards~ ID': F'' Mke0'. unfolds Mread_env. rewrite Rke0' in ID'. destruct kes1' as [|kea' kesa']; tryfalse.
        rew_refl in ID'. unfolds Mread_env. applys~ EFormula_rewriting_complete_env C1 Rke0' ID'. simpl. rew_refl~.
       forwards~ ID': F'' Mke1'. rewrite~ <- Mread_env_None in Nke1'. unfolds Mread_env. rewrite Nke1' in ID'.
        rew_refl~ in ID'.
Qed.

Lemma frameMembraneInterface : forall M Phi Phi',
  frameMembrane M Phi = Some Phi' →
  double_incl (MOuterInterface M) (EFormulaInterface Phi').
Proof.
  introv E. unfolds frameMembrane. let_name. cases_if as I; tryfalse.
  cases_if as I'; tryfalse. destruct Phi as [phi ext M0]. simpls.
  destruct MembraneCheckCompose eqn: E'; tryfalse. inverts E. unfolds EFormulaInterface.
  forwards EC: MembraneCheckCompose_MembraneCompose E'. rewrite <- EC. simpl.
  apply~ mem_double_incl. intro h. extens. unfolds Local_Comparable_ALoc.
  repeat erewrite MOuterInterface_correct. rew_refl. apply~ MembraneCompose_same_interface.
Qed.

Lemma frameMembraneEInterface : forall M Phi Phi',
  frameMembrane M Phi = Some Phi' →
  double_incl (MEOuterInterface M) (EFormulaEInterface Phi').
Proof.
  introv E. unfolds frameMembrane. let_name. cases_if as I; tryfalse.
  cases_if as I'; tryfalse. destruct Phi as [phi ext M0]. simpls.
  destruct MembraneCheckCompose eqn: E'; tryfalse. inverts E. unfolds EFormulaEInterface.
  forwards EC: MembraneCheckCompose_MembraneCompose E'. rewrite <- EC. simpl.
  apply~ mem_double_incl. intro ke. extens. repeat erewrite MEOuterInterface_correct.
  rew_refl. apply~ MembraneCompose_same_interface_env.
Qed.

Lemma frameMembraneInDomain : forall M Phi Phi' h,
  frameMembrane M Phi = Some Phi' →
  EFormulaInDomain _ _ Phi h ↔ EFormulaInDomain _ _ Phi' h.
Proof.
  skip. (* Admitted *)
Qed.

Lemma frameMembraneInEDomain : forall M Phi Phi' ke,
  frameMembrane M Phi = Some Phi' →
  EFormulaInEDomain _ Phi ke ↔ EFormulaInEDomain _ Phi' ke.
Proof.
  skip. (* Admitted *)
Qed.

Definition DFrameMembrane : @DMembrane L K KE _ _ _ → DFormula → option DFormula.
  introv (M&OKM) (Phi&OKPhi). destruct (frameMembrane M Phi) as [Phi'|] eqn: E.
   apply Some. applys exist Phi'. apply* frameMembrane_correct.
   exact None.
Defined.

Notation "M 'D°' Phi" := (DFrameMembrane M Phi)
  (at level 10) : sep_scope.


(** *** Framing Formulae **)

(** This operation stars a formula [Phi] with a context formula [phi],
  living outside the membrane of [Phi].  The formula [phi] has to make
  sense in the interface of [Phi].  If not, it is easy to add a
  membrane to [Phi] refering to the new elements of [phi] using
  [DframeMembrane]. **)
Definition frameFormula (phi : IFormula) (Phi : EFormula) : option EFormula :=
  ifb (incl (IFNames phi) (EFormulaInterface Phi)) then
    'let Phi' := renameEFormula Phi (IFNames phi) (IFENames phi) in
    LibOption.map (fun phi' => EFormula_with_Formula Phi' (phi' * EFormulaFormula Phi'))
      (passThroughMembrane phi (EFormulaMembrane Phi'))
  else None.

Lemma frameFormula_correct : forall Phi Phi' phi,
  EFormula_correct Phi →
  frameFormula phi Phi = Some Phi' →
  EFormula_correct Phi'.
Proof.
  introv C E. unfolds in E. let_name. cases_if as I. fold_bool. rew_refl in I.
  destruct passThroughMembrane as [phi'|] eqn: E'; inverts E. constructors~.
  (** Membrane **)
  forwards C': renameEFormula_correct C. rewrite <- EQPhi'0 in C'.
  destruct Phi'0 as [phi0 ext0 M0]. apply~ C'.
  (** Naming **)
Admitted.

Lemma frameFormulaInDomain : forall phi Phi Phi' h,
  frameFormula phi Phi = Some Phi' →
  EFormulaInDomain _ _ Phi h ↔ EFormulaInDomain _ _ Phi' h.
Proof.
  skip. (* Admitted *)
Qed.

Lemma frameFormulaInEDomain : forall phi Phi Phi' ke,
  frameFormula phi Phi = Some Phi' →
  EFormulaInEDomain _ Phi ke ↔ EFormulaInEDomain _ Phi' ke.
Proof.
  skip. (* Admitted *)
Qed.

Definition DFrameFormula : IFormula → DFormula → option DFormula.
  introv phi (Phi&OKPhi). destruct (frameFormula phi Phi) as [Phi'|] eqn: E.
   apply Some. applys exist Phi'. apply* frameFormula_correct.
   exact None.
Defined.

(* Note how the notation “phi '|' '*' ')' Phi” is not a good idea, hence the “D”. *)
Notation "phi 'D*' Phi" := (DFrameFormula phi Phi)
  (at level 10) : sep_scope.

End General.

